/////////////////////////////////////////////////////////////////////////////
// FTP.cpp

#include "stdafx.h"
#include "base.h"
#include "zcftpres.h"
#include "filename.h"
#include "FTP.h"
#include "RASCtrl.h"

AFX_EXT_CLASS CFTP ftp;

CArray<FTPInfo, FTPInfo&> CFTP::m_arrFTP;
DisConnMode CFTP::m_disconn;
CString CFTP::m_sTextMask;
UINT CFTP::m_transfer;
static TCHAR sFTPURL[] = _T("ftp://");
BOOL CFTP::m_bPassv = FALSE;

CFTP::CFTP()
{
	m_hINET = NULL;
	m_hFTP = NULL;
	m_pInfo = NULL;
	m_uCbMsg = 0;
	m_bIgnoreError = FALSE;
	m_dwError = 0;
}

CFTP::~CFTP()
{
	Close();
}

BOOL CFTP::ParseURL(CString sURL, CString& sServer, CString& sObject)
{
	DWORD dwType;
	INTERNET_PORT nPort;
	if(!AfxParseURL(sURL, dwType, sServer, sObject, nPort)) {
		CString sFTP = sFTPURL;
		sFTP += sURL;
		if(!AfxParseURL(sFTP, dwType, sServer, sObject, nPort))
			return FALSE;
	}
	if(dwType == AFX_INET_SERVICE_FTP) {	// ### 1.05 %20->space if FTP
		sObject.Replace(_T("%20"), _T(" "));
	}
	return TRUE;
}

CString CFTP::GetRoot(CString sURL)
{
	CString sServer, sObject;
	ParseURL(sURL, sServer, sObject);
	return sObject;
}

CString CFTP::GetServer(CString sURL)
{
	CString sServer, sObject;
	ParseURL(sURL, sServer, sObject);
	return sServer;
}

CString CFTP::MakeURL(CString sObject)
{
	CString sURL;
	if(m_pInfo)
		sURL = sFTPURL + GetServer(m_pInfo->m_sServer) + sObject;
	return sURL;
}

BOOL CFTP::IsFTPPath(CString sPath)
{
	return (sPath.GetLength() >= 2 && !CFileName::IsSlash(sPath[0]) && sPath[1] != TCHAR(':'));
}

static UINT ThreadConnect(LPVOID pParam)
{
	CFTP* pFTP = (CFTP*)pParam;
	LPCTSTR pUserID = NULL;
	LPCTSTR pPassword = NULL;
	if(!pFTP->m_pInfo->m_sUserID.IsEmpty()) pUserID = pFTP->m_pInfo->m_sUserID;
	if(!pFTP->m_pInfo->m_sPassword.IsEmpty()) pPassword = pFTP->m_pInfo->m_sPassword;
	HWND hWnd = ::GetForegroundWindow();
	pFTP->m_hFTP = ::InternetConnect(pFTP->m_hINET, pFTP->GetServer(pFTP->m_pInfo->m_sServer), pFTP->m_pInfo->m_nPort, 
		pUserID, pPassword, INTERNET_SERVICE_FTP, pFTP->m_bPassv ? INTERNET_FLAG_PASSIVE : 0, CONTEXTID);
	::SetForegroundWindow(hWnd);
	if(!pFTP->m_hFTP) {
		pFTP->m_dwError = ::GetLastError();
		if(!pFTP->m_bIgnoreError)	// ###1.23
			AfxMessageBox(GetErrorMessage());
	}
	else if(pFTP->m_uCbMsg)
		AfxGetMainWnd()->PostMessage(pFTP->m_uCbMsg);
	return 0;
}

BOOL CFTP::Connect(FTPInfo* pInfo, UINT uCbMsg, BOOL bWait, BOOL bIgnoreError)
{
	m_dwError = 0;
	if(IsConnect()) {
		if(m_pInfo == pInfo) return TRUE;
	}
	Close();
	if(pInfo) m_pInfo = pInfo;
	if(!m_pInfo) return FALSE;
	m_hINET = ::InternetOpen(AfxGetAppName(), INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, NULL);
	if(!m_hINET) return FALSE;
	if(uCbMsg)
		m_uCbMsg = uCbMsg;
	m_bIgnoreError = bIgnoreError;
	if(bWait) {
		ThreadConnect((LPVOID)this);
		return m_hFTP != NULL;
	}
	AfxBeginThread(ThreadConnect, (LPVOID)this);
	return TRUE;
}

void CFTP::Close()
{
	if(m_hFTP) {
		::InternetCloseHandle(m_hFTP);
		m_hFTP = NULL;
		if(m_hINET) {
			::InternetCloseHandle(m_hINET);
			m_hINET = NULL;
		}
	}
}

void CFTP::QueryHangUp()
{
	if(m_disconn != DISCONN_KEEP) {
		if(IsRasConnected(TRUE)) {
			if(m_disconn == DISCONN_HANGUP || m_disconn == DISCONN_QUERY &&
				AfxMessageBox(IDS_QUERYHANGUP, MB_YESNO | MB_ICONQUESTION) == IDYES)
				HangUp();
		}
	}
}

static const TCHAR sRegDefault[] = _T("Settings");
CString ConvertPassword(CString sPwd, BOOL bEncode);

void CFTP::Load()
{
	CWinApp* pApp = AfxGetApp();
	int nCount = (int)pApp->GetProfileInt(sRegDefault, _T("FTPServerCount"), 0);
	m_arrFTP.RemoveAll();	// ###1.21
	for_to(i, nCount) {
		FTPInfo info;
		CString sRegFTP;
		sRegFTP.Format(_T("FTP%d"), i+1);
		info.m_sServer = pApp->GetProfileString(sRegFTP, _T("Server"));
		info.m_sUserID = pApp->GetProfileString(sRegFTP, _T("UserID"));
		info.m_sPassword = ConvertPassword(pApp->GetProfileString(sRegFTP, _T("Password")), FALSE);
		info.m_nPort = pApp->GetProfileInt(sRegFTP, _T("Port"), INTERNET_DEFAULT_FTP_PORT); 
		m_arrFTP.Add(info);
	}
	m_disconn = (DisConnMode)pApp->GetProfileInt(sRegDefault, _T("DisConnect"), DISCONN_KEEP);
	CString sTextMask; sTextMask.LoadString(IDS_TEXTMASKS);
	m_sTextMask = pApp->GetProfileString(sRegDefault, _T("TextMask"), sTextMask);
	m_transfer = pApp->GetProfileInt(sRegDefault, _T("TransferType"), FTP_TRANSFER_TYPE_UNKNOWN);
	m_bPassv = pApp->GetProfileInt(sRegDefault, _T("PassiveFTP"), FALSE);
}

void CFTP::Save()
{
	CWinApp* pApp = AfxGetApp();
	int nFTP = m_arrFTP.GetSize();
	pApp->WriteProfileInt(sRegDefault, _T("FTPServerCount"), nFTP);
	for_to(i, nFTP) {
		CString sRegFTP;
		sRegFTP.Format(_T("FTP%d"), i+1);
		pApp->WriteProfileString(sRegFTP, _T("Server"), m_arrFTP[i].m_sServer);
		pApp->WriteProfileString(sRegFTP, _T("UserID"), m_arrFTP[i].m_sUserID);
		pApp->WriteProfileString(sRegFTP, _T("Password"), ConvertPassword(m_arrFTP[i].m_sPassword, TRUE));
		pApp->WriteProfileInt(sRegFTP, _T("Port"), m_arrFTP[i].m_nPort);
	}
	pApp->WriteProfileInt(sRegDefault, _T("DisConnect"), m_disconn);
	pApp->WriteProfileString(sRegDefault, _T("TextMask"), m_sTextMask);
	pApp->WriteProfileInt(sRegDefault, _T("TransferType"), m_transfer);
	pApp->WriteProfileInt(sRegDefault, _T("PassiveFTP"), m_bPassv);
}

FTPInfo* CFTP::FindServer(CString sServer) 
{
	for_to(i, m_arrFTP.GetSize()) {
//		if(sServer == m_arrFTP[i].m_sServer) {
			if(_tcsncmp(sServer, m_arrFTP[i].m_sServer, sServer.GetLength()) == 0) {	// ###1.20
			return &m_arrFTP[i];
		}
	}
	return NULL;
}

BOOL CFTP::RemoveServer(CString sServer)
{
	if(!IsConnect()) {
		for_to(i, m_arrFTP.GetSize()) {
//			if(sServer == m_arrFTP[i].m_sServer) {
			if (_tcsncmp(sServer, m_arrFTP[i].m_sServer, sServer.GetLength()) == 0) {	// ###1.20
				m_arrFTP.RemoveAt(i);
				return TRUE;
			}
		}
	}
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////

#define PWD_MIN 0x20
#define PWD_MAX 0x7F
#define PWD_RANGE (PWD_MAX - PWD_MIN)

CString ConvertPassword(CString sPwd, BOOL bEncode)
{
	CString sResult;
	if(!sPwd.IsEmpty()) {
		if(bEncode) {
			WORD dwSeed = LOWORD(GetTickCount()) + sPwd[0];
			BYTE b1 = dwSeed % PWD_RANGE;
			BYTE b2 = (dwSeed >> 7) % PWD_RANGE;
			srand((b2 << 7) + b1);
			sResult = CString(TCHAR(b1 + PWD_MIN)) + TCHAR(b2 + PWD_MIN);
			for_to(i, sPwd.GetLength()) {
				int c = sPwd[i] + (rand() % PWD_RANGE);
				if(c >= PWD_MAX) c -= PWD_RANGE;
				sResult += (TCHAR)c;
			}
		} else if(sPwd.GetLength() >= 2) {
			srand(((sPwd[1] - PWD_MIN) << 7) + (sPwd[0] - PWD_MIN));
			for(int i=2; i < sPwd.GetLength(); i++) {
				int c = sPwd[i] - (rand() % PWD_RANGE);
				if(c < PWD_MIN) c += PWD_RANGE;
				sResult += (TCHAR)c;
			}
			TRACE("FTP Password:%s\n", sResult);
		}
	}
	return sResult;
}
