/////////////////////////////////////////////////////////////////////////////
// FTPFile.cpp

#include "stdafx.h"
#include "zcftpres.h"
#include "FTPFile.h"
#include "FTP.h"
#include "ErrorDlg.h"
#include "base.h"
#ifndef _UNICODE
#include <mbstring.h>
#endif
#pragma comment( lib, "wininet" )

//#define DEBUG_VER103

HANDLE CFTPFindFiles::FindFirstFile(LPCTSTR pPath, HANDLE hFTP, WIN32_FIND_DATA* pFindData)
{
	if(!hFTP)
		return CFindFiles::FindFirstFile(pPath, NULL, pFindData);
	else
		return ::FtpFindFirstFile(hFTP, pPath, pFindData, INTERNET_FLAG_RELOAD, CONTEXTID);
}

BOOL CFTPFindFiles::FindNextFile(HANDLE hFindFile, HANDLE hFTP, WIN32_FIND_DATA* pFindData)
{
	if(!hFTP)
		return CFindFiles::FindNextFile(hFindFile, NULL, pFindData);
	else
		return ::InternetFindNextFile(hFindFile, pFindData);
}

void CFTPFindFiles::FindClose(HANDLE hFindFile, HANDLE hFTP)
{
	if(!hFTP)
		::FindClose(hFindFile);
	else
		::InternetCloseHandle(hFindFile);
}

CFTPFile::CFTPFile()
{
	m_hFile = NULL;
	m_hFTP = NULL;
}

CFTPFile::~CFTPFile()
{
	Close();
}

BOOL CFTPFile::Open(LPCTSTR pFile, UINT mode, UINT transfer)
{
	Close();
	DWORD dwAccess = (mode & OF_WRITE) ? GENERIC_WRITE : GENERIC_READ;
//	DWORD dwShareMode = (mode & OF_WRITE) ? 0 : FILE_SHARE_READ;
	DWORD dwShareMode = (mode & OF_WRITE) ? FILE_SHARE_READ : (FILE_SHARE_READ | FILE_SHARE_WRITE); // ###1.03
	DWORD dwCreateFlag = (mode & OF_CREATE) ? CREATE_NEW : (mode & OF_WRITE) ? CREATE_ALWAYS : OPEN_EXISTING;
	if(!m_hFTP) {
		HANDLE  hFile = ::CreateFile(pFile, dwAccess, dwShareMode, NULL, dwCreateFlag, FILE_ATTRIBUTE_NORMAL, NULL);
		if(hFile != INVALID_HANDLE_VALUE)
			m_hFile = hFile;
	} else
		m_hFile = ::FtpOpenFile(m_hFTP, pFile, dwAccess, transfer, CONTEXTID);
	return m_hFile != NULL;
}

void CFTPFile::Close()
{
	if(m_hFile) {
		DWORD dwErr = ::GetLastError();
		if(!m_hFTP)
			::CloseHandle(m_hFile);
		else
			::InternetCloseHandle(m_hFile);
		m_hFile = NULL;
		::SetLastError(dwErr);
	}
}

DWORD CFTPFile::Read(LPVOID pBuff, DWORD dwSize)
{
	ASSERT(m_hFile);
	DWORD dwRead = 0;
	BOOL bResult;
	if(!m_hFTP)
		bResult = ::ReadFile(m_hFile, pBuff, dwSize, &dwRead, NULL);
	else
		bResult = ::InternetReadFile(m_hFile, pBuff, dwSize, &dwRead);
	return bResult ? dwRead : INVALID;
}

DWORD CFTPFile::Write(LPVOID pBuff, DWORD dwSize)
{
	ASSERT(m_hFile);
	DWORD dwWrite = 0;
	BOOL bResult;
	if(!m_hFTP)
		bResult = ::WriteFile(m_hFile, pBuff, dwSize, &dwWrite, NULL);
	else
		bResult = ::InternetWriteFile(m_hFile, pBuff, dwSize, &dwWrite);
	return bResult ? dwWrite : INVALID;
}

BOOL CFTPFile::DeleteFile(LPCTSTR pFile)
{
//	TRACE(_T("Delete %s\n)", pFile);
//	return TRUE;
	if(!m_hFTP)
		return ::DeleteFile( pFile);
	else
		return ::FtpDeleteFile(m_hFTP, pFile);
}

BOOL CFTPFile::CreateDirectory(LPCTSTR pDir)
{
	if(!m_hFTP)
		return ::CreateDirectory(pDir, NULL);
	else
		return ::FtpCreateDirectory(m_hFTP, pDir);
}

BOOL CFTPFile::RemoveDirectory(LPCTSTR pDir)
{
	if(!m_hFTP)
		return ::RemoveDirectory(pDir);
	else
		return ::FtpRemoveDirectory(m_hFTP, pDir);
}

BOOL CFTPFile::Rename(LPCTSTR pPathOld, LPCTSTR pPathNew)
{
	if(!m_hFTP)
		return ::MoveFile(pPathOld, pPathNew);
	else
		return ::FtpRenameFile(m_hFTP, pPathOld, pPathNew);
}

DWORD CFTPFile::GetAttributes(LPCTSTR pFile)
{
	if(!m_hFTP)
		return ::GetFileAttributes(pFile);
	else
		return 0;
}

BOOL CFTPFile::SetAttributes(LPCTSTR pFile, DWORD dwAttr)
{
	if(!m_hFTP)
		return ::SetFileAttributes(pFile, dwAttr);
	else
		return TRUE;
}

BOOL CFTPFile::GetTimes(FileTimes* pTimes)
{
	ASSERT(m_hFile);
	return ::GetFileTime(m_hFile, &pTimes->ftCreation, &pTimes->ftLastAccess, &pTimes->ftLastWrite);
}

BOOL CFTPFile::SetTimes(FileTimes* pTimes)
{
	ASSERT(m_hFile);
	return ::SetFileTime(m_hFile, &pTimes->ftCreation, &pTimes->ftLastAccess, &pTimes->ftLastWrite);
}

CFileName CFTPFile::AddPath(CFileName sPath, LPCTSTR pDir)
{
	if(!sPath.IsEmpty() && pDir) {
		CFileName sDir(pDir);
		sPath.AddSlash(IsFTP());
		if(IsFTP())
			sDir.BackslashToSlash();
		sPath += sDir;
	}
	return sPath;
}

/////////////////////////////////////////////////////////////////////////////
// CFTPFileOp class

#define ERRORBIT_READONLY 1
#define ERRORBIT_FILEEXIST 2

CFTPFileOp::CFTPFileOp()
{
	m_pBuffer = m_pBuffer2 = NULL;
	m_dwBuffer = 32 * 1024;
	m_transfer = FTP_TRANSFER_TYPE_BINARY;
	m_bDone = FALSE;
	m_dwBytesPerCluster = 0;	// ###1.20
	m_dwlTotalBytes = 0;
}

CFTPFileOp::~CFTPFileOp()
{
	if(m_pBuffer)
		MemFree(m_pBuffer);
	if(m_pBuffer2)
		MemFree(m_pBuffer2);
}

void CFTPFileOp::InitFileOp()
{
#ifndef NO_FTP
	ParseTextMask();
#endif
	m_errbits = 0;
	m_bError = FALSE;
	m_bIgnoreError = FALSE;
	m_nTotalFiles = 0;
	m_nDoneFiles = 0;
	m_dwlDoneBytes = 0;
	m_dwlDoneDiskBytes = 0;		// ###1.20
}

void CFTPFileOp::InitSubCounter()
{
	m_dwlSubTotalBytes = m_dwlSubDoneBytes = 0;
	m_nSubTotlaFiles = m_nSubDoneFiles = 0;
}

void CFTPFileOp::SetTotalParameters()
{
	if(m_dwlTotalBytes == 0)
		m_dwlTotalBytes = m_dwBytesPerCluster ? m_dwlDoneDiskBytes : m_dwlDoneBytes;
	m_nTotalFiles = m_nDoneFiles;
	m_nDoneFiles = 0;
	m_dwlDoneBytes = 0;
	m_dwlDoneDiskBytes = 0;
}

BOOL CFTPFileOp::InitTotalDiskBytes(CFileName sPath)	// ###1.20
{
	if(sPath.IsRoot()) {
		DWORD dwTotalClusters, dwFreeClusters, dwBytesPerCluster;
		if(GetDriveInfo(sPath, &dwTotalClusters, &dwFreeClusters, &dwBytesPerCluster)) {
			m_dwlTotalBytes = dwTotalClusters - dwFreeClusters;
			m_dwlTotalBytes *= dwBytesPerCluster;
			m_dwBytesPerCluster = dwBytesPerCluster;
			return TRUE;
		}
	}
	return FALSE;
}

BOOL CFTPFileOp::FileOperate(UINT wFunc, LPCTSTR pSrcDir, LPCTSTR pDstDir, CFindData* pSrcData, CFindData* pDstData)
{
	LPCTSTR pFileName = NULL;
	LPCTSTR pFileNameDst = NULL;
	CFileName sPathName, sPathNameDst;
	if(pSrcData) {
		sPathName = pSrcData->GetSubPathName(m_fileSrc.IsFTP());		// ###1.20
		pFileName = pFileNameDst = sPathName;
		if(pDstData) {
			sPathNameDst = pDstData->GetSubPathName(m_fileDst.IsFTP());
			pFileNameDst = sPathNameDst;
		}
	}
	CFileName sSrcFile = m_fileSrc.AddPath(CFileName(pSrcDir), pFileName);
	CFileName sDstFile = m_fileDst.AddPath(CFileName(pDstDir), pFileNameDst);
ReOperate:
	BOOL bResult = FALSE;
	DWORD dwAttr;
	BOOL bDest = pDstDir && *pDstDir;
	CFileName& sFile = bDest ? sDstFile : sSrcFile;
	LPCTSTR pDir = bDest ? pDstDir : pSrcDir;
	CFTPFile& file = bDest ? m_fileDst : m_fileSrc;
Retry:
	switch(wFunc) {
	case FO_COPY:
		bResult = CopyFile(sSrcFile, sDstFile, pSrcData);
		break;
	case FO_RENAME:
		bResult = ::MoveFile(sSrcFile, sDstFile);
		break;
	case FO_MOVE:
		if(bResult = CopyFile(sSrcFile, sDstFile, pSrcData)) {
			wFunc = FO_DELETE;
			pDstDir = NULL;
			goto ReOperate;
		}
		break;
	case FO_DELETE:
		if(pSrcData) {
			bResult = file.DeleteFile(sFile);
		} else {
			bResult = file.RemoveDirectory(pDir);
			if(!bResult && LOWORD(::GetLastError()) == ERROR_DIR_NOT_EMPTY)
				bResult = TRUE;
		}
		break;
	default:
		bResult = OperateDefault(wFunc, sSrcFile, sDstFile, pSrcData, pDstData);
		break;
	}
	if(IsCanceled()) return FALSE;
	if(!bResult) {
ErrorRedo:
		DWORD dwErr = ::GetLastError();
		m_nErrCode = LOWORD(dwErr);
		UINT nMsg = 0;
		UINT errbit = 0;
		switch(m_nErrCode) {
		case 0:
			if((wFunc == FO_COPY || wFunc == FO_MOVE) && m_fileSrc.IsFTP() && m_fileDst.IsFTP()) // ### 1.23
				nMsg = IDS_ERR_FTPTOOLARGE;
			break;
		case ERROR_ACCESS_DENIED:
			dwAttr = file.GetAttributes(sFile);
			if(dwAttr == UINT_MAX) goto ErrorRedo;
			if(dwAttr & FILE_ATTRIBUTE_READONLY) {
				nMsg = IDS_ERR_READONLY;
				errbit = ERRORBIT_READONLY;
			}
			break;
		case ERROR_ALREADY_EXISTS:
			if(!pSrcData) return FALSE;
			nMsg = IDS_ERR_FILEEXIST;
			errbit = ERRORBIT_FILEEXIST;
			break;
		case ERROR_PATH_NOT_FOUND:	// ###1.24
			if(wFunc == FO_RENAME) {
				CreateDirectory(pDstDir, pSrcDir);
				m_nErrCode = 0;
				goto Retry;
			}
			break;
		}
		int nResult = 0;
//		m_bError = TRUE;
		if(nMsg) {
			if(m_errbits & errbit) nResult = IDYES;
			else nResult = ErrorMessageDlg(pFileName, nMsg);
			switch(nResult) {
			case IDYESALL:
				m_errbits |= errbit;
			case IDYES:
				switch(errbit) {
				case ERRORBIT_READONLY:
					if(!file.SetAttributes(sFile, dwAttr & ~FILE_ATTRIBUTE_READONLY))
						goto ErrorRedo;
					goto Retry;
				case ERRORBIT_FILEEXIST:
					if(!m_fileDst.DeleteFile(sDstFile)) goto ErrorRedo;
					goto Retry;
				}
			case IDNO:
				break;
			case IDCANCEL:
				Cancel();
				break;
			}
		} else {
			m_bError = TRUE;		// ###0.93
			for_to(i, m_dwSkipErrors.GetSize()) {
				if(m_dwSkipErrors[i] == dwErr) { nResult = IDYES; dwErr = 0; break; }
			}
			if(!nResult)
				nResult = ErrorMessageDlg(pFileName);
			switch(nResult) {
			case IDYESALL:
				if(dwErr) m_dwSkipErrors.Add(dwErr);
			case IDYES:
				break;
			case IDNO:
				goto Retry;
			case IDCANCEL:
				Cancel();
				break;
			}
		}
	}
	if(bResult) m_bDone = TRUE;
	return bResult;
}

BOOL CFTPFileOp::OperateDefault(UINT wFunc, LPCTSTR pSrcFile, LPCTSTR pDstFile, CFindData* pSrcData, CFindData* pDstData)
{
	ASSERT(TRUE);
	return TRUE;
}

BOOL CFTPFileOp::CopyFile(LPCTSTR pSrcFile, LPCTSTR pDstFile, CFindData* pSrcData)
{
//	return ::CopyFile(pSrcFile, pDstFile, NULL);
	if(!AllocBuffer()) return FALSE;
	CFileName sFileName;
	if(!_tcsicmp(pSrcFile, pDstFile)) {
		CString sTitle = CFileName(pDstFile).GetNoExt();
		CString sExt = CFileName(pDstFile).GetExt();
		for(int n = 1; ; n++) {
			sFileName.Format(_T("%s%u%s"), sTitle, n, sExt); // ###_UNICODE
			if(!m_fileDst.Open(sFileName, OF_READ, m_transfer)) break;
			m_fileDst.Close();
		}
		pDstFile = sFileName;
	}
//	if(!OpenWriteFile(pSrcFile, pDstFile)) return FALSE; // ###0.93

	BOOL bResult = FALSE;
	DWORDLONG dwlWriteTotal = 0;

	if(!(m_fileSrc.IsFTP() && m_fileDst.IsFTP())) {
		if(m_fileSrc.Open(pSrcFile, OF_READ, m_transfer)) {
			for(;;) {
				DWORD dwRead = m_fileSrc.Read(m_pBuffer, m_dwBuffer);
				if(dwRead == INVALID) break;
				if(dwRead < m_dwBuffer)
					m_fileSrc.Close();
				if(IsCanceled()) break;
				if(!m_fileDst.IsOpened() && !OpenWriteFile(pSrcFile, pDstFile)) break; // ###0.93
				DWORD dwWrite = 0;
				if(dwRead) {	// ###1.04
					dwWrite = m_fileDst.Write(m_pBuffer, dwRead);
					if(dwWrite == INVALID) {
						m_fileDst.Close();		// ###1.02
						m_fileDst.DeleteFile(pDstFile);
						break;
					}
				}
				ASSERT(dwRead == dwWrite);
				dwlWriteTotal += dwWrite;	// ###1.23
				if(dwRead < m_dwBuffer) {
					m_fileDst.SetTimes(&pSrcData->ft);
					m_fileDst.Close();
					m_fileDst.SetAttributes(pDstFile, pSrcData->dwAttr);
					bResult = TRUE;
					break;
				}
				DrawStatus(dwlWriteTotal);
				if(IsCanceled()) break;
			}
			m_fileSrc.Close();
		}	
		m_fileDst.Close();
		if (IsCanceled() && !bResult) { // ###1.25(08.12.16)
			m_fileDst.DeleteFile(pDstFile);
		}
	} else {	// Both FTP ###1.23
		for(;;) {
			if(!m_fileSrc.Open(pSrcFile, OF_READ, m_transfer)) break;
			DWORD dwRead = m_fileSrc.Read(m_pBuffer, m_dwBuffer);
			m_fileSrc.Close();
			if(dwRead == INVALID || dwRead == m_dwBuffer) break;
			if(IsCanceled()) break;
			if(!OpenWriteFile(pSrcFile, pDstFile)) break; // ###0.93
			DWORD dwWrite = 0;
			if(dwRead) {	// ###1.04
				dwWrite = m_fileDst.Write(m_pBuffer, dwRead);
				m_fileDst.SetTimes(&pSrcData->ft);
				m_fileDst.Close();
				if(dwWrite == INVALID) {
					m_fileDst.DeleteFile(pDstFile);
					break;
				}
			}
			ASSERT(dwRead == dwWrite);
			dwlWriteTotal += dwWrite;
			m_fileDst.SetAttributes(pDstFile, pSrcData->dwAttr);
			bResult = TRUE;
			DrawStatus(dwlWriteTotal);
			break;
		}
	}
	return bResult;
}

BOOL CFTPFileOp::OpenWriteFile(LPCTSTR pSrcFile, LPCTSTR pDstFile)
{
Retry:
	if(!m_fileDst.Open(pDstFile, OF_WRITE, m_transfer)) {
		int nErr = LOWORD(::GetLastError());		// ###0.92
		if(nErr == ERROR_PATH_NOT_FOUND || nErr == ERROR_INTERNET_EXTENDED_ERROR
			|| nErr == ERROR_FILE_NOT_FOUND) {		// ###1.03
			CFileName sSrcPath(pSrcFile), sDstPath(pDstFile);
			if(sDstPath.GetParentDirectory()) {
				sSrcPath.GetParentDirectory();
				if(CreateDirectory(sDstPath, sSrcPath)) goto Retry;
			}
		}
#ifdef DEBUG_VER103
		else {
			CString sMsg;
			sMsg.Format("Open for write '%s' failed, (code=%d)", pDstFile, nErr);
			AfxMessageBox(sMsg);
		}
#endif
		return FALSE;
	}
	return TRUE;
}

BOOL CFTPFileOp::AllocBuffer()
{
	if(!m_pBuffer)
		m_pBuffer = MemAlloc(m_dwBuffer);
	return m_pBuffer != NULL;
}

BOOL CFTPFileOp::CreateDirectory(LPCTSTR pDstPath, LPCTSTR pSrcPath)
{
	if(!m_fileDst.CreateDirectory(pDstPath)) {
		int nErr = LOWORD(::GetLastError());		// ###0.92
		if(nErr == ERROR_PATH_NOT_FOUND || nErr == ERROR_INTERNET_EXTENDED_ERROR
			|| nErr == ERROR_FILE_NOT_FOUND) {		// ###1.03
			CFileName sSrcPath;
			CFileName sDstPath(pDstPath);
			if(sDstPath.GetParentDirectory()) {
				if(pSrcPath) {
					sSrcPath = CString(pSrcPath);
					pSrcPath = NULL;
					if(sSrcPath.GetParentDirectory())
						pSrcPath = sSrcPath;
				}
				return CreateDirectory(sDstPath, pSrcPath);
			}
		}
#ifdef DEBUG_VER103
		else {
			CString sMsg;
			sMsg.Format("CreateDirectory '%s' failed, (code=%d)", pDstPath, nErr);
			AfxMessageBox(sMsg);
		}
#endif
		return FALSE;
	}
	if(pSrcPath) {
		DWORD dwAttr = m_fileSrc.GetAttributes(pSrcPath);
		if(dwAttr & ~FILE_ATTRIBUTE_DIRECTORY)
			m_fileDst.SetAttributes(pDstPath, dwAttr);
	}
	return TRUE;
}

#ifndef NO_FTP
void CFTPFileOp::ParseTextMask()
{
	LPCTSTR pMask = ftp.m_sTextMask;
	for(;;) {
		CString sTerm = GetNextTerm(pMask);
		if(sTerm.IsEmpty()) break;
		m_TextMasks.Add(sTerm);
	}
}

UINT CFTPFileOp::GetFileType(CFileName sFile)
{
	if(sFile[0] == TCHAR('.')) return FTP_TRANSFER_TYPE_ASCII;
	for_to(i, m_TextMasks.GetSize()) {
		if(!sFile.CompareMask(m_TextMasks[i], FALSE)) return FTP_TRANSFER_TYPE_ASCII;
	}
	return FTP_TRANSFER_TYPE_BINARY;
}

#endif

CString CFTPFileOp::GetNextTerm(LPCTSTR& pMask)
{
	CString sTerm;
	LPCTSTR p0;
	while(*pMask) {
		TCHAR c = *pMask++;
		if(c == _T(' ')) continue;
		if(c == _T('\"')) {
			p0 = pMask;
			while(*pMask && *pMask != _T('\"')) pMask++;
			if(!*pMask) break;
			if(pMask == p0) { pMask++; continue; }
		} else {
			p0 = pMask - 1;
			while(*pMask && *pMask != _T(' ')) pMask++;
		}
		sTerm = CString(p0, pMask - p0);
		if(*pMask) pMask++;
		break;
	}
	return sTerm;
}

int CFTPFileOp::ErrorMessageBox(LPCTSTR pCaption)
{
	CString sMsg = GetErrorMessage();
//	AddLog(sMsg);
	if(m_bIgnoreError) return IDYESALL;		// ###0.93
	return AfxGetMainWnd()->MessageBox(sMsg, pCaption, MB_ICONHAND + MB_RETRYCANCEL);
}

int CFTPFileOp::ErrorMessageDlg(LPCTSTR pFile, UINT nMsg)
{
	CString sMsg;
	if(!nMsg) {
		sMsg = GetErrorMessage();
//		AddLog(sMsg);
	}
	else
		sMsg.LoadString(nMsg);
	if(m_bIgnoreError) return IDYESALL;		// ###0.93
	CString sCaption;
	sCaption.Format(_T("%s Error (%d)"), m_sOpMode, m_nErrCode);
	CErrorDlg errDlg;
	errDlg.SetInfo(sCaption, pFile, sMsg, !nMsg);
	return errDlg.DoModal();
}

// File Compare ###1.20 <-- CZCFind

static int CompMem(LPCVOID pBuf1, LPCVOID pBuf2, DWORD dwSize);

int CFTPFileOp::CompareFile(LPCTSTR pSrcFile, LPCTSTR pDstFile) // ###1.20
{
	int	nResult = FALSE;	// TRUE:Match, INVALID:Mismatch
	BOOL bDstOpened = FALSE;
	DWORDLONG dwlTotal = 0;
	if(AllocBuffer()) {
		if(m_fileSrc.Open(pSrcFile, OF_READ, m_transfer)) {
			for(;;) {
				DWORD dwSrc = m_fileSrc.Read(m_pBuffer, m_dwBuffer);
				if(dwSrc == 0) {			// ###1.04
					nResult = TRUE;
					break;
				}
				if(dwSrc == INVALID) break;
				if(dwSrc < m_dwBuffer)
					m_fileSrc.Close();
				if(IsCanceled()) break;
				if(!bDstOpened) {
					if(!m_fileDst.Open(pDstFile, OF_READ, m_transfer)) break;
					bDstOpened = TRUE;
				}
				DWORD dwDst = m_fileDst.Read(m_pBuffer2, dwSrc);
				if(dwDst == INVALID) break;
				ASSERT(dwSrc == dwDst);
				if(CompMem(m_pBuffer, m_pBuffer2, dwSrc)) {
					nResult = INVALID;
					break;
				}
				if(dwSrc < m_dwBuffer) {
					nResult = TRUE;
					break;
				}
				dwlTotal += dwDst;
				DrawStatus(dwlTotal);
				if(IsCanceled()) break;
			}
			m_fileSrc.Close();
			m_fileDst.Close();
		}
	}
	return nResult;
}

int CompMem(LPCVOID pBuf1, LPCVOID pBuf2, DWORD dwSize)
{
	int nResult = 1;
	_asm {
			push	esi
			push	edi
			mov		esi, pBuf1
			mov		edi, pBuf2
			mov		ecx, dwSize
			cmp		ecx, 4			// ###0.92
			jg		CompDword
	repe	cmpsb
			jne		MisMatch
			jmp		short Match
CompDword:	mov		edx, ecx
			shr		ecx, 2
			and		edx, 3
			cld
	repe	cmpsd
			jne		MisMatch
			mov		ecx, edx
	repe	cmpsb
			jne		MisMatch
Match:		mov		nResult, 0
MisMatch:
			pop		edi
			pop		esi
	}
	return nResult;
}

