// findfile.cpp : implementation of the CFindFiles class
//

#include "stdafx.h"
#include "FindFile.h"

static SortMode sortmode = SortByName;
static BOOL bReverse = FALSE;

#define POOL_GLOWBY 1024

WIN32_FIND_DATA CFindFiles::m_fdata;

int CFindData::Compare(CFindData* p1, CFindData* p2)
{
	int r = 0;
	switch(sortmode) {
	case SortByType:
		r = _tcsicmp(CFileName(p1->GetFileName()).GetExt(), CFileName(p2->GetFileName()).GetExt());
		break;
	case SortBySize:
		r = (int)(p1->dwlSize - p2->dwlSize);
		break;
	case SortByTime:
		r = ::CompareFileTime(&p1->ft.ftLastWrite, &p2->ft.ftLastWrite);
		break;
	case SortByAttr:
		r = (int)(p1->dwAttr - p2->dwAttr);
		break;
	case SortByFolder:
		r = _tcsicmp(p1->sSubFolder, p2->sSubFolder);
		break;
	}
	if(bReverse)
		r = -r;
	if(r == 0) {
		r = _tcsicmp(p1->GetFileName(), p2->GetFileName());
		if(sortmode == SortByName && bReverse)
			r = -r;
	}
	return r;
};

int CFindData::CompareTo(CFindData* p2, SortMode _sortmode, BOOL _bReverse)
{
	sortmode = _sortmode;
	bReverse = _bReverse; 
	return Compare(this, p2);
}

BOOL CFindData::IsDots(LPCTSTR p)
{
	if(*p == '.') {
		if(*(p+1) == '\0' || *(p+1) == '.' && *(p+2) == '\0') return TRUE;
	}
	return FALSE;
}

void CFindData::SetFileName(LPTSTR pName)
{
	if(_tcslen(pName) < NAME_BUFFER_SIZE - 1) {
		_tcscpy(cFileName, pName);
		sLongName.Empty();
	} else {
		cFileName[0] = TCHAR('\0');
		sLongName = CString(pName);
	}
}

CFileName CFindData::GetSubPathName(BOOL bSlash)		// ###1.20
{
	if(sSubFolder.IsEmpty()) return CFileName(GetFileName());
	CFileName sPathName(sSubFolder);
	sPathName.AddSlash(bSlash);
	sPathName += GetFileName();
	return sPathName;
}

/////////////////////////////////////////////////////////////////////////////

CFindFiles::CFindFiles()
{
	SetSize(0, POOL_GLOWBY);
	m_hFind = NULL;
	m_pHead = NULL;	// set when Sort() called
}

CFindFiles::~CFindFiles()
{
	Init();
}

void CFindFiles::Init()
{
	RemoveAll();
	SetSubFolder();
	Close();
}

void CFindFiles::Close()
{
	if(m_hFind) {
		FindClose(m_hFind, m_hReserve);
		m_hFind = NULL;
	}
}

BOOL CFindFiles::FindAll(LPCTSTR pPath, HANDLE hReserve, UINT type, BOOL* pCanceled)
{
//	TRACE("Find:%s\n", pPath);
	CFindData* pFindData = FindFirst(pPath, hReserve, type);
	if(!pFindData) {
		Close();
		DWORD err = LOWORD(::GetLastError());
		return (err == 0 || err == ERROR_NO_MORE_FILES || err == ERROR_INVALID_HANDLE) ? TRUE : FALSE;
	}
	while(pFindData) {
		pFindData = FindNext();
		if(pCanceled && *pCanceled) break;
	}
	Close();
	return TRUE;
}

CFindData* CFindFiles::FindFirst(LPCTSTR pPath, HANDLE hReserve, UINT type)
{
	if(m_sSubFolder.IsEmpty())
		Init();
	m_hReserve = hReserve;
	m_sFindDir = CFileName(pPath);
	CFileName sPath(pPath);
	if(!m_sSubFolder.IsEmpty()) {	// ###1.20
		sPath.AddSlash((BOOL)m_hReserve);
		sPath += m_sSubFolder;
	}
	if(m_hReserve) {
		if(!::FtpSetCurrentDirectory(m_hReserve, sPath))	// ### 1.05
			return NULL;	// ###1.22
		sPath = _T("*");
	} else if(!(type & FIND_ARC)) {
		sPath.AddSlash(FALSE);
		sPath += _T("*.*");
	}
	m_type = type;
	m_hFind = FindFirstFile(sPath, hReserve, &m_fdata);
	if(!m_hFind || m_hFind == INVALID_HANDLE_VALUE) return NULL;
	return SetFindData();
}

CFindData* CFindFiles::FindNext()
{
	if(!FindNextFile(m_hFind, m_hReserve, &m_fdata)) return NULL;
	return SetFindData();
}

CFindData* CFindFiles::FindFile(LPCTSTR pFile, HANDLE hReserve)
{
	Init();
	m_hReserve = hReserve;
	CFileName m_sFindDir(pFile);
	m_sFindDir.GetParentDirectory();
	m_type = FIND_ALL;
	m_hFind = FindFirstFile(pFile, hReserve, &m_fdata);
	if(!m_hFind || m_hFind == INVALID_HANDLE_VALUE) return NULL;
	CFindData* pFindData = SetFindData();
	Close();
	return pFindData;
}

HANDLE CFindFiles::FindFirstFile(LPCTSTR pPath, HANDLE hReserve, WIN32_FIND_DATA* pFindData)
{
	return ::FindFirstFile(pPath, pFindData);
}

BOOL CFindFiles::FindNextFile(HANDLE hFindFile, HANDLE hReserve, WIN32_FIND_DATA* pFindData)
{
	return ::FindNextFile(hFindFile, pFindData);
}

void CFindFiles::FindClose(HANDLE hFindFile, HANDLE hReserve)
{
	::FindClose(hFindFile);
}

#define FILEATTR_RECYCLED  (FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_DIRECTORY)

CFindData* CFindFiles::SetFindData()
{
	CFindData find;
Redo:
	if(!(m_type & FIND_ARC)) {
		for(;;) {
			BOOL bDir = (m_fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0;
			BOOL bDot = CFindData::IsDots((LPCTSTR)m_fdata.cFileName);
			BOOL bRecycled = ((m_fdata.dwFileAttributes & FILEATTR_RECYCLED) == FILEATTR_RECYCLED);	// ###1.03
			if(bRecycled)
				bRecycled = (_tcsicmp((LPCTSTR)m_fdata.cFileName, _T("RECYCLED")) == 0);
			if(!bRecycled && !(bDir &&  !(m_type & FIND_DIR) || !bDir && !(m_type & FIND_FILE) || bDot && !(m_type & FIND_DOTS)))
				break;
			if(!FindNextFile(m_hFind, m_hReserve, &m_fdata)) return NULL;
		}
		find.dwAttr = m_fdata.dwFileAttributes;
		find.dwSize = m_fdata.nFileSizeLow;
		find.dwSizeHigh = m_fdata.nFileSizeHigh;	// ###1.21
		find.ft.ftCreation = m_fdata.ftCreationTime;
		find.ft.ftLastAccess = m_fdata.ftLastAccessTime;
		find.ft.ftLastWrite = m_fdata.ftLastWriteTime;
		find.SetFileName(m_fdata.cFileName);
		if(!m_sSubFolder.IsEmpty())
			find.sSubFolder = m_sSubFolder;
	}
	if(!PostFind(&find)) {
		if(!FindNextFile(m_hFind, m_hReserve, &m_fdata)) return NULL;
		goto Redo;
	}
	return &ElementAt(Add(find));
}

void CFindFiles::Sort(SortMode _sortmode, BOOL _bReverse)
{
	sortmode = _sortmode;
	bReverse = _bReverse;
	if(GetCount() == 0) return;
	m_pHead = ElementAt(0).Sort(GetCount());
}

CFileName CFindFiles::GetFullPath(CFindData* pData)
{
	CFileName sPath = GetFindDir();
	if(!pData->sSubFolder.IsEmpty()) {	// ###1.20
		sPath.AddSlash((BOOL)m_hReserve);
		sPath += pData->sSubFolder;
	}
	sPath.AddSlash((BOOL)m_hReserve);
	sPath += pData->GetFileName();
	return sPath;
}

int CFindFiles::SearchFile(CFileName sFile)
{
	for_to(i, GetCount()) {
		if(sFile.CompareNoCase(GetFullPath(i)) == 0)
			return i;
	}
	return INVALID;
}

void CFindFiles::ClearMarkAll()
{
	for_to(i, GetCount())
		ElementAt(i).ClearMark();
}
