// findfile.h : interface of the CFindFiles class
//
// ### add subfolder info (03/07/25)

#ifndef FINDFILE_H
#define FINDFILE_H

#include "filename.h"
#include "sortobj.h"

enum SortMode { SortNone, SortByName, SortByType, SortBySize, SortByTime, SortByAttr, SortByFolder };

#define FILEATTR_MARK	0x80000000
#define NAME_BUFFER_SIZE	16

#define FIND_FILE	 1
#define FIND_DIR	 2
#define FIND_DIRFILE 3
#define FIND_DOTS	 4
#define FIND_ALL	 7
#define FIND_ARC	16	// ###1.24

struct FileTimes
{
	FILETIME ftCreation;
	FILETIME ftLastAccess;
	FILETIME ftLastWrite;
};

class CFindData : public CSortObject<CFindData>
{
public:
	DWORD dwAttr;
	union {
		struct {
			DWORD dwSize;
			DWORD dwSizeHigh;	// ###1.21
		};
		DWORDLONG dwlSize;
	};
	DWORD dwPackedSize;		// ###1.24
	DWORD dwRatio;			// ###1.24
	FileTimes ft;
	TCHAR cFileName[NAME_BUFFER_SIZE];
	CString sLongName;
	CString sSubFolder;
	void SetMark() { dwAttr |= FILEATTR_MARK; }
	void ClearMark() { dwAttr &= ~FILEATTR_MARK; }
	BOOL IsReadOnly() { return (dwAttr & FILE_ATTRIBUTE_READONLY) != 0; }
	BOOL IsHidden() { return (dwAttr & FILE_ATTRIBUTE_HIDDEN) != 0; }
	BOOL IsSystem() { return (dwAttr & FILE_ATTRIBUTE_SYSTEM) != 0; }
	BOOL IsDirectory() { return (dwAttr & FILE_ATTRIBUTE_DIRECTORY) != 0; }
	BOOL IsArchive() { return (dwAttr & FILE_ATTRIBUTE_ARCHIVE) != 0; }
	BOOL IsCompressed() { return (dwAttr & FILE_ATTRIBUTE_COMPRESSED) != 0; }

	BOOL IsMarked() { return (dwAttr & FILEATTR_MARK) != 0; }
	BOOL IsDots() { return IsDots(GetFileName()); }
	static BOOL IsDots(LPCTSTR p);
	LPCTSTR GetFileName() { return sLongName.IsEmpty() ? cFileName : (LPCTSTR)sLongName; }
	AFX_EXT_CLASS CFileName GetSubPathName(BOOL bSlash = FALSE);	// ###1.20
	AFX_EXT_CLASS void SetFileName(LPTSTR pName);
	AFX_EXT_CLASS int CompareTo(CFindData* p2, SortMode _sortmode, BOOL _bReverse = FALSE);
protected:
	AFX_EXT_CLASS virtual int Compare(CFindData* p1, CFindData* p2);
};

class AFX_EXT_CLASS CFindFiles : public CArray<CFindData, CFindData&>
{
public:
	CFindFiles();
	virtual ~CFindFiles();
	BOOL FindAll(LPCTSTR pPath, HANDLE hReserve, UINT type = FIND_FILE, BOOL* pCanceled = NULL);
	CFindData* FindFirst(LPCTSTR pPath, HANDLE hReserve, UINT type = FIND_FILE);
	CFindData* FindNext();
	CFindData* FindFile(LPCTSTR pFile, HANDLE hReserve);
	void Close();
	void Init();
	void Sort(SortMode _sortmode, BOOL _bReverse = FALSE);
	void ClearMarkAll();
	CFindData* GetHead() { return m_pHead; }
	int GetCount() { return GetSize(); }
	CFileName GetFindDir() { return m_sFindDir; }
	CFileName GetFullPath(int i) { return GetFullPath(&ElementAt(i)); }
	CFileName GetFullPath(CFindData* pData); // ###1.24
	int SearchFile(CFileName sFile);
	void SetSubFolder(CFileName sFolder) { m_sSubFolder = sFolder; }
	void SetSubFolder() { m_sSubFolder.Empty(); }
	CFileName GetSubFolder() { return m_sSubFolder; }
	virtual HANDLE FindFirstFile(LPCTSTR pPath, HANDLE hReserve, WIN32_FIND_DATA* pFindData);
	virtual BOOL FindNextFile(HANDLE hFindFile, HANDLE hReserve, WIN32_FIND_DATA* pFindData);
	virtual void FindClose(HANDLE hFindFile, HANDLE hReserve);
	virtual BOOL PostFind(CFindData* pData) { return TRUE; }
private:
	static WIN32_FIND_DATA m_fdata;
	HANDLE m_hFind;
	HANDLE m_hReserve;
	CFindData* m_pHead;
	BOOL m_type;
	CFileName m_sFindDir;
	CFindData* SetFindData();
	CFileName m_sSubFolder;
};

#endif
