// LogDlg.cpp : implementation file
//

#include "stdafx.h"
#include "ZCopy.h"
#include "ZCInfo.h"
#include "LogDlg.h"
#include "ZCFind.h"
#include "base.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define PROG_MAX 256
#define IDTIMER 2
#define TIMER_ELAPSE 1000
#define KB_SHIFT 10

/////////////////////////////////////////////////////////////////////////////
// CLogDlg dialog


CLogDlg::CLogDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CLogDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CLogDlg)
	m_bHideDlg = options.m_bHideDlg;
	//}}AFX_DATA_INIT
	m_pParent = pParent;
	m_bInit = FALSE;
	m_nPos = -1;
	m_dwlTotalBytes = m_dwlDoneBytes = 0;
}


void CLogDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CLogDlg)
	DDX_Control(pDX, IDS_PATH, m_stPath);
	DDX_Control(pDX, IDS_FILENAME, m_stFileName);
	DDX_Control(pDX, IDS_COUNTER, m_stCounter);
	DDX_Control(pDX, IDS_REMAIN, m_stRemain);
	DDX_Control(pDX, IDC_PROGRESS, m_progress);
	DDX_Control(pDX, IDE_LOG, m_editLog);
	DDX_Check(pDX, IDB_HIDEDLG, m_bHideDlg);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CLogDlg, CDialog)
	//{{AFX_MSG_MAP(CLogDlg)
	ON_BN_CLICKED(IDB_LOGOPEN, OnLogOpen)
	ON_BN_CLICKED(IDB_HIDEDLG, OnHideDlg)
	ON_WM_TIMER()
	ON_BN_CLICKED(IDB_CANCEL, OnCancel)
	ON_WM_CLOSE()
	//}}AFX_MSG_MAP
	ON_STN_CLICKED(IDS_PATH, &CLogDlg::OnStnClickedPath)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CLogDlg message handlers

UINT ThreadZCopy(LPVOID pParam);

BOOL CLogDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
	CRect rcDlg;
	GetWindowRect(&rcDlg);
	m_nHeight = rcDlg.Height();
	m_bLogOpen = TRUE;
	OpenLog(m_bExec == DO_LIST);
	SetWindowText(m_pInfo->GetTitle());
	m_progress.SetRange(0, PROG_MAX);
	m_editLog.SetTextMode(TM_PLAINTEXT);
//	m_editLog.SetLangOptions(0);
//	m_editLog.SetLineSpacing();
	m_bDone = FALSE;
	m_bCanceled = FALSE;

	SetTimer(IDTIMER, TIMER_ELAPSE, NULL);

	if(m_bHideDlg)
		OnHideDlg();
	AfxBeginThread(ThreadZCopy, (LPVOID)this);
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CLogDlg::OnCancel() 
{
	// TODO: Add your control notification handler code here
	if(m_bDone)
		EndDialog(TRUE);
	else
		m_bCanceled = TRUE;
}

void CLogDlg::OnClose() 
{
	// TODO: Add your message handler code here and/or call default
	EndDialog(FALSE);
//	CDialog::OnClose();
}

void CLogDlg::Done()
{
	m_bDone = TRUE;
	GetDlgItem(IDB_CANCEL)->SetWindowText(_T("Close"));
}

void CLogDlg::ClearLog()
{
	m_editLog.SetReadOnly(FALSE);		// ##1.05
	m_editLog.SetSel(0, -1);
	m_editLog.Clear();
	m_editLog.SetReadOnly(TRUE);
}

void CLogDlg::AddLog(CString sText)
{
	if(m_bInit) return;
//	m_editLog.SetSel(-1, -1);			// ###1.03 ???
	m_editLog.SetSel(-1, 0);
	if(sText.Right(2) != _T("\r\n"))	// ###0.93
		sText += _T("\r\n");
	m_editLog.ReplaceSel(sText);
}

void CLogDlg::InitLog()
{
	ClearLog();
	CString sHead = m_pInfo->GetOpModeString() + _T(" ") + ParseEnv(m_pInfo->m_sSrcPath);
	if(m_pInfo->m_opmode != OPE_DELETE && m_pInfo->m_opmode != OPE_FIND)
		sHead += GetArrow() + ParseEnv(m_pInfo->m_sDstPath);
	AddLog(sHead);
	::GetLocalTime(&m_stStart);
}

void CLogDlg::CloseLog()
{
	::GetLocalTime(&m_stEnd);
}

void CLogDlg::DrawPath(CString sSrcPath, CString sDstPath)
{
	if(m_bInit) return;
	if(!sDstPath.IsEmpty())
		sSrcPath += GetArrow() + sDstPath;
	m_stPath.SetWindowText(sSrcPath);
	m_stFileName.SetWindowText(_T(""));
	// m_progress.ShowWindow(SW_HIDE);
	// m_stCounter.SetWindowText("");
}

CString CLogDlg::GetArrow()
{
	CString sArrow;
	if(m_pInfo->m_opmode != OPE_DELETE && m_pInfo->m_opmode != OPE_FIND)
		sArrow = (m_pInfo->m_opmode == OPE_COMPARE) ? _T(" <-> ") : _T(" --> ");
	return sArrow;
}

void CLogDlg::DrawFileName(LPCTSTR pFileName)
{
	if(m_bInit) return;
	m_stFileName.SetWindowText(pFileName);
}

void CLogDlg::DrawStatus(DWORDLONG dwlTotalBytes, DWORDLONG dwlDoneBytes, int nTotlaFiles, int nDoneFiles)
{
	if(m_bInit) return;
	DrawStatus(dwlTotalBytes, dwlDoneBytes);
	CString sCounter;
	if(nTotlaFiles)
		sCounter.Format(_T("%d/%d"), nDoneFiles, nTotlaFiles);
	else
		sCounter.Format(_T("%d"), nDoneFiles);
	m_stCounter.SetWindowText(sCounter);
}

void CLogDlg::DrawStatus(DWORDLONG dwlTotalBytes, DWORDLONG dwlDoneBytes)
{
	if(m_bInit) return;
	// m_progress.ShowWindow(SW_SHOW);
	// m_dwlTotalBytes = dwlTotalBytes;
	// m_dwlDoneBytes = dwlDoneBytes;

	int nPos = dwlTotalBytes ? (int)(dwlDoneBytes * PROG_MAX / dwlTotalBytes) : PROG_MAX;
	if(nPos != m_nPos) {
		m_nPos = nPos;
	//	dwTotalBytes /= 16;
	//	dwDoneBytes /= 16;
		m_progress.SetPos(nPos);
	}
}

void CLogDlg::OpenLog(BOOL bOpen)
{
	if(bOpen != m_bLogOpen) {
		m_bLogOpen = bOpen;
		RECT rcDlg, rcEdit;
		GetWindowRect(&rcDlg);
		m_editLog.GetWindowRect(&rcEdit);
		rcDlg.bottom = bOpen ? rcDlg.top + m_nHeight : rcEdit.top;
		MoveWindow(&rcDlg);
		m_editLog.ShowWindow(bOpen ? SW_SHOW : SW_HIDE);
	}
	return;
}

void CLogDlg::OnLogOpen() 
{
	// TODO: Add your control notification handler code here
	OpenLog(!m_bLogOpen);
}

void CLogDlg::OnHideDlg() 
{
	// TODO: Add your control notification handler code here
	UpdateData(TRUE);
	options.m_bHideDlg = m_bHideDlg;
	options.Touch();
	m_pParent->ShowWindow(!m_bHideDlg && !m_bTray ? SW_SHOW : SW_HIDE);
}

void CLogDlg::OnTimer(UINT nIDEvent) 
{
	// TODO: Add your message handler code here and/or call default
	
	int nTotalKB = (int)(m_dwlTotalBytes >> KB_SHIFT);
	int nDoneKB  = (int)(m_dwlDoneBytes  >> KB_SHIFT);

	if(m_bInit || nTotalKB <= 0 || nDoneKB <= 0) return;
	DWORD dwTick = ::GetTickCount();
	if(dwTick < m_dwTickStart) {
		m_dwTickStart = dwTick;
		return;
	}
	int nTickRemain = MulDiv(dwTick - m_dwTickStart, nTotalKB - nDoneKB, nDoneKB);
	if(nTickRemain > 0) {
		nTickRemain /= 1000;
		CString sMsg;
		sMsg.Format(IDS_OPERATE_REMAIN, nTickRemain / 60, nTickRemain % 60);
		m_stRemain.SetWindowText(sMsg);
	}
}

void CLogDlg::SetInitializeMode(BOOL bInit)
{
	m_bInit = bInit;
	if(bInit) {
		CString sMsg;
		sMsg.LoadString(IDS_OPERATE_INIT);
		m_stFileName.SetWindowText(sMsg);
		m_progress.ShowWindow(SW_HIDE);
	}
}

void CLogDlg::StartOperation(BOOL bShowProgress)
{
	m_progress.ShowWindow(bShowProgress ? SW_SHOW : SW_HIDE);
	m_dwTickStart = ::GetTickCount();
	m_dwlTotalBytes = 0;
	m_dwlDoneBytes = 0;
}

void CLogDlg::EndOperation()
{
	KillTimer(IDTIMER);
	m_progress.ShowWindow(SW_HIDE);
	m_stRemain.ShowWindow(SW_HIDE);
}

/*
CString GetLogPath()
{
	CString sPath;
	LPTSTR pBuffer = sPath.GetBuffer(_MAX_PATH);
	::GetModuleFileName(NULL, pBuffer, _MAX_PATH - 1);
	LPTSTR p;
	p = _tcsrchr(pBuffer, TCHAR('\\'));
	_tcscpy(p + 1, _T("zcopylog.txt"));
	sPath.ReleaseBuffer();
	return sPath;
}
*/

static void WriteString(CFile& file, CString& str)
{
#ifndef _UNICODE
	file.Write((LPCTSTR)str, str.GetLength());
#else
	CT2A pszA(str);
	file.Write(pszA, str.GetLength());
#endif
}

void CLogDlg::SaveLog()
{
	CFile file;
	if(file.Open(GetLogPath(), CFile::modeCreate | CFile::modeNoTruncate | CFile::modeReadWrite)) {
		TCHAR *pBuffer = NULL;
		DWORD dwSize = (DWORD)file.GetLength();
		if(dwSize) {
			if(options.m_nSaveLogHist > 1) {
				pBuffer = new TCHAR[dwSize + 1];
				file.Read(pBuffer, dwSize);
				*(pBuffer + dwSize) = TCHAR('\0');
				TCHAR *p = pBuffer;
				for_to(i, options.m_nSaveLogHist - 1) {
					p += 6;
					p = _tcsstr(p, _T(">>>>> "));
					if(!p) break;
				}
				if(p)
					dwSize = p - pBuffer;
			}
			file.SetLength(0);
		}
		CString sLog;
		sLog.Format(_T(">>>>> %s [%s] Start\r\n"), GetDateTimeString(&m_stStart), m_pInfo->GetTitle());
		WriteString(file, sLog);
		m_editLog.GetWindowText(sLog);
		WriteString(file, sLog);
		sLog.Format(_T("<<<<< %s [%s] End\r\n\r\n"), GetDateTimeString(&m_stEnd), m_pInfo->GetTitle());
		WriteString(file, sLog);
		if(pBuffer) file.Write(pBuffer, dwSize);		// ###1.04
		::SetEndOfFile((HANDLE)file.m_hFile);
		file.Close();
		if(pBuffer) delete pBuffer;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Thread

UINT ThreadZCopy(LPVOID pParam)
{
	CLogDlg* pLogDlg = (CLogDlg*)pParam;
#ifndef NO_FTP
	if(pLogDlg->m_pInfo->m_option & OPT_FTP && !ftp.IsConnect()) {
		CString sMsg;
		sMsg.Format(_T("Connect to \"%s\" ..."), ftp.m_pInfo->m_sServer);
		pLogDlg->DrawPath(sMsg, "");
		if(!ftp.Connect(NULL, WM_FTP_CONNECT, TRUE, (pLogDlg->m_pInfo->m_option & OPT_IGNORE_ERROR) != 0)) {
			if(pLogDlg->m_bExec == DO_AUTOEXEC && options.m_bSaveLog) {	// ###1.23
				pLogDlg->InitLog();
				CString sMsg;
				sMsg.Format("%s (%d)", GetErrorMessage(ftp.m_dwError), LOWORD(ftp.m_dwError));
				pLogDlg->AddLog(sMsg);
				pLogDlg->CloseLog();
				pLogDlg->SaveLog();
			}
			pLogDlg->EndDialog(0);
			return 0;
		}
	}
#endif
	CZCFind finder(pLogDlg);
	if(finder.Init()) {
		if(pLogDlg->m_bExec && finder.m_bValidTotal)
			finder.CalcTotalBytes();
		pLogDlg->InitLog();
		if(!pLogDlg->m_bExec || !finder.m_bValidTotal || options.m_bNoCalcTotal || finder.m_dwlTotalBytes
			|| !(pLogDlg->m_pInfo->m_option & OPT_SUBDIR)	// ###1.22
			|| pLogDlg->m_pInfo->m_opmode == OPE_SYNC && finder.m_nFileCount) {		// ###1.22
			finder.m_nFileCount = 0;	 // ###1.22
			pLogDlg->StartOperation(pLogDlg->m_bExec && finder.m_bValidTotal);
			finder.ScanDir();
			pLogDlg->EndOperation();
		}
		pLogDlg->CloseLog();
		if(!finder.Done() || pLogDlg->m_bExec == DO_AUTOEXEC)
			// pLogDlg->EndDialog(0);	###1.24
			pLogDlg->PostMessage(WM_CLOSE);
		else
			pLogDlg->Done();
	} else
		// pLogDlg->EndDialog(0);	###1.24
		pLogDlg->PostMessage(WM_CLOSE);
	return 0;
}


void CLogDlg::OnStnClickedPath()
{
	// TODO: ɃRg[ʒmnh[ R[hǉ܂B
}
