/////////////////////////////////////////////////////////////////////////////
// TrayIcon.cpp

#include "stdafx.h"
#include "trayicon.h"
#include <afxpriv.h>		// for AfxLoadString

IMPLEMENT_DYNAMIC(CTrayIcon, CCmdTarget)

CTrayIcon::CTrayIcon()
{
	ZeroMem(m_nid);
	m_bMenuOpened = FALSE;
}

CTrayIcon::~CTrayIcon()
{
	SetIcon(0); // remove icon from system tray
}

//////////////////
// Set notification window. It must created already.
//
void CTrayIcon::Init(CWnd* pNotifyWnd, UINT uID, UINT uCbMsg)
{
	// Initialize NOTIFYICONDATA
	m_nid.cbSize = sizeof(m_nid);
	m_nid.uID = uID;	// never changes after construction

	// Use resource string as tip if there is one
	AfxLoadString(uID, m_nid.szTip, sizeof(m_nid.szTip));
	// If the following assert fails, you're probably
	// calling me before you created your window. Oops.
	ASSERT(pNotifyWnd==NULL || ::IsWindow(pNotifyWnd->GetSafeHwnd()));
	m_nid.hWnd = pNotifyWnd->GetSafeHwnd();

	ASSERT(uCbMsg==0 || uCbMsg>=WM_USER);
	m_nid.uCallbackMessage = uCbMsg;
}

//////////////////
// Common SetIcon for all overloads. 
//
BOOL CTrayIcon::SetIcon(HICON hicon, LPCTSTR lpTip) 
{
	UINT msg;
	m_nid.uFlags = 0;

	// Set the icon
	if (hicon) {
		// Add or replace icon in system tray
		msg = m_nid.hIcon ? NIM_MODIFY : NIM_ADD;
		m_nid.hIcon = hicon;
		m_nid.uFlags |= NIF_ICON;
	} else { // remove icon from tray
		if (m_nid.hIcon==NULL)
			return TRUE;		// already deleted
		msg = NIM_DELETE;
	}

	// Use the tip, if any
	if (lpTip)
		_tcsncpy(m_nid.szTip, lpTip, sizeof(m_nid.szTip));
	if (m_nid.szTip[0])
		m_nid.uFlags |= NIF_TIP;

	// Use callback if any
	if (m_nid.uCallbackMessage && m_nid.hWnd)
		m_nid.uFlags |= NIF_MESSAGE;

	// Do it
	BOOL bRet = Shell_NotifyIcon(msg, &m_nid);
	if (msg==NIM_DELETE || !bRet)
		m_nid.hIcon = NULL;	// failed
	return bRet;
}

void CTrayIcon::PopupMenu(CMenu* pMenu)
{
	if(m_bMenuOpened) return;
	CMenu menu;
	if(pMenu == NULL) {
		if(!menu.LoadMenu(m_nid.uID)) return;
		pMenu = menu.GetSubMenu(0);
		if(!pMenu) return;
	}
	m_bMenuOpened = TRUE;
	CPoint pt;
	GetCursorPos(&pt);
	CWnd* pWnd = CWnd::FromHandle(m_nid.hWnd);
	pWnd->SetForegroundWindow();
	pMenu->TrackPopupMenu(0, pt.x, pt.y, pWnd);
	pWnd->SetForegroundWindow();
	m_bMenuOpened = FALSE;
}
