// ZBListVw.cpp : implementation of the CZBListView class
//

#include "stdafx.h"
#include "ZBrowse.h"
#include "MainFrm.h"
#include "ZBTreeVw.h"
#include "ZBListVw.h"
#include "pidl.h"
#include "base.h"
#include "FTP.h"
#ifndef _UNICODE
#include <mbstring.h>
#else
#include "filename.h"
#endif
#include <afxpriv.h>
#include "ZBFileOp.h"
#include "archiver.h"
//#include "Shlwapi.h"
//#pragma comment( lib, "shlwapi" )

//#define THUMBNAIL_PROVIDER

#ifdef THUMBNAIL_PROVIDER
#define __RPC__in
#define __RPC__out
#define __RPC__in_opt
#define __RPC__inout_opt
#define __RPC__deref_out_opt
#include <Shobjidl.h>
#include <Thumbcache.h>
#endif

//#define XIMAGE

#ifdef XIMAGE
#include "ximage/ximage.h"
#endif

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define WM_FILL_COMPLETE (WM_USER + 10)
#define WM_REFRESH_LIST  (WM_USER + 13)	// ###1.23
#define COLUMN_MAX 8

#define THUMB_CX 96
#define THUMB_CY 80
#define THUMB_ICON_SIZE 48
#define THUMB_BPP 24

enum HeaderInfo { HEAD_NAME, HEAD_SIZE, HEAD_TYPE, HEAD_DATE, HEAD_ATTR, HEAD_USE, HEAD_FOLDER,
				  HEAD_DRVNAME, HEAD_DRVTYPE, HEAD_DRVTOTAL, HEAD_DRVFREE, HEAD_DRVCLUSTER,
				  HEAD_PACKED, HEAD_RATIO, HEADINFO_COUNT };

static int s_nColumnWidth[HEADINFO_COUNT] = { 120, 60, 40, 80, 32, 32, 60, 120, 80, 60, 60, 60, 60, 32 };
static int s_nColumnFormat[HEADINFO_COUNT] = { LVCFMT_LEFT, LVCFMT_RIGHT, LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_RIGHT, LVCFMT_LEFT, 
										     LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_RIGHT, LVCFMT_RIGHT, LVCFMT_RIGHT, LVCFMT_RIGHT, LVCFMT_RIGHT };
static int s_nColumnLabel[HEADINFO_COUNT] = { IDS_HEADER_NAME, IDS_HEADER_SIZE, IDS_HEADER_TYPE, IDS_HEADER_DATE, IDS_HEADER_ATTR, IDS_HEADER_USE, IDS_HEADER_FOLDER,
										    IDS_HEADER_NAME, IDS_HEADER_TYPE, IDS_HEADER_TOTAL, IDS_HEADER_FREE, IDS_HEADER_CLUSTER, IDS_HEADER_PACKED, IDS_HEADER_RATIO };
static int s_nColumnIndex[HEADINFO_COUNT];

static int s_nColumnCount[LISTTYPE_MAX] = { 5, 6, 6, 5, 8 };

static int s_nColumnTable[LISTTYPE_MAX][COLUMN_MAX] = {
	{ HEAD_NAME, HEAD_SIZE, HEAD_TYPE, HEAD_DATE, HEAD_ATTR, -1, -1, -1 },
	{ HEAD_NAME, HEAD_SIZE, HEAD_USE, HEAD_TYPE, HEAD_DATE, HEAD_ATTR, -1, -1 },
	{ HEAD_NAME, HEAD_FOLDER, HEAD_SIZE, HEAD_TYPE, HEAD_DATE, HEAD_ATTR, -1, -1 },
	{ HEAD_DRVNAME, HEAD_DRVTYPE, HEAD_DRVTOTAL, HEAD_DRVFREE, HEAD_DRVCLUSTER, -1, -1, -1 },
	{ HEAD_NAME, HEAD_FOLDER, HEAD_SIZE, HEAD_PACKED, HEAD_RATIO, HEAD_TYPE, HEAD_DATE, HEAD_ATTR },
};

static int s_nColumnOrder[LISTTYPE_MAX][COLUMN_MAX];

static TCHAR *sDriveType[] = { 
	_T(""), _T(""), _T("Removable Disk"), _T("Fixed Disk"), _T("Remote Drive"), _T("CD-ROM"), _T("RAM Disk")
};

static struct FillRequest
{
	BOOL bRequest;
	BOOL bDrive;
	CString sPath;
	BOOL bFTP;
	LPSHELLFOLDER pSF;
	LPITEMIDLIST pIDLParent;
//	BOOL bDoneMessage;
	UINT uIDRequest;
	BOOL bThread;
	BOOL bArchive;
} req;

#ifndef NO_FTP
CFTP ftp2;
#endif
CArchiver m_archiver;	// ###1.24
static HANDLE m_hFTP = NULL;

/////////////////////////////////////////////////////////////////////////////
// CZBListView diagnostics

#ifdef _DEBUG
void CZBListView::AssertValid() const
{
	CListView::AssertValid();
}

void CZBListView::Dump(CDumpContext& dc) const
{
	CListView::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CZBListView

IMPLEMENT_DYNCREATE(CZBListView, CListView)

BEGIN_MESSAGE_MAP(CZBListView, CListView)
	//{{AFX_MSG_MAP(CZBListView)
	ON_WM_DESTROY()
	ON_NOTIFY_REFLECT(LVN_COLUMNCLICK, OnColumnClick)
	ON_NOTIFY_REFLECT(NM_DBLCLK, OnDblclk)
	ON_NOTIFY_REFLECT(LVN_DELETEITEM, OnDeleteItem)
	ON_NOTIFY_REFLECT(LVN_ENDLABELEDIT, OnEndLabelEdit)
	ON_WM_SETCURSOR()
	ON_COMMAND(ID_FILE_REMOVE, OnFileRemove)
	ON_UPDATE_COMMAND_UI(ID_FILE_REMOVE, OnUpdateFileRemove)
	ON_NOTIFY_REFLECT(LVN_BEGINDRAG, OnBeginDrag)
	ON_WM_CREATE()
	ON_WM_KEYDOWN()
	ON_NOTIFY_REFLECT(NM_CLICK, OnClick)
	ON_WM_MOUSEACTIVATE()
	ON_NOTIFY_REFLECT(NM_RCLICK, OnRClick)
	ON_COMMAND(ID_FILE_RENAME, OnFileRename)
	ON_UPDATE_COMMAND_UI(ID_FILE_RENAME, OnUpdateFileRename)
	ON_COMMAND(ID_FILE_CONTEXTMENU, OnFileContextMenu)
	ON_NOTIFY_REFLECT(LVN_GETDISPINFO, OnGetdispinfo)
	ON_COMMAND(ID_FILE_ENTER, OnFileEnter)
	ON_WM_VSCROLL()
	//}}AFX_MSG_MAP
	ON_MESSAGE(WM_FILL_COMPLETE, OnFillComplete)
	ON_MESSAGE(WM_FILEOP_COMPLETE, OnFileOpComplete)
	ON_MESSAGE(WM_REFRESH_LIST, OnRefreshList)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CZBListView construction/destruction

CZBListView* me; // for ComapreProc

CZBListView::CZBListView()
{
	m_listtype = LIST_NONE;
	m_nSort = 0;
	m_bReverse = FALSE;
	m_pTreeSize = NULL;
	m_bTreeSize = FALSE;
	m_bBusy = FALSE;
	me = this;
	req.bRequest = FALSE;
	m_pDragFinder = NULL;
	m_bRefresh = TRUE;
	m_hChangeNotify = NULL;
	m_bUnfold = m_bTaskBusy = m_bNotSort = FALSE;
	req.uIDRequest = 0;
	m_iPrevTopIndex = 0;
	m_pDrawThread = NULL;	// ###1.24
}

CZBListView::~CZBListView()
{
	if(m_pTreeSize) delete m_pTreeSize;
	if(m_hChangeNotify) ::FindCloseChangeNotification(m_hChangeNotify);
	if(m_pDrawThread) delete m_pDrawThread;
}

BOOL CZBListView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs
	m_style = (options.m_viewStyle & LVS_REPORT) ? LISTVIEW_REPORT : LISTVIEW_ICON;
	cs.style |= options.m_viewStyle & LVS_TYPEMASK | LVS_AUTOARRANGE;	// ###1.24
	return CListView::PreCreateWindow(cs);
}

int CZBListView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CListView::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	// TODO: Add your specialized creation code here
	CListCtrl& list = GetListCtrl();
	HWND hListView = list.GetSafeHwnd();

	m_dropTarget.Register(this);
	m_hCsrBusy = AfxGetApp()->LoadStandardCursor(IDC_WAIT);

	ListView_SetImageList(hListView, GetShellImageList(SHGFI_SMALLICON), LVSIL_SMALL);
	ListView_SetImageList(hListView, GetShellImageList(SHGFI_LARGEICON), LVSIL_NORMAL);

	m_imageThumbs.Create(THUMB_CX, THUMB_CY, ILC_COLOR24, 0, 16);

	DWORD dwStyle = LVS_EX_HEADERDRAGDROP;
	if(options.m_viewStyle & STYLE_FULLROW)
		dwStyle |= LVS_EX_FULLROWSELECT;
	list.SetExtendedStyle(dwStyle);

	SetHeaderImageList();
//	LoadProfiles();
	return 0;
}

void CZBListView::OnDestroy() 
{
	ListView_SetImageList(GetListCtrl().GetSafeHwnd(), NULL, LVSIL_SMALL);
	ListView_SetImageList(GetListCtrl().GetSafeHwnd(), NULL, LVSIL_NORMAL);
	CListView::OnDestroy();
	
	// TODO: Add your message handler code here
	GetColumnWidth();

	DeleteTempFiles();
//	SaveProfiles();	
}

void CZBListView::SetHeaderImageList() // ###1.24
{
	CHeaderCtrl* pHeader = GetListCtrl().GetHeaderCtrl( );
	if(pHeader) {
		if(m_imageSort.m_hImageList == NULL)
			m_imageSort.Create(IDB_SORT, 16, 2, RGB(192, 192, 192));
		pHeader->SetImageList(&m_imageSort);
	}
}

void CZBListView::SetStyle(UINT style)
{
	options.m_viewStyle = options.m_viewStyle & ~LVS_TYPEMASK | style;
	::SetWindowLong(GetListCtrl().GetSafeHwnd(), GWL_STYLE, (GetListCtrl().GetStyle() & ~LVS_TYPEMASK) | style); 
}

void CZBListView::SetListStyle(UINT style, BOOL bAuto) // ###1.24
{
	CListCtrl& list = GetListCtrl();
	if(style != GetListStyle()) {
		if(m_style == LISTVIEW_REPORT)
			GetColumnWidth();
		if(bAuto) list.DeleteAllItems();
		SetStyle((style == LISTVIEW_REPORT) ? LVS_REPORT : LVS_ICON);
		if(style == LISTVIEW_REPORT)
			SetHeaderImageList();
		if(style == LISTVIEW_THUMBNAIL || m_style & LISTVIEW_THUMBNAIL) {
			if(style == LISTVIEW_THUMBNAIL) {
				list.SetImageList(&m_imageThumbs, LVSIL_NORMAL);
				m_style |= LISTVIEW_THUMBNAIL;
			} else {
				ListView_SetImageList(list.GetSafeHwnd(), GetShellImageList(SHGFI_LARGEICON), LVSIL_NORMAL);
				m_style = style;
			}
			if(!bAuto) {
				SetThumbPath();
				FillItems(NULL, m_hFTP != NULL);
			}
		} else
			m_style = style;
	}
}

/////////////////////////////////////////////////////////////////////////////
// CZBTreeView initializing

void CZBOptions::LoadListView()
{
	GetProfileBinary(_T("ColumnWidth"), s_nColumnWidth);
	for_to(i, LISTTYPE_MAX)
		for_to(j, COLUMN_MAX)
			s_nColumnOrder[i][j] = j;
	GetProfileBinary(_T("ColumnOrder"), s_nColumnOrder, sizeof(s_nColumnOrder));
}

void CZBOptions::SaveListView()
{
	WriteProfileBinary(_T("ColumnWidth"), (LPBYTE)s_nColumnWidth, sizeof(s_nColumnWidth));
	WriteProfileBinary(_T("ColumnOrder"), (LPBYTE)s_nColumnOrder, sizeof(s_nColumnOrder));
}

/////////////////////////////////////////////////////////////////////////////
// CZBFindFiles class (for Achiver ###1.24)

static CStringArray m_sarrFileMask;
static CDWordArray m_arrMaskFlag;
static BOOL m_bFileMask = FALSE;

class CZBFindFiles : public CFTPFindFiles
{
public:
	CZBFindFiles();

	virtual HANDLE FindFirstFile(LPCTSTR pPath, HANDLE hFTP, WIN32_FIND_DATA* pFindData);
	virtual BOOL FindNextFile(HANDLE hFindFile, HANDLE hFTP, WIN32_FIND_DATA* pFindData);
	virtual void FindClose(HANDLE hFindFile, HANDLE hFTP);
	virtual BOOL PostFind(CFindData* pData);

	BOOL IsArchive() { return m_bArchive; }
	BOOL m_bArchive;
};

static CZBFindFiles m_finder;

CZBFindFiles::CZBFindFiles()
{
	m_bArchive = FALSE;
}

HANDLE CZBFindFiles::FindFirstFile(LPCTSTR pPath, HANDLE hFTP, WIN32_FIND_DATA* pFindData)
{
	if(!IsArchive())
		return CFTPFindFiles::FindFirstFile(pPath, hFTP, pFindData);
	HANDLE hResult = (HANDLE)m_archiver.OpenArchive(AfxGetMainWnd()->GetSafeHwnd(), pPath);
	if(!hResult) return NULL;
	int result = m_archiver.FindFirst();
	if(result == TRUE) return hResult;
	return NULL;
}

BOOL CZBFindFiles::FindNextFile(HANDLE hFindFile, HANDLE hFTP, WIN32_FIND_DATA* pFindData)
{
	if(!IsArchive())
		return CFTPFindFiles::FindNextFile(hFindFile, hFTP, pFindData);
	return m_archiver.FindNext() == TRUE;
}

void CZBFindFiles::FindClose(HANDLE hFindFile, HANDLE hFTP)
{
	if(!IsArchive())
		CFTPFindFiles::FindClose(hFindFile, hFTP);
	else
		m_archiver.CloseArchive();
}

BOOL CZBFindFiles::PostFind(CFindData* pData)
{
	CString sSubFolder;

	if(IsArchive()) {
		LPCTSTR pName = m_archiver.GetFileName();
		int nFolder = m_archiver.GetFileFolderLength();
		if(nFolder) {
			pData->sSubFolder = CString(pName, nFolder);
			pName += nFolder;
			if(*pName == TCHAR('\0')) return FALSE;
		}
		pData->SetFileName((LPTSTR)pName);
		pData->dwAttr = 0;
		pData->dwSize = m_archiver.GetFileSize();
		pData->dwSizeHigh = 0;
		pData->dwPackedSize = m_archiver.GetFilePackedSize();
		pData->dwRatio = m_archiver.GetFileRatio();
		m_archiver.GetFileTime(pData->ft.ftLastWrite);
		pData->dwAttr = m_archiver.GetFileAttributes();
	}
	if(!m_bFileMask || pData->IsDirectory()) return TRUE;
	CFileName sFile = pData->GetFileName();
	for_to(i, m_sarrFileMask.GetSize()) {
		if(!sFile.CompareMask(m_sarrFileMask[i], FALSE, m_arrMaskFlag[i]))
			return TRUE;
	}
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
// FillItemsThread

struct ShellInfo
{
	CString sDisplayName;
	CString sTypeName;
	int iIcon;
	int iIconAlt;
};

static CArray<ShellInfo, ShellInfo&> s_arrShellInfo;

UINT CZBListView::FillItemsThread(LPVOID pParam)
{
	BOOL bResult = TRUE;
	CZBListView* pView = (CZBListView*)pParam;
	CListCtrl& list = pView->GetListCtrl();
Redo:
	s_arrShellInfo.RemoveAll();
	if(req.bDrive) {
		pView->Init(LIST_DRIVE);

		LPSHELLFOLDER pSF = req.pSF;
		LPITEMIDLIST pIDLParent = req.pIDLParent;
		req.bRequest = FALSE;
		LPITEMIDLIST pIDL = NULL;
		LPENUMIDLIST pEnumIDL = NULL;
		int nItemCount = 0;
		if(SUCCEEDED(pSF->EnumObjects(AfxGetMainWnd()->m_hWnd, SHCONTF_FOLDERS | SHCONTF_NONFOLDERS, &pEnumIDL)) &&  pEnumIDL) {
			while(pEnumIDL->Next(1, &pIDL, NULL) == NOERROR && !req.bRequest) {
				LPITEMIDLIST pIDLAbs = ConcatPIDLs(pIDLParent, pIDL);
				if(pView->InsertDriveItem(pIDLAbs))
					list.SortItems(CompareProc, 0);
				else
					g_pMalloc->Free(pIDLAbs);
				g_pMalloc->Free(pIDL);
				pIDL = NULL;
				nItemCount++;
			}
			pEnumIDL->Release();	
		}
	} else {
		if(!req.sPath.IsEmpty()) pView->m_sPath = req.sPath;

		BOOL bFTP = req.bFTP;
		req.bRequest = FALSE;
//		req.bDoneMessage = FALSE;

		pView->m_bTreeSize = pView->m_pTreeSize && pView->FindNode(pView->m_sPath);

		if(pView->IsUnfold()) {
			if(m_finder.GetSubFolder().IsEmpty())
				pView->Init(LIST_UNFOLD);
		}
		else
			pView->Init(req.bArchive ? LIST_ARC : pView->m_bTreeSize ? LIST_TREE : LIST_FILE);

		bResult = FillSubFolders(pView, CFileName(""));
	}
	if(req.bRequest) goto Redo;
	if(bResult)
		pView->PostMessage(WM_FILL_COMPLETE);
	if(req.uIDRequest) {
		GetMainFrame()->PostMessage(WM_COMMAND, req.uIDRequest);
		req.uIDRequest = 0;
	}
	return bResult;
}

BOOL CZBListView::FillSubFolders(CZBListView* pView, CFileName sSubFolder)
{
	BOOL bFTP = req.bFTP;
	req.bRequest = FALSE;

#ifndef NO_FTP
	if (bFTP) {
		CString sURL = pView->m_sPath;
		CString sServer, sPath;
		if(!ftp2.ParseURL(sURL, sServer, sPath)) return FALSE;
		if(!ftp2.Connect(ftp2.FindServer(sServer), 0, TRUE)) return FALSE;
		if(req.bRequest) return FALSE;
		pView->m_sPath = sPath;
		m_hFTP = ftp2.m_hFTP;
	} else
#endif
		m_hFTP = NULL;


	int nCountPrev = m_finder.GetCount();
	m_finder.SetSubFolder(sSubFolder);
	m_finder.m_bArchive = req.bArchive;
	if(!m_finder.FindAll(pView->m_sPath, m_hFTP, FIND_DIRFILE | (FIND_ARC * req.bArchive), &req.bRequest)) {
		AfxMessageBox(GetErrorMessage(), MB_ICONHAND);
		return FALSE;
	}
	if(req.bRequest) return FALSE;

	if(pView->m_bTreeSize) {	// ###1.20
		for_to(i, m_finder.GetCount()) {
			CFindData& data = m_finder[i];
			if(data.IsDirectory()) {
				CTreeNode* pNode = pView->FindNode(m_finder.GetFullPath(i));
				if(pNode) {
					data.dwSize = (DWORD)((pNode->m_dwlFileSize + 1023)/ 1024);
				}
			}
		}
	}

	CStringArray sarrSubFolders;
	if(pView->IsUnfold()) {
		for(int i = nCountPrev; i < m_finder.GetCount(); i++) {
			CFindData& data = m_finder[i];
			if(data.IsDirectory()) {
				sarrSubFolders.Add(data.GetFileName());
				m_finder.RemoveAt(i--);
			}
		}
	}

	if(pView->IsUnfold()) {
		for_to(i, sarrSubFolders.GetSize()) {
			CFileName sFolder = sSubFolder;
			if(!sFolder.IsEmpty())
				sFolder.AddSlash(bFTP);
			sFolder += sarrSubFolders[i];
			if(!FillSubFolders(pView, sFolder)) return FALSE;
		}
		if(nCountPrev < m_finder.GetCount()) {
			pView->OnFillComplete(TRUE, 0);
	/*		if(!req.bDoneMessage) {
				req.bDoneMessage = TRUE;
				pView->PostMessage(WM_FILL_COMPLETE, TRUE);
			} */
		}
		if(!req.bThread) {
			if(!((CZBrowseApp*)AfxGetApp())->MessageLoop()) return FALSE;
		}
		if(!pView->IsTaskBusy()) return FALSE;
	}
	return TRUE;
}

LRESULT CZBListView::OnFillComplete(WPARAM bUnfold, LPARAM)
{
	if(IsFileList()) {
		CListCtrl& list = GetListCtrl();
//		BOOL bRealName = m_hFTP || (options.m_viewStyle & STYLE_REALNAME);

		int nShiPrev = s_arrShellInfo.GetSize();
		s_arrShellInfo.SetSize(m_finder.GetCount());
		for(int i = nShiPrev; i < s_arrShellInfo.GetSize(); i++) {
			s_arrShellInfo[i].iIcon = INVALID;
			s_arrShellInfo[i].iIconAlt = INVALID;
		}

		list.SetItemCount(m_finder.GetCount());
		if(IsThumbnail())
			m_imageThumbs.SetImageCount(m_finder.GetCount());

		LV_ITEM lvitem;
		lvitem.iSubItem = 0;
		lvitem.pszText = LPSTR_TEXTCALLBACK;
		lvitem.iImage = I_IMAGECALLBACK;

		int nCountPrev = list.GetItemCount();
		int nCountPage = list.GetCountPerPage();

		for(i = nCountPrev; i < m_finder.GetCount(); i++) {
			lvitem.iItem = i;
			lvitem.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;
			lvitem.lParam = i;
			if(!m_sPrevSelect.IsEmpty()) {
				if(!m_sPrevSelect.CompareNoCase((LPCTSTR)m_finder[i].GetFileName())) {
					lvitem.mask |= LVIF_STATE;
					lvitem.state = lvitem.stateMask = LVIS_SELECTED | LVIS_FOCUSED;
				}
			}
			lvitem.iItem = list.InsertItem(&lvitem);
		}
//		s_arrShellInfo.RemoveAll();
		if(!bUnfold) {
			m_bNotSort = FALSE;
			DrawHeaderImage();
			list.SortItems(CompareProc, 0);
			int iSel = GetFirstSelectedItem();
			if(iSel != INVALID) {
				list.EnsureVisible(iSel, FALSE);
			} else if(m_iPrevTopIndex > 0) {	// ###1.20
				list.EnsureVisible(m_iPrevTopIndex + list.GetCountPerPage() - 1, TRUE);
				m_iPrevTopIndex = 0;
			}
		} /* else {
			if(nCountPrev < nCountPage) {
		//	list.EnsureVisible(0, FALSE);
				TRACE(_T("*** Count %d/%d/%d\n"), nCountPrev, m_finder.GetCount(), nCountPage);
				list.RedrawItems(nCountPrev, nCountPage);
				list.UpdateWindow();
			} 
		} */
	}
	m_sPrevSelect.Empty();
	if(m_bBusy) {
		m_bBusy = FALSE;
		PostMessage(WM_SETCURSOR);
	}
	if(!bUnfold) {
		m_bTaskBusy = FALSE;
	}
#ifndef NO_FTP
	if (req.bFTP && !ftp.IsConnect())
		GetTreeView()->FillFTPItems(&m_finder);	// ### 1.23
#endif
	DisplayInfo();
	// req.bDoneMessage = FALSE;
	if(!bUnfold && m_finder.GetCount() && !(m_hFTP || (m_finder.IsArchive()) && !IsThumbnail())) {
		m_pDrawThread = AfxBeginThread(DrawIconsThread, (LPVOID)this, THREAD_PRIORITY_NORMAL, 0, CREATE_SUSPENDED);
		m_pDrawThread->m_bAutoDelete = FALSE;
		m_pDrawThread->ResumeThread();
	}
	return 1;
}

void CZBListView::OnGetdispinfo(NMHDR* pNMHDR, LRESULT* pResult) 
{
	LV_DISPINFO* pDispInfo = (LV_DISPINFO*)pNMHDR;
	// TODO: Add your control notification handler code here

	static  CFileName sText;
	if(IsFileList()) {
		int i = pDispInfo->item.lParam;
		if(i != INVALID && !m_bBusy) {
			CFindData& data = m_finder[i];
			CFileName sFullPath = m_finder.GetFullPath(i);
			ShellInfo& shi = s_arrShellInfo[i];

			if(shi.iIcon == INVALID) {
				SHFILEINFO sfi;
				if(!IsThumbnail()) {
//					if(m_hFTP || m_finder.IsArchive())
						::SHGetFileInfo(data.GetFileName(), data.dwAttr, &sfi, sizeof(SHFILEINFO), SHGFI_USEFILEATTRIBUTES | SHGFI_DISPLAYNAME | SHGFI_TYPENAME | SHGFI_SYSICONINDEX);
//					else
//						::SHGetFileInfo(sFullPath, 0, &sfi, sizeof(SHFILEINFO), SHGFI_DISPLAYNAME | SHGFI_TYPENAME | SHGFI_SYSICONINDEX);
					shi.sDisplayName = (m_hFTP || (options.m_viewStyle & STYLE_REALNAME)) ? (LPCTSTR)data.GetFileName() : sfi.szDisplayName;
					shi.sTypeName = sfi.szTypeName;
	//				shi.iIcon = sfi.iIcon < m_iSortImage ? sfi.iIcon : sfi.iIcon + 2;
					shi.iIcon = sfi.iIcon;
				} else {
/*					if(!data.IsDirectory())
						shi.iIcon = ExtractThumbImage(sFullPath);
					if(shi.iIcon == INVALID) {
						DWORD dwAttr = (m_hFTP || m_finder.IsArchive()) ? SHGFI_USEFILEATTRIBUTES : SHGFI_DISPLAYNAME;
						::SHGetFileInfo(sFullPath, data.dwAttr, &sfi, sizeof(SHFILEINFO), dwAttr | SHGFI_ICON);
						shi.iIcon = m_imageThumbs.Add(sfi.hIcon);
					}
*/				}
			}
			if(pDispInfo->item.mask & LVIF_IMAGE){
				//  Fill in the image information.
				pDispInfo->item.iImage = (shi.iIconAlt != INVALID) ? shi.iIconAlt : shi.iIcon;
			}

			if(pDispInfo->item.mask & LVIF_TEXT){
				//  Fill in the text information.
				switch (m_pnColumnTable[pDispInfo->item.iSubItem]){
				case HEAD_NAME:
					//  Copy the main item text.
					pDispInfo->item.pszText = IsThumbnail() ? (LPTSTR)(LPCTSTR)data.GetFileName() : (LPTSTR)(LPCTSTR)shi.sDisplayName;
					break;
				case HEAD_SIZE:
					if(data.IsDirectory()) {
						if(m_bTreeSize) {
						//	CTreeNode* pNode = FindNode(m_finder.GetFullPath(i));
						//	if(pNode) {
						//		data.dwSize = (DWORD)((pNode->m_dwlFileSize + 1023)/ 1024);
								sText = SeparateByComma(data.dwSize);
								sText += _T("KB");
						//	}
						} else
							sText.Empty();
					} else
						sText = FormatFileSize(data.dwlSize);
					if(!sText.IsEmpty())
						pDispInfo->item.pszText = (LPTSTR)(LPCTSTR)sText;
					break;
				case HEAD_TYPE:
					//  Copy subitem 1 text.
					pDispInfo->item.pszText = (LPTSTR)(LPCTSTR)shi.sTypeName;
					break;
				case HEAD_DATE:
					sText = FormatFileTime(&data.ft.ftLastWrite);
					pDispInfo->item.pszText = (LPTSTR)(LPCTSTR)sText;
					break;
				case HEAD_ATTR:
					sText.Empty();
					if(data.IsReadOnly()) sText += TCHAR('R');
					if(data.IsHidden()) sText += TCHAR('H');
					if(data.IsSystem()) sText += TCHAR('S');
					if(data.IsArchive()) sText += TCHAR('A');
					if(data.IsCompressed()) sText += TCHAR('C');
					pDispInfo->item.pszText = (LPTSTR)(LPCTSTR)sText;
					break;
				case HEAD_USE:
					if(data.IsDirectory()) {
						CTreeNode* pNode = FindNode(m_finder.GetFullPath(i));
						if(pNode && pNode->m_dwlDiskSize) {
							sText.Format(_T("%u%%"), (DWORD)(pNode->m_dwlFileSize * 100 / pNode->m_dwlDiskSize));
							pDispInfo->item.pszText = (LPTSTR)(LPCTSTR)sText;
						}
					}
					break;
				case HEAD_FOLDER:
					sText = data.sSubFolder;
					if(sText.IsEmpty())
						sText = _T(".");
					else
						sText.AddSlash(req.bFTP);
					pDispInfo->item.pszText = (LPTSTR)(LPCTSTR)sText;
					break;
				case HEAD_PACKED:
					sText = FormatFileSize(data.dwPackedSize);
					pDispInfo->item.pszText = (LPTSTR)(LPCTSTR)sText;
					break;
				case HEAD_RATIO:
					sText.Format(_T("%d%%"), data.dwRatio / 10);
					pDispInfo->item.pszText = (LPTSTR)(LPCTSTR)sText;
					break;
				}
			}
		}
	}
	*pResult = 0;
}

void CZBListView::FillItems(LPCTSTR pPath, BOOL bFTP)
{
	if(pPath != NULL)
		TRACE(_T("List:FillItems %s\n"), pPath);
	if(pPath) {
		if(req.bArchive = IsArchiveFile(CFileName(pPath), TRUE)) {	// ###1.24
			SetUnfold(FALSE);
			// m_bFileMask = FALSE;
		}
		if(!bFTP)
			options.m_sBrowseFolder = pPath;
		if(req.sPath.CompareNoCase(pPath) == 0) {
			if(!m_bRefresh) return;
		} else {
			if(m_hChangeNotify) {
				::FindCloseChangeNotification(m_hChangeNotify);
				m_hChangeNotify = NULL;
			}
			if(!bFTP && !req.bArchive) {
				m_hChangeNotify = ::FindFirstChangeNotification(pPath, FALSE,
					FILE_NOTIFY_CHANGE_FILE_NAME |
					FILE_NOTIFY_CHANGE_DIR_NAME |
					FILE_NOTIFY_CHANGE_ATTRIBUTES |
					FILE_NOTIFY_CHANGE_SIZE |
					FILE_NOTIFY_CHANGE_LAST_WRITE |
					FILE_NOTIFY_CHANGE_CREATION
				);
				if(m_hChangeNotify == INVALID_HANDLE_VALUE)
					m_hChangeNotify = NULL;
			}
		}
		req.sPath = pPath;
		BOOL bThumbView = IsThumbPath(pPath);
//		if(bThumbView && !IsThumbnail()) SetListStyle(LISTVIEW_THUMBNAIL, TRUE);
//		else if(!bThumbView && IsThumbnail())  SetListStyle(m_style & ~LISTVIEW_THUMBNAIL, TRUE);

	}
	m_bRefresh = FALSE;
	req.bThread = TRUE;
	req.bDrive = FALSE;
	req.bFTP = bFTP;
	req.bRequest = TRUE;
	req.uIDRequest = 0;
	m_bNotSort = FALSE;
	GetMainFrame()->SetCaption(pPath);
	GetMainFrame()->ReportMessage(IDS_INFO_LISTING, req.sPath);
	m_finder.SetSubFolder();
	TerminateThread(&m_pDrawThread);	// ###1.24
	if(!m_bBusy) {
		if(IsUnfold()) {
			m_bTaskBusy = TRUE;
			m_bNotSort = TRUE;
			if(!bFTP) req.bThread = FALSE;
		}
		if(req.bThread) {
			m_bBusy = TRUE;
			TRACE(_T("*** List:AfxBeginThread %s\n"), req.sPath);
			AfxBeginThread(FillItemsThread, (LPVOID)this);
		} else {
			GetMainFrame()->OnUpdateCmdUI();
			TRACE(_T("*** List:FillItemsThread %s\n"), req.sPath);
			FillItemsThread((LPVOID)this);
		}
	}
}

LRESULT CZBListView::OnRefreshList(WPARAM hItem, LPARAM)
{
	FillItems(NULL, m_hFTP != NULL);
	return 1;
}

void CZBListView::OnViewUnfold()
{
	if(m_bTaskBusy) {
		StopOperation();
		req.uIDRequest = ID_VIEW_UNFOLD;
	} else {
		AfxGetApp()->OnIdle(0);
		SetUnfold(!IsUnfold());
		FillItems(NULL, m_hFTP != NULL);
	}
}

void CZBListView::FillDriveItems(LPCTSTR pPath, LPSHELLFOLDER pSF, LPITEMIDLIST pIDLParent)
{
	if(m_hChangeNotify) {
		::FindCloseChangeNotification(m_hChangeNotify);
		m_hChangeNotify = NULL;
	}
	options.m_sBrowseFolder.Empty();
	req.bDrive = TRUE;
	req.sPath = pPath;
	req.pSF = pSF;
	req.pIDLParent = pIDLParent;
	req.bRequest = TRUE;
	GetMainFrame()->SetCaption(pPath, FALSE);
	if(!m_bBusy) {
		m_bBusy = TRUE;
		AfxBeginThread(FillItemsThread, (LPVOID)this);
	}
	GetMainFrame()->SetStatusBarText(1, _T(""));
}

BOOL CZBListView::InsertDriveItem(LPITEMIDLIST pIDL)
{
	CListCtrl& list = GetListCtrl();

	SHFILEINFO sfi;
	::SHGetFileInfo((LPCTSTR)pIDL, 0, &sfi, sizeof(SHFILEINFO), SHGFI_PIDL | SHGFI_DISPLAYNAME | SHGFI_ATTRIBUTES | SHGFI_SYSICONINDEX);

	if(sfi.dwAttributes & SFGAO_FILESYSANCESTOR) {
		LV_ITEM lvitem;
		lvitem.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;
		lvitem.iItem = lvitem.iSubItem = 0;
		lvitem.pszText = sfi.szDisplayName;
		lvitem.iImage = sfi.iIcon;
		lvitem.lParam = (LONG)pIDL;
		lvitem.iItem = list.InsertItem(&lvitem);
		lvitem.mask = LVIF_TEXT;

		TCHAR buff[MAX_PATH];
		if(::SHGetPathFromIDList(pIDL, buff)) {
			UINT nType = GetDriveType(buff);
			lvitem.pszText = sDriveType[nType];
			lvitem.iSubItem = s_nColumnIndex[HEAD_DRVTYPE];
			list.SetItem(&lvitem);
			if(!(sfi.dwAttributes & SFGAO_REMOVABLE)) {
				CString sTotal, sFree, sUnit;
				DWORD dwBytesPerCluster;
				if(GetDriveInfoString(buff, sTotal, sFree, &dwBytesPerCluster)) {
					lvitem.pszText = (LPTSTR)(LPCTSTR)sTotal;
					lvitem.iSubItem = s_nColumnIndex[HEAD_DRVTOTAL];
					list.SetItem(&lvitem);
					lvitem.pszText = (LPTSTR)(LPCTSTR)sFree;
					lvitem.iSubItem = s_nColumnIndex[HEAD_DRVFREE];
					list.SetItem(&lvitem);
					sUnit = SeparateByComma(dwBytesPerCluster);
					lvitem.pszText = (LPTSTR)(LPCTSTR)sUnit;
					lvitem.iSubItem = s_nColumnIndex[HEAD_DRVCLUSTER];
					list.SetItem(&lvitem);
				}
			}
		}
		return TRUE;
	}
	return FALSE;
}

void CZBListView::SetFileMask(LPCTSTR pMask)
{
	m_bFileMask = FALSE;
	if(pMask) {
		m_sFileMasks = pMask;
		m_sarrFileMask.RemoveAll();
		m_arrMaskFlag.RemoveAll();
		CString sMask = pMask;
		LPTSTR p = sMask.GetBuffer(0);
		for(;;) {
			p = _tcstok(p, _T(" ;"));
			if(!p) break;
			CString s;
			UINT uFlags = GetMBStringType(p);		// ###1.20
			m_arrMaskFlag.Add(uFlags);
			if(!(uFlags & (MBST_WILDCARD | MBST_PATH))) {
			/*
				if(uFlags & MBST_MB) {
					if(uFlags & MBST_MB_ASCII) {
						s = HanToZen((LPCTSTR)p, 0, TRUE);
						p = (LPTSTR)(LPCTSTR)s;
					}
				} else
			*/
					_tcslwr(p);
			}
			m_sarrFileMask.Add(p);
			m_bFileMask = TRUE;
			p = NULL;
		}

		sMask.ReleaseBuffer();
	} else
		m_sFileMasks.Empty();
}

void CZBListView::OnEditFilemask()
{
	if(!m_bFileMask && m_sarrFileMask.GetSize() == 0) {
		CWnd* pWnd = &GetMainFrame()->m_wndToolBar.m_combo;
		pWnd->SetWindowText(_T(""));
		pWnd->SetFocus();
	} else {
		m_bFileMask = !m_bFileMask;
		FillItems(NULL);
	}
}

BOOL CZBListView::IsFileMaskMode()
{
	return m_bFileMask;
}

/////////////////////////////////////////////////////////////////////////////
// new listview

void CZBListView::SetPrevSelect(LPCTSTR pNextPath) // ###1.24
{
	int len = _tcslen(pNextPath);
	if(len < m_sPath.GetLength() && _tcsnicmp(m_sPath, pNextPath, len) == 0) {
		m_sPrevSelect = (CFileName)m_sPath.Mid(len + 1);
	}
}

void CZBListView::Init(ListType type)
{
	CListCtrl& list = GetListCtrl();
	list.DeleteAllItems();

	if(type != m_listtype) {
		int iSortHeader = HEAD_NAME;
		if(m_listtype != LIST_NONE) {
			GetColumnWidth();
			for_to(i, s_nColumnCount[m_listtype])
				list.DeleteColumn(0);
			iSortHeader = s_nColumnTable[m_listtype][m_nSort];
		}

		for_to(i, HEADINFO_COUNT)
			s_nColumnIndex[i] = -1;
		int nCount = s_nColumnCount[type];
		m_pnColumnTable = s_nColumnTable[type];
		for_to_(i, nCount) {
			int iHeader = m_pnColumnTable[i];
			s_nColumnIndex[iHeader] = i;
			LV_COLUMN lvcol;
			lvcol.mask = LVCF_TEXT | LVCF_WIDTH | LVCF_FMT;
			lvcol.fmt = s_nColumnFormat[iHeader];
			CString sLabel; sLabel.LoadString(s_nColumnLabel[iHeader]);
			lvcol.pszText = (LPTSTR)(LPCTSTR)sLabel;
			lvcol.cx = s_nColumnWidth[iHeader];
			list.InsertColumn(i, &lvcol);
		}
		m_listtype = type;
		m_nSort = s_nColumnIndex[iSortHeader];
		if(m_nSort == -1) m_nSort = 0;
		list.SetColumnOrderArray(nCount, s_nColumnOrder[m_listtype]);
	}
	LONG lStyle = list.GetStyle();
	if(IsFileList())
		lStyle |= LVS_EDITLABELS;
	else
		lStyle &= ~LVS_EDITLABELS;
	::SetWindowLong(list.GetSafeHwnd(), GWL_STYLE, lStyle);
	DrawHeaderImage();
	m_finder.Init();
	if(IsThumbnail())
		m_imageThumbs.SetImageCount(0);
}

void CZBListView::DrawHeaderImage()
{
	CListCtrl& list = GetListCtrl();
	int nCount = s_nColumnCount[m_listtype];
	for_to(i, nCount) {
		LV_COLUMN lvcol;
		lvcol.mask = LVCF_FMT;
		lvcol.fmt = s_nColumnFormat[m_pnColumnTable[i]];
		if(m_nSort == i && !m_bNotSort) {
			lvcol.mask = LVCF_FMT | LVCF_IMAGE;
			lvcol.fmt |= LVCFMT_BITMAP_ON_RIGHT | LVCFMT_IMAGE ;
			lvcol.iImage = m_bReverse;
		}
		list.SetColumn(i, &lvcol);

	}
}

void CZBListView::GetColumnWidth()
{
	if(m_listtype != LIST_NONE && options.m_viewStyle & LVS_REPORT) {
		int nCount = s_nColumnCount[m_listtype];
		for_to(i, nCount) {
			s_nColumnWidth[m_pnColumnTable[i]] = GetListCtrl().GetColumnWidth(i);
		}
		GetListCtrl().GetColumnOrderArray(s_nColumnOrder[m_listtype]);
	}
}

#define BUFFER_MAX 128

CString CZBListView::FormatFileSize(DWORDLONG dwlSize)
{
	return SeparateByComma(dwlSize);
}

CString CZBListView::FormatFileTime(FILETIME* pFT)
{
	FILETIME ftLocal;
	SYSTEMTIME st;
	FileTimeToLocalFileTime(pFT, &ftLocal);
	FileTimeToSystemTime(&ftLocal, &st);
	CString sResult;
	LPTSTR pBuff = sResult.GetBuffer(BUFFER_MAX);
    GetDateFormat(LOCALE_USER_DEFAULT, DATE_SHORTDATE, &st, NULL, pBuff, BUFFER_MAX);
	int len = _tcslen(pBuff);
	pBuff += len;
	*pBuff++ = TCHAR(' ');
    GetTimeFormat(LOCALE_USER_DEFAULT, /*TIME_NOSECONDS*/0, &st, NULL, pBuff, BUFFER_MAX - len - 1);
	sResult.ReleaseBuffer();
	return sResult;
}

int CALLBACK CZBListView::CompareProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	if(me->m_listtype != LIST_DRIVE) {
		CFindData& data1 = m_finder[lParam1];
		CFindData& data2 = m_finder[lParam2];
		if(data1.IsDirectory() != data2.IsDirectory()) {
			return data1.IsDirectory() ? -1 : 1;
		}
		switch(me->m_pnColumnTable[me->m_nSort]) {
		case HEAD_NAME:
			return data1.CompareTo(&data2, SortByName, me->m_bReverse);
		case HEAD_SIZE:
			return data1.CompareTo(&data2, SortBySize, !me->m_bReverse);
		case HEAD_TYPE:
			return data1.CompareTo(&data2, SortByType, me->m_bReverse);
		case HEAD_DATE:
			return data1.CompareTo(&data2, SortByTime, !me->m_bReverse);
		case HEAD_ATTR:
			return data1.CompareTo(&data2, SortByAttr, me->m_bReverse);
		case HEAD_FOLDER:
			return data1.CompareTo(&data2, SortByFolder, me->m_bReverse);
		}
	} else 
		return ComparePIDs(g_pSFRoot, (LPITEMIDLIST)lParam1, (LPITEMIDLIST)lParam2);
	return 0;
}

void CZBListView::DisplayInfo()
{
	CListCtrl& list = GetListCtrl();
	CString sMsg;

	if(list.GetItemCount()) {
		if(IsFileList()) {
			if(m_bTaskBusy) {
				sMsg.Format(IDS_INFO_UNFOLD, m_finder.GetCount(), m_finder.GetSubFolder());
			} else {
				BOOL bSelected = (list.GetSelectedCount() != 0);

				DWORDLONG dwlFileSize = 0;	// ###1.20
				DWORD dwTreeSize = 0;
				for_to(i, list.GetItemCount()) {
					if(!bSelected || list.GetItemState(i, LVIS_SELECTED) & LVIS_SELECTED) {
						int index = list.GetItemData(i);
						if(index != -1) {	// ###1.23
							CFindData& data = m_finder[index];
							if(data.IsDirectory())
								dwTreeSize += data.dwSize;
							else
								dwlFileSize += data.dwlSize;
						}
					}
				}
				DWORD dwFileSize = (DWORD)((dwlFileSize + 1023) / 1024);
				CString sSize = SeparateByComma(dwFileSize + dwTreeSize);

				if(bSelected)
					sMsg.Format(IDS_INFO_SELECT, list.GetSelectedCount(), sSize);
				else
					sMsg.Format(IDS_INFO_TOTAL, list.GetItemCount(), sSize);
			}
		} else
			sMsg.Format(IDS_INFO_OBJECTS, list.GetItemCount());
	} else
		sMsg.LoadString(IDS_INFO_NONE);

	GetMainFrame()->SetStatusBarText(0, sMsg);

	if(!m_bTaskBusy) {
		CString sFree;
		if(IsFileList()) {
			if(!m_hFTP) {
				CString sTotal;
				CString sDrive(m_sPath);
		#ifdef _INC_SHLWAPI
				LPTSTR pDrive = sDrive.GetBuffer(0);
				PathStripToRoot(pDrive);
				sDrive.ReleaseBuffer();
		#else
				sDrive = sDrive.Left(3);
		#endif
				GetDriveInfoString(sDrive, sTotal, sFree);
				if(!sFree.IsEmpty())
					sFree = sDrive.Left(2) + _T(" ") + sFree + _T(" free");
			}
		}
		GetMainFrame()->SetStatusBarText(1, sFree);
	}
}

/////////////////////////////////////////////////////////////////////////////
// CTreeSize : calculate tree size

CFTPFindFiles CTreeSize::m_finder;
CZBListView* CTreeSize::m_pView = NULL;
static DWORD s_dwBytesPerCluster;

void CZBListView::CalcTreeSize()
{
	s_dwBytesPerCluster = GetBytesPerCluster();
	if(s_dwBytesPerCluster == 0) return;
	if(!m_pTreeSize)
		m_pTreeSize = new CTreeSize;
	m_pTreeSize->m_pView = this;

	CString sDrive = m_finder.GetFindDir().Left(3);
	int i = m_pTreeSize->FindNode(sDrive);
	if(i >= 0)
		m_pTreeSize->RemoveAt(i);

	m_bTaskBusy = TRUE;
	AfxGetApp()->OnIdle(0);
	BOOL bResult = m_pTreeSize->AddNode(NULL, sDrive, NULL);
	m_bTaskBusy = FALSE;
	if(bResult)
		FillItems(NULL);
}

DWORD CZBListView::GetBytesPerCluster()
{
	DWORD dwResult = 0;
	CString sDrive = m_finder.GetFindDir().Left(3);
	GetDriveInfo(sDrive, NULL, NULL, &dwResult);
	return dwResult;
}

CTreeNode::CTreeNode()
{
	m_pSubTree = NULL;
	m_dwlDiskSize = m_dwlFileSize = 0;
}

CTreeNode::~CTreeNode()
{
	if(m_pSubTree)
		delete m_pSubTree;
}

CTreeSize::CTreeSize()
{
}

CTreeSize::~CTreeSize()
{
}

BOOL CTreeSize::AddNode(LPCTSTR pRoot, LPCTSTR pDir, CTreeNode* pParent)
{
	CTreeNode nodeNew;
	CTreeNode& node = ElementAt(Add(nodeNew));
	node.m_sName = pDir;
	CFileName sPath;
	if(pRoot) {
		sPath = CString(pRoot);
		sPath.AddSlash();
	}
	sPath += pDir;
	GetMainFrame()->ReportMessage(sPath);
	if(!m_finder.FindAll(sPath, NULL, FIND_DIRFILE)) return TRUE;

	CStringArray arrsDir;
	for_to(i, m_finder.GetCount()) {
		CFindData& data = m_finder[i];
		if(data.IsDirectory())
			arrsDir.Add(data.GetFileName());
		else {
			node.m_dwlFileSize += data.dwlSize;
			node.m_dwlDiskSize += ((data.dwlSize + (s_dwBytesPerCluster - 1)) & ~(s_dwBytesPerCluster-1));
		}
	}
	m_finder.RemoveAll();

	if(!((CZBrowseApp*)AfxGetApp())->MessageLoop()) return FALSE;
	if(!m_pView->IsTaskBusy()) return FALSE;

	for_to_(i, arrsDir.GetSize()) {
		if(!node.m_pSubTree)
			node.m_pSubTree = new CTreeSize;
		if(!node.m_pSubTree->AddNode(sPath, arrsDir[i], &node)) return FALSE;
	}
	if(pParent) {
		pParent->m_dwlFileSize += node.m_dwlFileSize;
		pParent->m_dwlDiskSize += node.m_dwlDiskSize;
	}
	return TRUE;
}

int CTreeSize::FindNode(LPCTSTR pName)
{
	for_to(i, GetSize()) {
		if(!ElementAt(i).m_sName.CompareNoCase(pName)) return i;
	}
	return INVALID;
}

CTreeNode* CZBListView::FindNode(LPCTSTR pPath)
{
	static TCHAR* sSep = _T("\\/");
	if(!m_pTreeSize) return NULL;
	CTreeSize* pTree = m_pTreeSize;
	CTreeNode* pNode = NULL;
	CString sPath(pPath);
	LPTSTR p = sPath.GetBuffer(0);
	CString sDrive;
	p = (LPTSTR)_tcstok(p, sSep);
	if(_tcslen(p) == 2 && p[1] == ':') {
		sDrive = (LPCTSTR)p;
		sDrive += '\\';
		p = (LPTSTR)(LPCTSTR)sDrive;
	}
	while(p && pTree) {
		int i = pTree->FindNode((LPCTSTR)p);
		if(i == INVALID) { pNode = NULL; break; }
		pNode = &pTree->ElementAt(i);
		pTree = pNode->m_pSubTree;
		p = (LPTSTR)_tcstok(NULL, sSep);
	}
	sPath.ReleaseBuffer();
	return pNode;
}

void CZBListView::ReCalcTreeSize()
{
	CFileName sPath = m_sPath;
	do {
		CTreeNode* pTree = FindNode(sPath);
		if(pTree) {
			pTree->m_dwlDiskSize -= m_fileOp.dwlDoneDiskBytes;
			pTree->m_dwlFileSize -= m_fileOp.dwlDoneFileBytes;
		}
	}
	while(sPath.GetParentDirectory());
}

/////////////////////////////////////////////////////////////////////////////
// CZBListView operations

void CZBListView::NewItem(BOOL bFolder)
{
	m_bNewFolder = bFolder;
	if(!IsFileList()) return;
	CListCtrl& list = GetListCtrl();
	LV_ITEM lvitem;
	lvitem.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;
	lvitem.iItem = list.GetItemCount();
	lvitem.iSubItem = 0;
	lvitem.pszText = NULL;
	lvitem.iImage = GetFolderIcon(NULL, bFolder ? FILE_ATTRIBUTE_DIRECTORY : 0, SHGFI_SMALLICON);
	lvitem.lParam = INVALID;
	list.EditLabel(list.InsertItem(&lvitem));
}

void CZBListView::SelectAll(BOOL bInvert)
{
	CListCtrl& list = GetListCtrl();
	for_to(i, list.GetItemCount()) {
		UINT nState = (!bInvert || list.GetItemState(i, LVIS_SELECTED) == 0) ? LVIS_SELECTED : 0;
		list.SetItemState(i, nState, LVIS_SELECTED);
	}
	DisplayInfo();
}

BOOL CZBListView::ClearAll()
{
	CListCtrl& list = GetListCtrl();

	BOOL bClear = FALSE;
	POSITION pos = list.GetFirstSelectedItemPosition();
	while(pos) {
		bClear = TRUE;
		int nItem = list.GetNextSelectedItem(pos);
		list.SetItemState(nItem, 0, LVIS_SELECTED);
	}
	return bClear;
}

CString CZBListView::GetFullPath(int i)
{
	CString sPath = m_finder.GetFullPath(i);
#ifndef NO_FTP
	if (m_hFTP)
		sPath = ftp2.MakeURL(sPath);
#endif
	return sPath;
}

/////////////////////////////////////////////////////////////////////////////
// CZBListView message handlers

int CZBListView::GetFirstSelectedItem()
{
	POSITION pos = GetListCtrl().GetFirstSelectedItemPosition();
	return pos ? GetListCtrl().GetNextSelectedItem(pos) : INVALID;
}

void CZBListView::OnColumnClick(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_LISTVIEW* pNMListView = (NM_LISTVIEW*)pNMHDR;
	// TODO: Add your control notification handler code here
	if(m_nSort != pNMListView->iSubItem || m_bNotSort) {
		m_bNotSort = FALSE;
		m_nSort = pNMListView->iSubItem;
		m_bReverse = FALSE;
	} else
		m_bReverse = !m_bReverse;
	DrawHeaderImage();
	GetListCtrl().SortItems(CompareProc, 0);
	*pResult = 0;
}

void CZBListView::OnFileEnter() 
{
	// TODO: Add your command handler code here
	DblClickItem(GetFirstSelectedItem());
}

void CZBListView::OnDblclk(NMHDR* pNMHDR, LRESULT* pResult) 
{
//	NM_LISTVIEW* pNMListView = (NM_LISTVIEW*)pNMHDR;
	// TODO: Add your control notification handler code here
	POINT pt;
	::GetCursorPos(&pt);
	ScreenToClient(&pt);
	DblClickItem(GetListCtrl().HitTest(pt));
	*pResult = 0;
}

void CZBListView::DblClickItem(int iItem)
{
	if(iItem != INVALID) {
		DWORD lParam = GetListCtrl().GetItemData(iItem);
		if(IsFileList()){
			CFileName sPath = GetFullPath(lParam);
			CString sWorkDir = m_finder.GetFindDir();
			DWORD dwAttr = SFGAO_FILESYSTEM;
			if(!m_hFTP) dwAttr = GetFolderAttr(sPath);
			if(m_finder[lParam].IsDirectory() && dwAttr & SFGAO_FILESYSTEM) {
				FillItems(sPath, m_hFTP != NULL);
				GetMainFrame()->BrowseFrom(sPath);
			} else if(!m_hFTP && !IsArcList() && IsArchiveFile(sPath, TRUE)) {
				FillItems(sPath, m_hFTP != NULL);
				GetMainFrame()->BrowseFrom(sPath);
			} else {
				if(IsArcList()) {
					if(!ExtractArchive(m_finder[lParam], sPath)) return;
				}
				DWORD dwResult = (DWORD)::ShellExecute(AfxGetMainWnd()->GetSafeHwnd(), _T("open") , sPath, NULL, sWorkDir, SW_SHOWNORMAL);
				TRACE(_T("ShellExecute result=%d\n"), dwResult);
				if(dwResult <= 32 && m_finder[lParam].dwSize < 1024*1024) {
					CString sEditorPath;
					::FindExecutable(_T("zcopy.txt"), NULL, sEditorPath.GetBuffer(_MAX_PATH));	// ###1.60
					sEditorPath.ReleaseBuffer();
					if(sEditorPath.IsEmpty()) 
						::ShellExecute(AfxGetMainWnd()->GetSafeHwnd(), _T("edit") , sPath, NULL, NULL, SW_SHOWNORMAL);
					else
						::ShellExecute(AfxGetMainWnd()->GetSafeHwnd(), _T("open") , sEditorPath, sPath, NULL, SW_SHOWNORMAL);
				}
			}
		} else {
			TCHAR buff[MAX_PATH];
			if(::SHGetPathFromIDList((LPITEMIDLIST)lParam, buff)) {
				FillItems(buff, FALSE);
				GetMainFrame()->BrowseFrom(buff);
			}
		}
	}
}

void CZBListView::OnDeleteItem(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_LISTVIEW* pNMListView = (NM_LISTVIEW*)pNMHDR;
	// TODO: Add your control notification handler code here
	if(m_listtype == LIST_DRIVE) {
		if(pNMListView->lParam != 0)
			g_pMalloc->Free((LPVOID)pNMListView->lParam);
	}
	*pResult = 0;
}

void CZBListView::OnEndLabelEdit(NMHDR* pNMHDR, LRESULT* pResult) 
{
	BOOL bResult = FALSE;
	LV_DISPINFO* pDispInfo = (LV_DISPINFO*)pNMHDR;
	// TODO: Add your control notification handler code here
	int i = pDispInfo->item.lParam;
	if(pDispInfo->item.pszText == NULL || pDispInfo->item.pszText[0] == TCHAR('\0')) {
		if(i == INVALID)  // FolderNew
			GetListCtrl().DeleteItem(pDispInfo->item.iItem);
	} else {
		CFTPFile file;
		file.m_hFTP = m_hFTP;
		CString sPathNew = file.AddPath(m_finder.GetFindDir(),(LPCTSTR)pDispInfo->item.pszText);
		if(i == INVALID) {
			if(m_bNewFolder) {
				bResult = file.CreateDirectory(sPathNew);
				if(bResult) {
					if(m_hFTP) {
						PostMessage(WM_REFRESH_LIST);
						// GetTreeView()->InsertFTPItem((LPCTSTR)pDispInfo->item.pszText);	// ###1.23
					} else
						::SHChangeNotify(SHCNE_MKDIR, SHCNF_PATH, (LPCTSTR)sPathNew, NULL);	// ###1.20
				}
			} else {
				bResult = file.Open(sPathNew, OF_WRITE|OF_CREATE);
				if(bResult) file.Close();
			}
		} else {
			CString sPathOld = m_finder.GetFullPath(i);
			bResult = file.Rename(sPathOld, sPathNew);
			if(bResult) {
				if(m_hFTP)
					PostMessage(WM_REFRESH_LIST);	// ###1.23
				else
					::SHChangeNotify(SHCNE_RENAMEFOLDER, SHCNF_PATH, (LPCTSTR)sPathOld,  (LPCTSTR)sPathNew); // ###1.20
			}
		}
		if(bResult) {
			m_sPrevSelect = (LPCTSTR)pDispInfo->item.pszText;
			IsChangeNotify();
			/* if(i != INVALID && m_finder[i].IsDirectory()) { // ###1.03	// ###1.20
				SetRefresh();
				GetMainFrame()->PostMessage(WM_COMMAND, ID_VIEW_REFRESH);
			} */
		} else {
			AfxMessageBox(GetErrorMessage(), MB_ICONHAND);
			if(i == INVALID)
				GetListCtrl().PostMessage(LVM_EDITLABEL, pDispInfo->item.iItem);
		}
	}
	*pResult = bResult;
}

void CZBListView::OnFileRename() 
{
	// TODO: Add your command handler code here
	int iItem = GetFirstSelectedItem();
	if(iItem >= 0)
		GetListCtrl().EditLabel(iItem);
}

void CZBListView::OnUpdateFileRename(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(GetListCtrl().GetSelectedCount() == 1);	
}

BOOL CZBListView::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message) 
{
	// TODO: Add your message handler code here and/or call default
	if(nHitTest == HTCLIENT && pWnd == this && m_bBusy) {
		SetCursor(m_hCsrBusy);
		return TRUE;
	}
	return CListView::OnSetCursor(pWnd, nHitTest, message);
}

BOOL CZBListView::PreTranslateMessage(MSG* pMsg) 
{
	// TODO: Add your specialized code here and/or call the base class
	if(pMsg->message == WM_PAINT) {
		if(m_bBusy)
			return TRUE;
	}
	return CListView::PreTranslateMessage(pMsg);
}

void CZBListView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// TODO: Add your message handler code here and/or call default
	switch(nChar) {
	case VK_TAB:
		if(::GetKeyState(VK_SHIFT) < 0)
			GetMainFrame()->SetActiveView(GetTreeView());
		else
			GetMainFrame()->m_wndToolBar.SetFocus();
		break;
	}
	CListView::OnKeyDown(nChar, nRepCnt, nFlags);
	DisplayInfo();	// ###1.22
}

int CZBListView::OnMouseActivate(CWnd* pDesktopWnd, UINT nHitTest, UINT message)
{
	int nResult = CWnd::OnMouseActivate(pDesktopWnd, nHitTest, message);

	if (nResult == MA_NOACTIVATE || nResult == MA_NOACTIVATEANDEAT
		|| nHitTest == HTHSCROLL ||  nHitTest == HTVSCROLL)
		return nResult;   // frame does not want to activate

	CFrameWnd* pParentFrame = GetParentFrame();
	if (pParentFrame != NULL)
	{
		// eat it if this will cause activation
		ASSERT(pParentFrame == pDesktopWnd || pDesktopWnd->IsChild(pParentFrame));

		// either re-activate the current view, or set this view to be active
		CView* pView = pParentFrame->GetActiveView();
		HWND hWndFocus = ::GetFocus();
		if (pView == this &&
			m_hWnd != hWndFocus && !::IsChild(m_hWnd, hWndFocus))
		{
			// re-activate this view
			OnActivateView(TRUE, this, this);
		}
		else
		{
			// activate this view
			pParentFrame->SetActiveView(this, FALSE);	// bNotify = FALSE!!
		}
	}
	return nResult;
}

void CZBListView::OnUpdateViewRealName(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(IsFileList() && !m_hFTP);
	pCmdUI->SetCheck(options.m_viewStyle & STYLE_REALNAME);
}

void CZBListView::IsChangeNotify()
{
	// TODO: Add your message handler code here and/or call default
	if(m_hChangeNotify) {
		BOOL bChanged = FALSE;
		while(::WaitForSingleObject(m_hChangeNotify, 0) == WAIT_OBJECT_0) {
			::FindNextChangeNotification(m_hChangeNotify);
			bChanged = TRUE;
		}
		if(bChanged && !m_bRefresh) {
			TRACE(_T("*** List Changed\n"));
			int nCount = GetListCtrl().GetSelectedCount();		// ###1.20
			if(nCount <= 1) {
				if(nCount == 1) {
					int iItem = GetFirstSelectedItem();
					if(iItem >= 0)
						m_sPrevSelect = GetListCtrl().GetItemText(iItem, 0);
				}
				FillItems(NULL);
			}
		}
	}
}

/* ###1.22 

void CZBListView::OnItemChanged(NMHDR* pNMHDR, LRESULT* pResult)  // ###1.20
{
	NM_LISTVIEW* pNMListView = (NM_LISTVIEW*)pNMHDR;
	// TODO: Add your control notification handler code here
	if(pNMListView && pNMListView->uNewState & LVIS_SELECTED)
		DisplayInfo();
	*pResult = 0;
}
*/

/////////////////////////////////////////////////////////////////////////////
// CZBListView OLE Drag and Drop

extern CImageList* g_pDragImage;

void CZBListView::OnBeginDrag(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_LISTVIEW* pNMListView = (NM_LISTVIEW*)pNMHDR;
	*pResult = 0;
	// TODO: Add your control notification handler code here
	static UINT cfShellIDList = 0;
	if(!IsFileList()) return;
	if(g_pDragImage) delete g_pDragImage;
	CPoint pt;
	g_pDragImage = GetListCtrl().CreateDragImage(pNMListView->iItem, &pt);
	g_pDragImage->BeginDrag(0, CPoint(8, 8));
	pt = pNMListView->ptAction;
	ClientToScreen(&pt);
	g_pDragImage->DragEnter(NULL, pt);
	COleDataSource dataSrc;
	HGLOBAL hDrop = CreateHDrop();
	dataSrc.CacheGlobalData(CF_HDROP, hDrop);
	HGLOBAL hShellIDL = CreateShellIDList();
	if(hShellIDL) {
		if(cfShellIDList == 0)
			cfShellIDList = ::RegisterClipboardFormat(CFSTR_SHELLIDLIST);
		dataSrc.CacheGlobalData(cfShellIDList, hShellIDL);
	}
	m_pDragFinder = &m_finder;
	dataSrc.DoDragDrop(DROPEFFECT_COPY | DROPEFFECT_MOVE | DROPEFFECT_LINK);
	g_pDragImage->DragLeave(NULL);
	g_pDragImage->EndDrag();
	delete g_pDragImage;
	g_pDragImage = NULL;
	if(m_iDropItem != INVALID)
		GetListCtrl().SetItemState(m_iDropItem, 0, LVIS_DROPHILITED);
	m_pDragFinder = NULL;
}

DROPEFFECT CZBListView::OnDragEnter(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point) 
{
	// TODO: Add your specialized code here and/or call the base class
	if(!pDataObject->IsDataAvailable(CF_HDROP) || !IsFileList()) return DROPEFFECT_NONE;
	HDROP hDrop = (HDROP)pDataObject->GetGlobalData(CF_HDROP);
	if(!hDrop) return DROPEFFECT_NONE;
	m_sDropSrcDir = HDropGetFileName(hDrop);
	m_sDropSrcDir.GetParentDirectory();
	::GlobalFree(hDrop);
	m_iDropItem = INVALID;
	return OnDragOver(pDataObject, dwKeyState, point);
}

DROPEFFECT CZBListView::OnDragOver(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point) 
{
	// TODO: Add your specialized code here and/or call the base class
	if(!pDataObject->IsDataAvailable(CF_HDROP) || !IsFileList()) return DROPEFFECT_NONE;
	if(g_pDragImage) {
		CPoint pt(point);
		ClientToScreen(&pt);
		g_pDragImage->DragMove(pt);
	}
	CListCtrl& list = GetListCtrl();
	int iItem = list.HitTest(point);
	int i = INVALID;
	if(iItem != INVALID) {
		i = list.GetItemData(iItem);
	}
	if(iItem == INVALID || !m_finder[i].IsDirectory()) {
		m_sDropDir = m_finder.GetFindDir();
#ifndef NO_FTP
		if (m_hFTP)
			m_sDropDir = ftp2.MakeURL(m_sDropDir);
#endif
		iItem = INVALID;
	} else
		m_sDropDir = GetFullPath(i);

	if(iItem != m_iDropItem) {
		if(g_pDragImage)
			g_pDragImage->DragShowNolock(FALSE);
		if(m_iDropItem != INVALID) {
			list.SetItemState(m_iDropItem, 0, LVIS_DROPHILITED);
			list.RedrawItems(m_iDropItem, m_iDropItem);
		}
		m_iDropItem = iItem;
		if(m_iDropItem != INVALID) {
			list.SetItemState(m_iDropItem, LVIS_DROPHILITED, LVIS_DROPHILITED);
			list.RedrawItems(m_iDropItem, m_iDropItem);
		}
		list.UpdateWindow();
		if(g_pDragImage)
			g_pDragImage->DragShowNolock(TRUE);
	}
	DROPEFFECT de = DROPEFFECT_NONE;
	if(m_sDropDir.ComparePath(m_sDropSrcDir) == 0) {
		if(!(dwKeyState & MK_SHIFT))
			de = DROPEFFECT_COPY;
	} else {
		de = (m_sDropDir.GetDrive() == m_sDropSrcDir.GetDrive()) ? DROPEFFECT_MOVE : DROPEFFECT_COPY;
		if(dwKeyState & MK_CONTROL) de = DROPEFFECT_COPY;
		if(dwKeyState & MK_SHIFT)   de = DROPEFFECT_MOVE;
	}
	return de;
}

void CZBListView::OnDragLeave() 
{
	// TODO: Add your specialized code here and/or call the base class
	if(g_pDragImage)
		g_pDragImage->DragShowNolock(FALSE);
	if(m_iDropItem != INVALID)
		GetListCtrl().SetItemState(m_iDropItem, 0, LVIS_DROPHILITED);
	CListView::OnDragLeave();
}

// static FileOpParam m_fileOp;	###1.20 ???

BOOL CZBListView::OnDrop(COleDataObject* pDataObject, DROPEFFECT dropEffect, CPoint point) 
{
	// TODO: Add your specialized code here and/or call the base class
	m_fileOp.hDrop = (HDROP)pDataObject->GetGlobalData(CF_HDROP);
	if(!m_fileOp.hDrop) return FALSE;
	m_fileOp.wFunc = (dropEffect == DROPEFFECT_MOVE) ? FO_MOVE : FO_COPY;
	m_fileOp.sDstDir = m_sDropDir;
	m_fileOp.hwnd = GetSafeHwnd();
	m_fileOp.pFinder = m_pDragFinder;
	m_bBusy = TRUE;
	m_fileOp.Start();
	return TRUE;
}

void CZBListView::OnFileRemove() 
{
	// TODO: Add your command handler code here
	CListCtrl& list = GetListCtrl();
	CString sMsg;
	sMsg.Format(IDS_REMOVE_LIST, list.GetSelectedCount());
	if(AfxGetMainWnd()->MessageBox(sMsg, _T("Delete"), MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2) != IDYES)
		return;
	m_fileOp.hDrop = (HDROP)CreateHDrop();
	m_fileOp.wFunc = FO_DELETE;
	m_fileOp.sDstDir.Empty();
	m_fileOp.hwnd = GetSafeHwnd();
	m_fileOp.pFinder = &m_finder;
	m_bBusy = TRUE;

	m_fileOp.Start(m_bTreeSize ? GetBytesPerCluster() : 0);
}

void CZBListView::OnUpdateFileRemove(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(IsFileList() && GetListCtrl().GetSelectedCount() && !m_bBusy);
}

LRESULT CZBListView::OnFileOpComplete(WPARAM bDone, LPARAM)
{
	m_bBusy = FALSE;
	PostMessage(WM_SETCURSOR);
	m_fileOp.pDlg->DestroyWindow();
	if(bDone) {
		if(m_fileOp.wFunc == FO_COMPARE) {
			if(!m_fileOp.pDlg->m_bCanceled)
				ReSelectDuplicateFiles();
		} else {
			if(m_fileOp.wFunc == FO_DELETE && m_bTreeSize)
				ReCalcTreeSize();
			m_iPrevTopIndex = GetListCtrl().GetTopIndex();	// ###1.20
			IsChangeNotify();
	//		FillItems(NULL, m_hFTP != NULL);
			SetRefresh();
			GetMainFrame()->PostMessage(WM_COMMAND, ID_VIEW_REFRESH);
		}
	}
	if(m_fileOp.wFunc == FO_COMPARE) {
		m_finder.ClearMarkAll();
		m_fileOp.findPair.RemoveAll();
	}	
	return 1;
}

HGLOBAL CZBListView::CreateHDrop() // ##120721 MultiByte --> WideChar
{
	CListCtrl& list = GetListCtrl();
	int len = 0;
	for_to(i, list.GetItemCount()) {
		int iFind = list.GetItemData(i);
		if(list.GetItemState(i, LVIS_SELECTED) & LVIS_SELECTED) {
			CStringW sPath = CStringW(GetFullPath(iFind));
			len += sPath.GetLength() + 1;
			m_finder[iFind].SetMark();
		} else
			m_finder[iFind].ClearMark();
	}
	HGLOBAL hGlobal = ::GlobalAlloc(GHND | GMEM_SHARE, (DWORD)(sizeof(DROPFILES) + len * 2 + 2));
	LPDROPFILES pDropFiles = (LPDROPFILES)::GlobalLock(hGlobal);
   	pDropFiles->pFiles = sizeof(DROPFILES);
   	pDropFiles->fWide = TRUE; // FALSE;
	LPWSTR p = (LPWSTR)(pDropFiles + 1);
	for_to_(i, list.GetItemCount()) {
		if(list.GetItemState(i, LVIS_SELECTED) & LVIS_SELECTED) {
			CStringW sPath = CStringW(GetFullPath(list.GetItemData(i)));
			wcscpy(p, sPath);
			p += sPath.GetLength() + 1;
		}		
	}
	*p = 0;
	::GlobalUnlock(hGlobal);
 	return hGlobal;
}

HGLOBAL CZBListView::CreateShellIDList()
{
	if(m_hFTP) return NULL;

	CListCtrl& list = GetListCtrl();
	int nCount = list.GetSelectedCount();
	UINT nOfs = sizeof(CIDA) + sizeof(UINT) * nCount;
	LPITEMIDLIST pIDLParent = GetPIDLOf(g_pSFRoot, m_sPath);
	if(!pIDLParent) return NULL;
	UINT dwPIDL = GetPIDLSize(pIDLParent);
	DWORD dwTotal = nOfs + dwPIDL * (nCount + 1);
	HGLOBAL hGlobal = ::GlobalAlloc(GHND | GMEM_SHARE, dwTotal);
	LPIDA pIDA = (LPIDA)GlobalLock(hGlobal);
   	pIDA->cidl = nCount;
	pIDA->aoffset[0] = nOfs;
	memcpy((LPBYTE)(pIDA) + nOfs, (LPBYTE)pIDLParent, dwPIDL);
	nOfs += dwPIDL;
	LPSHELLFOLDER pSFParent = NULL;
	if(g_pSFRoot->BindToObject(pIDLParent, NULL, IID_IShellFolder, (LPVOID *)&pSFParent) != NOERROR) {
		g_pMalloc->Free((LPVOID)pIDLParent);
		::GlobalUnlock(hGlobal);
		::GlobalFree(hGlobal);
		return NULL;
	}
	int iItem = 0;
	for_to(i, list.GetItemCount()) {
		if(list.GetItemState(i, LVIS_SELECTED) & LVIS_SELECTED) {
			LPITEMIDLIST pIDL = GetPIDLOf(pSFParent, m_finder[list.GetItemData(i)].GetFileName());
			ASSERT(pIDL);
			if(pIDL) {
				pIDA->aoffset[iItem + 1] = nOfs;
				dwPIDL = GetPIDLSize(pIDL);
				if(dwTotal < nOfs + dwPIDL) {
					::GlobalUnlock(hGlobal);
					dwTotal += (nCount - iItem) * dwPIDL;
		 			hGlobal = ::GlobalReAlloc(hGlobal, dwTotal, GHND | GMEM_SHARE);
					pIDA = (LPIDA)GlobalLock(hGlobal);
				}
				memcpy((LPBYTE)(pIDA) + nOfs, (LPBYTE)pIDL, dwPIDL);
				nOfs += dwPIDL;
				iItem++;
			}
		}
	}
	pSFParent->Release();
	GlobalUnlock(hGlobal);
 	return hGlobal;
}

void CZBListView::OnClick(NMHDR* pNMHDR, LRESULT* pResult) 
{
	// TODO: Add your control notification handler code here
	DisplayInfo();
	*pResult = 0;
}

void CZBListView::OnRClick(NMHDR* pNMHDR, LRESULT* pResult) 
{
	// TODO: Add your control notification handler code here
//	NM_LISTVIEW* pNMListView = (NM_LISTVIEW*)pNMHDR;
	*pResult = 0;
	if(!IsFileList() || GetListCtrl().GetSelectedCount() == 0) return;
	POINT pt;
	::GetCursorPos(&pt);
	DoContextMenu(pt);
}

void CZBListView::OnFileContextMenu() 
{
	// TODO: Add your command handler code here
	CListCtrl& list = GetListCtrl();
	if(!IsFileList() || list.GetSelectedCount() == 0) return;
	int iLast = INVALID;
	for_to(i, list.GetItemCount()) {
		if(list.GetItemState(i, LVIS_SELECTED) & LVIS_SELECTED) iLast = i;
	}
	if(iLast != INVALID) {
		CRect rItem;
		list.GetItemRect(iLast, &rItem, LVIR_ICON);
		POINT pt = rItem.CenterPoint();
		ClientToScreen(&pt);
		DoContextMenu(pt);
	}
}

void CZBListView::DoContextMenu(POINT pt)
{
	CListCtrl& list = GetListCtrl();
	USES_CONVERSION;
	LPITEMIDLIST pIDLParent;
	if(pIDLParent = GetPIDLOf(g_pSFRoot, m_sPath)) {
		LPSHELLFOLDER pSFParent = NULL;
		if(g_pSFRoot->BindToObject(pIDLParent, NULL, IID_IShellFolder, (LPVOID *)&pSFParent) == NOERROR) {
			CArray<LPCITEMIDLIST, LPCITEMIDLIST> apIDL;
			for_to(i, list.GetItemCount()) {
				if(list.GetItemState(i, LVIS_SELECTED) & LVIS_SELECTED) {
					LPITEMIDLIST pIDL = NULL;
					if(pIDL = GetPIDLOf(pSFParent, m_finder[list.GetItemData(i)].GetFileName()))
						apIDL.Add(pIDL);
				}
			}
			if(apIDL.GetSize()) {
				GetMainFrame()->PopupContextMenu(pSFParent, apIDL.GetSize(), &apIDL[0], pt);
				for_to_(i, apIDL.GetSize())
					g_pMalloc->Free((LPVOID)apIDL[i]);
			}
			pSFParent->Release();
		}
		g_pMalloc->Free(pIDLParent);
	}
}

CString CZBListView::GetSelectedFileTitle()
{
	CString sResult;
	int iItem = GetFirstSelectedItem();
	if(iItem != -1) {
		CFindData& data = m_finder[GetListCtrl().GetItemData(iItem)];
		sResult += SeparateByComma(data.dwSize);
		sResult += " ";
		sResult += data.GetFileName();
	}
	return sResult;
}

/////////////////////////////////////////////////////////////////////////////
// Select Duplicate Files ###1.20

void CZBListView::SelectDuplicateFiles()
{
	CListCtrl& list = GetListCtrl();

	if(list.GetItemCount() == 0) return;
	ClearAll();
	m_finder.Sort(SortBySize, FALSE);
	ScanDuplicateFiles(FALSE);
	int nCount = list.GetSelectedCount();
	if(nCount == 0)
		AfxMessageBox(IDS_NODUPFILES);
	else {
		DisplayInfo();
		GetMainFrame()->SetActiveView(this);

		CString sMsg;
		sMsg.Format(IDS_DUPFILES_QUERY, nCount);
		if(AfxMessageBox(sMsg, MB_YESNO) == IDYES) {
			m_finder.ClearMarkAll();
			m_fileOp.findPair.RemoveAll();

			ScanDuplicateFiles(TRUE);

			m_fileOp.hDrop = NULL;
			m_fileOp.wFunc = FO_COMPARE;
			m_fileOp.sSrcDir = m_fileOp.sDstDir = m_sPath;
			m_fileOp.hwnd = GetSafeHwnd();
			m_fileOp.pFinder = NULL;
			m_bBusy = TRUE;
			m_fileOp.Start();
		}
	}
}

void CZBListView::ScanDuplicateFiles(BOOL bCompare)
{
	CListCtrl& list = GetListCtrl();

	CFindData* pData = m_finder.GetHead();
	CFindData* pDataSrc = NULL;
	CFindData* pDataDst = NULL;
	while(pData) {
		if(!pData->IsDirectory() && pData->dwlSize) {
			if(!pDataSrc) pDataSrc = pData;
			else {
				pDataDst = pData;
				if(pData->dwlSize == pDataSrc->dwlSize) {
					int nTimeDif = ::CompareFileTime(&pData->ft.ftLastWrite, &pDataSrc->ft.ftLastWrite);
					int nLenDif = _tcslen(pData->GetFileName()) - _tcslen(pDataSrc->GetFileName());
					BOOL bSwap = (
						   nTimeDif < 0
						|| nTimeDif == 0 && nLenDif < 0
						|| nTimeDif == 0 && nLenDif == 0 && _tcsicmp(pData->GetFileName(), pDataSrc->GetFileName()) < 0
					);
					if(bSwap)
						Swap(pDataSrc, pDataDst);
					if(!bCompare) {
						int nItem = FindItem(pDataDst);
						if(nItem >= 0) 
							list.SetItemState(nItem, LVIS_SELECTED, LVIS_SELECTED);
					} else {
						FindDataPair findPair;
						findPair.pDataSrc = pDataSrc;
						findPair.pDataDst = pDataDst;
						m_fileOp.findPair.Add(findPair);
					}
				} else 
					pDataSrc = pData;
			}
		}
		pData = pData->GetNext();
	}
}

int CZBListView::FindItem(CFindData* pData)
{
	int nIndex = pData - m_finder.GetData();
	LVFINDINFO findInfo;
	ZeroMem(findInfo);
	findInfo.flags = LVFI_PARAM;
	findInfo.lParam = nIndex;
	return GetListCtrl().FindItem(&findInfo);
}

void CZBListView::ReSelectDuplicateFiles()
{
	CListCtrl& list = GetListCtrl();

	for_to(i, m_finder.GetCount()) {
		CFindData* pData = &m_finder[i];
		if(pData->IsMarked()) {
			int nItem = FindItem(pData);
			if(nItem >= 0) 
				list.SetItemState(nItem, 0, LVIS_SELECTED);
		}
	}
	int nCount = list.GetSelectedCount();
	if(nCount) {
		CString sMsg;
		sMsg.Format(IDS_DUPFILES_DONE, nCount);
		AfxMessageBox(sMsg, MB_ICONINFORMATION);
	} else
		AfxMessageBox(IDS_DUPFILES_NONE, MB_ICONINFORMATION);
	DisplayInfo();
}

/////////////////////////////////////////////////////////////////////////////
// Thumbnail view ###1.24

CThumbImage::CThumbImage()
{
}

CThumbImage::~CThumbImage()
{
	m_bmpIcon.DeleteObject();
}

void CThumbImage::Init(CView* pView)
{
	m_pView = pView;
	CClientDC dc(pView);
	CDC dcMem;
	dcMem.CreateCompatibleDC(&dc);
	m_bmpIcon.CreateCompatibleBitmap(&dc, THUMB_CX, THUMB_CY);
	m_cIcon = CSize(::GetSystemMetrics(SM_CXICON), ::GetSystemMetrics(SM_CYICON));
}

HBITMAP CThumbImage::IconToBitmap(HICON hIcon)
{
	CClientDC dc(m_pView);
	CDC dcMem;
	dcMem.CreateCompatibleDC(&dc);
	CBitmap* pbmpOld = (CBitmap*)dcMem.SelectObject(&m_bmpIcon);
	dcMem.FillSolidRect(0, 0, THUMB_CX, THUMB_CY, ::GetSysColor(COLOR_WINDOW));
	::DrawIconEx(dcMem.GetSafeHdc(), (THUMB_CX - THUMB_ICON_SIZE) / 2, (THUMB_CY - THUMB_ICON_SIZE) / 2
		, hIcon, THUMB_ICON_SIZE, THUMB_ICON_SIZE, 0, NULL, DI_NORMAL);
	dcMem.SelectObject(pbmpOld);
	return (HBITMAP)m_bmpIcon.m_hObject;
}

/////////////////////////////////////////////////////////////////////////////
// Thumbnail by ximage 2011/2/1

#ifdef XIMAGE

int CThumbImage::GetXImageFormat(LPCTSTR pExt)
{
	if (!_tcsicmp(pExt, _T("jpg")) || !_tcsicmp(pExt, _T("jpeg"))) return CXIMAGE_FORMAT_JPG;
	if (!_tcsicmp(pExt, _T("png")))  return CXIMAGE_FORMAT_PNG;
	if (!_tcsicmp(pExt, _T("gif")))  return CXIMAGE_FORMAT_GIF;
	if (!_tcsicmp(pExt, _T("bmp")))  return CXIMAGE_FORMAT_BMP;
	if (!_tcsicmp(pExt, _T("ico")))  return CXIMAGE_FORMAT_ICO;
	return CXIMAGE_FORMAT_UNKNOWN;
}

HBITMAP CThumbImage::createImage(LPCTSTR pPath)
{
	LPCTSTR pExt = _tcsrchr(pPath, '.');
	if (!pExt) return NULL;
	int nImageType = GetXImageFormat(++pExt);
	if (nImageType == CXIMAGE_FORMAT_UNKNOWN) return NULL;

	CxImage image(CW2A(pPath), nImageType);
	if(!image.IsValid()) return NULL;

	// Calculate Rect to fit to canvas
	int x = 0;
	int y = 0;
	int cx = THUMB_CX;
	int cy = THUMB_CY;
	int width = image.GetWidth();
	int height = image.GetHeight();

	float fAspect = (float)THUMB_CY / THUMB_CX;
	float fImageAspect = (float)height /width;

	if (width <= THUMB_CX && height <= THUMB_CY) {
		x = (THUMB_CX - width) / 2;
		y = THUMB_CY - height;
		cx = width;
		cy = height;
	} else if(fImageAspect > fAspect) {
		cx = (int)(THUMB_CY / fImageAspect);
		x = (THUMB_CX - cx) / 2;
	} else {
		cy = (int)(THUMB_CX * fImageAspect);
		y = (cy <= (THUMB_CY / 2)) ? THUMB_CY - cy : (THUMB_CY - cy) / 2;
	}

	CClientDC dc(m_pView);
	CDC dcMem;
	dcMem.CreateCompatibleDC(&dc);
	CBitmap* pbmpOld = (CBitmap*)dcMem.SelectObject(&m_bmpIcon);
	dcMem.FillSolidRect(0, 0, THUMB_CX, THUMB_CY, ::GetSysColor(COLOR_WINDOW));
	// Draw Image
	image.Stretch(dcMem.GetSafeHdc(), x, y, cx, cy);
	dcMem.SelectObject(pbmpOld);
	return (HBITMAP)m_bmpIcon.m_hObject;
}
#else

HBITMAP CThumbImage::ResizeToFitBitmap(HBITMAP hBmp)
{
	BITMAP bm;
	GetObject(hBmp, sizeof(BITMAP), &bm);

	// Calculate Rect to fit to canvas
	int width = bm.bmWidth;
	int height = bm.bmHeight;
	int x = (THUMB_CX - width) / 2;
	int y = (THUMB_CY - height) / 2;
	int cx = width;
	int cy = height;
	//TRACE(_T("thumb(%dx%d)\n"), width, height);

	CClientDC dc(m_pView);
	CDC dcMem;
	dcMem.CreateCompatibleDC(&dc);
	CBitmap* pbmpOld = (CBitmap*)dcMem.SelectObject(&m_bmpIcon);
	dcMem.FillSolidRect(0, 0, THUMB_CX, THUMB_CY, ::GetSysColor(COLOR_WINDOW));

	CDC dcBmp;
	dcBmp.CreateCompatibleDC(&dc);
	HBITMAP hBmpOld = (HBITMAP)::SelectObject(dcBmp.GetSafeHdc(), hBmp);
	dcMem.BitBlt(x, y, cx, cy, &dcBmp, 0, 0, SRCCOPY);
	(HBITMAP)::SelectObject(dcBmp.GetSafeHdc(), hBmpOld);
	::DeleteObject(hBmp);

	dcMem.SelectObject(pbmpOld);
	return (HBITMAP)m_bmpIcon.m_hObject;

}

#endif

static LPSHELLFOLDER s_pSFParent = NULL;

int CZBListView::ExtractThumbImage(LPCTSTR pPath, CThumbImage* pThumbImage)
{
	int iImage = -1;

	LPCITEMIDLIST pIDL = GetPIDLOf(g_pSFRoot, pPath);
	if(pIDL) {
		LPCITEMIDLIST pIDLLast = GetLastPIDL(pIDL);
		LPEXTRACTIMAGE pIExtract = NULL;
		HRESULT hr = s_pSFParent->GetUIObjectOf(NULL, 1, &pIDLLast, IID_IExtractImage, NULL, (void**)&pIExtract);
		if(hr == NOERROR) {
		   OLECHAR wszPathBuffer[MAX_PATH];
		   DWORD dwPriority = 0; // IEI_PRIORITY_NORMAL is defined nowhere!
		   DWORD dwFlags = IEIFLAG_SCREEN;
		   HBITMAP hBmp = NULL;
		   CSize cBmp(THUMB_CX, THUMB_CY);
		   if(pIExtract->GetLocation(wszPathBuffer, MAX_PATH, &dwPriority, &cBmp, THUMB_BPP, &dwFlags) == NOERROR) {
			   if(pIExtract->Extract(&hBmp) == NOERROR) {
				   iImage = ImageList_Add(m_imageThumbs.GetSafeHandle(), hBmp, NULL);
				   if (iImage == -1) { // ImageList_Add failed when small or portrait aspect image
					   hBmp = pThumbImage->ResizeToFitBitmap(hBmp);
					   iImage = ImageList_Add(m_imageThumbs.GetSafeHandle(), hBmp, NULL);
				   } else {
					   ::DeleteObject(hBmp);
				   }
				   //TRACE(_T("iImage=%d: %s\n"), iImage, pPath);
			   }
			}
		   pIExtract->Release();
		}
#ifdef THUMBNAIL_PROVIDER
		else {
			IThumbnailProvider *pThumbnailProvider = NULL;
			HRESULT hr = s_pSFParent->GetUIObjectOf(NULL, 1, &pIDLLast, IID_IThumbnailProvider, NULL, (void**)&pThumbnailProvider);
			if(hr == NOERROR) {
				pThumbnailProvider->Release();
			}
		}
#endif
		g_pMalloc->Free((LPVOID)pIDL);
	}
	return iImage;
}

UINT CZBListView::DrawIconsThread(LPVOID pParam)
{
	CZBListView* pView = (CZBListView*)pParam;
	CListCtrl& list = pView->GetListCtrl();
	SHFILEINFO sfi;
	pView->m_bVScroll = FALSE;

	::CoInitialize(NULL);

	CThumbImage thumb;

	int iItem = 0;
	int nDoneCount = 0;
	int nItemCount = list.GetItemCount();

	if(pView->IsThumbnail()) {
		thumb.Init(pView);
		if(m_finder.IsArchive() || !pView->IsUnfold()) {
			if(!(s_pSFParent = BindToPath(g_pSFRoot, m_finder.IsArchive() ? pView->GetTempPath() : pView->m_sPath))) goto EndThread;
		}
	}

	while(pView->m_pDrawThread) {
		if(pView->m_bVScroll) {
			CRect rItem, rClient;
			list.GetItemRect(iItem, rItem, LVIR_BOUNDS);
			pView->GetClientRect(rClient);
			if(!rClient.IntersectRect(rClient, rItem)) {
				CPoint ptOrg;
				list.GetOrigin(&ptOrg);
				ptOrg.y -= rItem.Height();
				for_to(i, nItemCount) {
					CPoint ptItem;
					list.GetItemPosition(i, &ptItem);
					if(ptItem.y >= ptOrg.y) {
						iItem = i;
						break;
					}
				}
				pView->m_bVScroll = FALSE;
			}
		}
		int i = list.GetItemData(iItem);
		if(i != INVALID) {
			ShellInfo& shi = s_arrShellInfo[i];
			if(shi.iIconAlt == INVALID) {
				CFileName sFullPath = m_finder.GetFullPath(i);
				if(!pView->IsThumbnail()) {
					::SHGetFileInfo(sFullPath, 0, &sfi, sizeof(SHFILEINFO), SHGFI_DISPLAYNAME | SHGFI_SYSICONINDEX);
					shi.iIconAlt = sfi.iIcon;
				} else {
					CFindData& data = m_finder[i];
					if(!data.IsDirectory()) {
						if(m_finder.IsArchive()) {
							if(pView->ExtractArchive(data, sFullPath)) {
								shi.iIconAlt = pView->ExtractThumbImage(sFullPath, &thumb);
								pView->DeleteTempFiles();
							}
						} else {
							if(pView->IsUnfold()) {
								CFileName sPathParent = sFullPath;
								sPathParent.GetParentDirectory();
								s_pSFParent = BindToPath(g_pSFRoot, sPathParent);
							}
							if(s_pSFParent)
								shi.iIconAlt = pView->ExtractThumbImage(sFullPath, &thumb);
						}
					}
					if(shi.iIconAlt == INVALID) {
#ifdef XIMAGE
						HBITMAP hBmp = thumb.createImage(sFullPath);
						if (hBmp) {
							shi.iIconAlt = ImageList_Add(pView->m_imageThumbs.GetSafeHandle(), hBmp, NULL);
						} else {
#endif
							DWORD dwAttr = (m_hFTP || m_finder.IsArchive()) ? SHGFI_USEFILEATTRIBUTES : SHGFI_DISPLAYNAME;
							::SHGetFileInfo(sFullPath, data.dwAttr, &sfi, sizeof(SHFILEINFO), dwAttr | SHGFI_ICON);
							shi.iIconAlt = ImageList_Add(pView->m_imageThumbs.GetSafeHandle(), thumb.IconToBitmap(sfi.hIcon), NULL);
							::DestroyIcon(sfi.hIcon);
						}
#ifdef XIMAGE
					}
#endif
				}
				if(shi.iIconAlt != shi.iIcon)
					list.RedrawItems(iItem, iItem);
				nDoneCount++;
			}
		}
		if(++iItem >= nItemCount) {
			if(nDoneCount >= nItemCount) break;
			else iItem = 0;
		}
	}
EndThread:
	::CoUninitialize();
	TRACE(_T("DrawIconsThread Done.\n"));
//	pView->m_hDrawThread = NULL;
	return 1;
}

void CZBListView::TerminateThread(CWinThread** ppThread)
{
	CWinThread* pThread = *ppThread;
	if(pThread == NULL) return;
	*ppThread = NULL;
	DWORD dwExitCode = 0;
	// int count = 0;
	while(::GetExitCodeThread(pThread->m_hThread, &dwExitCode) && dwExitCode == STILL_ACTIVE) {
		// TRACE(_T"MessageLoop %d..."), count++);
		if(!((CZBrowseApp*)AfxGetApp())->MessageLoop()) return;
		// TRACE(_T"Done\n"));
	}
	delete pThread;
//	TRACE(_T("::WaitForSingleObject..."));
//	DWORD dwResult = ::WaitForSingleObject(hThread, 1000);
//	TRACE(_T("result=%d\n"), dwResult);
}

void CZBListView::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	// TODO: Add your message handler code here and/or call default
	m_bVScroll = TRUE;
	CListView::OnVScroll(nSBCode, nPos, pScrollBar);
}

void CZBListView::SetThumbPath()
{
	for_to(i, options.m_sThumbPath.GetSize()) {
		if(_tcsnicmp(m_sPath, options.m_sThumbPath[i], options.m_sThumbPath[i].GetLength()) == 0) {
			if(options.m_sThumbPath[i].CompareNoCase(m_sPath) == 0) {
				options.m_sThumbPath.RemoveAt(i);
				break;
			}
			if(IsThumbnail()) return;
		}
	}
	if(IsThumbnail()) {
		if(options.m_sThumbPath.GetSize() >= options.m_nHistMax)
			options.m_sThumbPath.SetSize(options.m_nHistMax);
		options.m_sThumbPath.InsertAt(0, m_sPath);
	}
}

BOOL CZBListView::IsThumbPath(LPCTSTR pPath)
{
	for_to(i, options.m_sThumbPath.GetSize()) {
		if(_tcsnicmp(pPath, options.m_sThumbPath[i], options.m_sThumbPath[i].GetLength()) == 0)
			return TRUE;
	}
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
// Archiver operation ###1.24


BOOL CZBListView::IsArchiveFile(LPCTSTR pPath, BOOL bLoadDll)
{
	if(m_archiver.IsArchive(pPath) != -1) {
		return m_archiver.LoadDll(pPath);
	}
	return FALSE;
}

BOOL CZBListView::IsListExtract()
{
	if(IsArcList()) return TRUE;
	int iSel = GetFirstSelectedItem();
	return iSel != -1 && IsArchiveFile(GetListCtrl().GetItemText(iSel, 0));
}

BOOL CZBListView::ExtractArchive(CFindData& data, CString& sPath)
{
	CFileName sFile = data.GetSubPathName(TRUE);
	m_archiver.SetArchive(AfxGetMainWnd()->GetSafeHwnd(), m_sPath);
	if(!m_archiver.Extract(sFile, GetTempPath())) return FALSE;
	sPath = GetTempPath() + data.GetFileName();
	m_sarrTmpFiles.Add(data.GetFileName());
	return TRUE;
}

void CZBListView::DeleteTempFiles()
{
	CString sTempPath = GetTempPath();
	for_to(i, m_sarrTmpFiles.GetSize()) {
		CString sTempFile = sTempPath + m_sarrTmpFiles[i];
		::DeleteFile(sTempFile);
	}
	m_sarrTmpFiles.RemoveAll();
}

CString CZBListView::GetTempPath()
{
	CString sPath;
	::GetTempPath(MAX_PATH, sPath.GetBuffer(MAX_PATH));
	sPath.ReleaseBuffer();
	return sPath;
}

void CZBListView::Extract()
{
	CListCtrl& list = GetListCtrl();
	int iSel = GetFirstSelectedItem();
	if(IsArcList()) {
		if(iSel == -1 || ::GetFocus() != GetSafeHwnd()) 
			ExtractAll(m_sPath);
		else
			ExtractSelectedFiles();
	} else if(iSel != -1) {
		POSITION pos = list.GetFirstSelectedItemPosition();
		while(pos) {
			ExtractAll(m_finder.GetFullPath(list.GetItemData(list.GetNextSelectedItem(pos))));
		}
	}
}

void CZBListView::ExtractAll(CFileName sArcFile)
{
	CFileName sExtractDir = sArcFile.GetNoExt();
	if(::CreateDirectory(sExtractDir, NULL))
		::SHChangeNotify(SHCNE_MKDIR, SHCNF_PATH, (LPCTSTR)sExtractDir, NULL);
	m_archiver.SetArchive(AfxGetMainWnd()->GetSafeHwnd(), sArcFile);
	sExtractDir.AddSlash();
	m_archiver.ExtractAll(sExtractDir);
}

void CZBListView::ExtractSelectedFiles()
{
}
