// ZBListVw.h : interface of the CZBListView class
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_ZBRVIEW_H__B884EA92_9BC3_11D2_AE74_00400533897F__INCLUDED_)
#define AFX_ZBRVIEW_H__B884EA92_9BC3_11D2_AE74_00400533897F__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include <afxcview.h>
#include <shlobj.h>
#include "FTP.h"
#include "FTPFile.h"

/////////////////////////////////////////////////////////////////////////////
// class CTreeSize

class CTreeSize;
class CFindData;
class CZBListView;

class CTreeNode
{
public:
	CTreeNode();
	~CTreeNode();
	CTreeSize* m_pSubTree;
	CString m_sName;
	DWORDLONG m_dwlDiskSize;
	DWORDLONG m_dwlFileSize;
};

class CTreeSize : public CArray<CTreeNode, CTreeNode&>
{
public:
	CTreeSize();
	~CTreeSize();
	BOOL AddNode(LPCTSTR pRoot, LPCTSTR pDir, CTreeNode* pParent);
	int FindNode(LPCTSTR pName);
	static CZBListView* m_pView;
private:
	static CFTPFindFiles m_finder;
};

/////////////////////////////////////////////////////////////////////////////
// class CThumbImage

class CThumbImage
{
public:
	CThumbImage();
	~CThumbImage();
	void Init(CView* pView);
	HBITMAP IconToBitmap(HICON hIcon);
#ifdef XIMAGE
	HBITMAP createImage(LPCTSTR pPath);
	int GetXImageFormat(LPCTSTR pExt);
#else
	HBITMAP ResizeToFitBitmap(HBITMAP hBmp);
#endif
protected:
	CView* m_pView;
	CBitmap m_bmpIcon;
	CSize m_cIcon;
	CPoint m_ptIcon;
};

/////////////////////////////////////////////////////////////////////////////
// class CZBListView

enum ListType { LIST_NONE = -1, LIST_FILE, LIST_TREE, LIST_UNFOLD, LIST_DRIVE, LIST_ARC, LISTTYPE_MAX };

#define LISTVIEW_ICON		0
#define LISTVIEW_REPORT		1
#define LISTVIEW_THUMBNAIL	2	// flag bit

#define STYLE_REALNAME 8
#define STYLE_FULLROW  16

class CBZTreeView;

class CZBListView : public CListView
{
protected: // create from serialization only
	CZBListView();
	DECLARE_DYNCREATE(CZBListView)

// Attributes
private:
	CImageList* m_pDragImage;
	CString m_sPath;
	ListType m_listtype;
	int *m_pnColumnTable;
	int m_nSort;
	BOOL m_bReverse;
	CTreeSize* m_pTreeSize;
	BOOL m_bTreeSize;
	BOOL m_bBusy;
	BOOL m_bRefresh;
	HCURSOR m_hCsrBusy;
	HANDLE m_hChangeNotify;
	CFileName m_sDropSrcDir;
	CFileName m_sDropDir;
	int m_iDropItem;
	COleDropTarget m_dropTarget;
	BOOL m_bNewFolder;
	CString m_sFileMasks;
	CImageList m_imageSort;
	BOOL m_bUnfold;			// ###1.20
	BOOL m_bTaskBusy;
	BOOL m_bNotSort;
	int m_iPrevTopIndex;	// ###1.20
	CStringArray m_sarrTmpFiles; // ###1.24
	UINT m_style;		// ###1.24
	CImageList m_imageThumbs;
	CWinThread* m_pDrawThread;
	BOOL m_bVScroll;
public:
	CFTPFindFiles* m_pDragFinder;
	CFileName m_sPrevSelect;

// Operations
public:
	void CalcTreeSize();
	void FillItems(LPCTSTR pPath, BOOL bFTP = FALSE);
	void FillDriveItems(LPCTSTR pPath, LPSHELLFOLDER pSF, LPITEMIDLIST pIDLParent);
	void NewItem(BOOL bFolder);
	BOOL IsFileList() { return m_listtype != LIST_DRIVE; }
	BOOL IsArcList() { return m_listtype == LIST_ARC; }		// ###1.24
	BOOL IsArchiveFile(LPCTSTR pPath, BOOL bLoadDll = FALSE);
	void SetRefresh() { m_bRefresh = TRUE; }
	void SelectAll(BOOL bInvert = FALSE);
	BOOL ClearAll();
	void Init(ListType type);
	void DisplayInfo();
	void OnUpdateViewRealName(CCmdUI* pCmdUI);
	void IsChangeNotify();
	void GetColumnWidth();
	void SetFileMask(LPCTSTR pMask = NULL);
	CString GetFileMask() { return m_sFileMasks; }
	void OnEditFilemask();
	BOOL IsFileMaskMode();
	void DblClickItem(int iItem);
	void SetUnfold(BOOL bUnFold) { m_bUnfold = bUnFold; }
	BOOL IsUnfold() { return m_bUnfold; }
	BOOL IsTaskBusy() { return m_bTaskBusy; }
	void StopOperation() { m_bTaskBusy = FALSE; }
	void OnViewUnfold();
	void SelectDuplicateFiles(); // ###1.20
	void SetPrevSelect(LPCTSTR pNextPath);
	BOOL IsListExtract();
	void Extract();
	void ExtractAll(CFileName sArcFile);
	void ExtractSelectedFiles();
	void SetListStyle(UINT style, BOOL bAuto = FALSE);
	UINT GetListStyle() { return (m_style > LISTVIEW_THUMBNAIL) ? LISTVIEW_THUMBNAIL : m_style; }
	BOOL IsThumbnail() { return m_style & LISTVIEW_THUMBNAIL; }
	CString GetSelectedFileTitle();

// Implementation
protected:
//	void LoadProfiles();
//	void SaveProfiles();
	CZBTreeView* GetTreeView() { return ((CZBTreeView*)GetNextWindow(GW_HWNDPREV)); }
	static UINT FillItemsThread(LPVOID pParam);
	static BOOL FillSubFolders(CZBListView* pView, CFileName sFolder);
	int GetFirstSelectedItem();
	void ScanDuplicateFiles(BOOL bCompare);
	void ReSelectDuplicateFiles();
	int FindItem(CFindData* pData);
	void ReCalcTreeSize();
	void SetStyle(UINT style);
	static UINT DrawIconsThread(LPVOID pParam); // ###1.24
	void TerminateThread(CWinThread** ppThread);

private:
	CString FormatFileSize(DWORDLONG dwlSize);
	CString FormatFileTime(FILETIME* pFT);
	static int CALLBACK CompareProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);
	BOOL InsertDriveItem(LPITEMIDLIST pIDL);
	 CTreeNode* FindNode(LPCTSTR pPath);
	HGLOBAL CreateHDrop();
	HGLOBAL CreateShellIDList();
	CString GetFullPath(int i);
	void DoContextMenu(POINT pt);
	void DrawHeaderImage();
	DWORD GetBytesPerCluster();
	BOOL ExtractArchive(CFindData& data, CString& sPath); // ###1.24
	void DeleteTempFiles();
	CString GetTempPath();
	int ExtractThumbImage(LPCTSTR pPath, CThumbImage *pThumbImage);
	void SetHeaderImageList(); // ###1.24
	void SetThumbPath();
	BOOL IsThumbPath(LPCTSTR pPath);

protected:
	virtual ~CZBListView();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CZBListView)
	public:
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
	virtual BOOL PreTranslateMessage(MSG* pMsg);
	virtual DROPEFFECT OnDragEnter(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point);
	virtual DROPEFFECT OnDragOver(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point);
	virtual void OnDragLeave();
	virtual BOOL OnDrop(COleDataObject* pDataObject, DROPEFFECT dropEffect, CPoint point);
	//}}AFX_VIRTUAL

// Generated message map functions
protected:
	//{{AFX_MSG(CZBListView)
	afx_msg void OnDestroy();
	afx_msg void OnColumnClick(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnDblclk(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnDeleteItem(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnEndLabelEdit(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg BOOL OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message);
	afx_msg void OnFileRemove();
	afx_msg void OnUpdateFileRemove(CCmdUI* pCmdUI);
	afx_msg void OnBeginDrag(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnClick(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg int OnMouseActivate(CWnd* pDesktopWnd, UINT nHitTest, UINT message);
	afx_msg void OnRClick(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnFileRename();
	afx_msg void OnUpdateFileRename(CCmdUI* pCmdUI);
	afx_msg void OnFileContextMenu();
	afx_msg void OnGetdispinfo(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnFileEnter();
	afx_msg void OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar);
	//}}AFX_MSG
	afx_msg LRESULT OnFillComplete(WPARAM hItem, LPARAM);
	afx_msg LRESULT OnFileOpComplete(WPARAM bDone, LPARAM);
	afx_msg LRESULT OnRefreshList(WPARAM hItem, LPARAM);
	DECLARE_MESSAGE_MAP()
};

/////////////////////////////////////////////////////////////////////////////

#ifndef NO_FTP
extern CFTP ftp2;
#endif

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_ZBRVIEW_H__B884EA92_9BC3_11D2_AE74_00400533897F__INCLUDED_)
