// ZBTreeVw.cpp : implementation file
//

#include "stdafx.h"
#include "ZBrowse.h"
#include "ZBTreeVw.h"
#include "ZBListVw.h"
#include "MainFrm.h"
#include "pidl.h"
#include "base.h"
#include "FTP.h"
#include "FTPFile.h"
#include "ZBFileOp.h"
#include <afxpriv.h>
#include "archiver.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define WM_FILL_COMPLETE (WM_USER + 10)
#define WM_REFRESH_TREE (WM_USER + 11)
#define WM_SHCHANGENOTIFY (WM_USER + 12)

void OpenSHChangeNotify();
void CloseSHChangeNotify();
void SetSHChangeNotify(HWND hwnd);

extern CArchiver m_archiver;

/////////////////////////////////////////////////////////////////////////////
// CZBTreeView diagnostics

#ifdef _DEBUG
void CZBTreeView::AssertValid() const
{
	CTreeView::AssertValid();
}

void CZBTreeView::Dump(CDumpContext& dc) const
{
	CTreeView::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CZBTreeView

IMPLEMENT_DYNCREATE(CZBTreeView, CTreeView)

BEGIN_MESSAGE_MAP(CZBTreeView, CTreeView)
	//{{AFX_MSG_MAP(CZBTreeView)
	ON_NOTIFY_REFLECT(TVN_ITEMEXPANDING, OnItemExpanding)
	ON_NOTIFY_REFLECT(TVN_DELETEITEM, OnDeleteItem)
	ON_NOTIFY_REFLECT(TVN_SELCHANGED, OnSelChanged)
	ON_WM_SETCURSOR()
	ON_COMMAND(ID_FILE_REMOVE, OnFileRemove)
	ON_UPDATE_COMMAND_UI(ID_FILE_REMOVE, OnUpdateFileRemove)
	ON_WM_CREATE()
	ON_WM_TIMER()
	ON_NOTIFY_REFLECT(TVN_BEGINDRAG, OnBeginDrag)
	ON_NOTIFY_REFLECT(TVN_KEYDOWN, OnKeyDown)
	ON_WM_DESTROY()
	ON_WM_MOUSEACTIVATE()
	ON_NOTIFY_REFLECT(TVN_BEGINLABELEDIT, OnBeginLabelEdit)
	ON_NOTIFY_REFLECT(TVN_ENDLABELEDIT, OnEndLabelEdit)
	ON_NOTIFY_REFLECT(NM_RCLICK, OnRClick)
	ON_COMMAND(ID_FILE_RENAME, OnFileRename)
	ON_UPDATE_COMMAND_UI(ID_FILE_RENAME, OnUpdateFileRename)
	ON_COMMAND(ID_FILE_CONTEXTMENU, OnFileContextMenu)
	ON_NOTIFY_REFLECT(NM_DBLCLK, OnDblclk)
	//}}AFX_MSG_MAP
	ON_MESSAGE(WM_FILL_COMPLETE, OnFillComplete)
	ON_MESSAGE(WM_REFRESH_TREE, OnRefreshTree)
	ON_MESSAGE(WM_FILEOP_COMPLETE, OnFileOpComplete)
	ON_MESSAGE(WM_SHCHANGENOTIFY, OnShellChangeNotify)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CZBTreeView construction/destruction

LPSHELLFOLDER g_pSFRoot = NULL;

CZBTreeView::CZBTreeView()
{
	m_hMyComputer = NULL;
	m_hNetwork = NULL;
	m_hFTPServer = NULL; // ###NO_FTP
	m_hSelected = NULL;
	m_bBusy = FALSE;
	m_bDeleting = FALSE;
	m_dwAttrSelected = 0;
	m_nScroll = 0;
	m_hTarget = NULL;
	m_bFindRetry = FALSE;
	m_pIDLNet = NULL;
}

CZBTreeView::~CZBTreeView()
{
	if(g_pSFRoot) g_pSFRoot->Release();
}

BOOL CZBTreeView::PreCreateWindow(CREATESTRUCT& cs) 
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs
	cs.style |= (TVS_HASLINES | TVS_LINESATROOT | TVS_HASBUTTONS | TVS_EDITLABELS | TVS_SHOWSELALWAYS); //###1.04
	return CTreeView::PreCreateWindow(cs);
}

int CZBTreeView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CTreeView::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	// TODO: Add your specialized creation code here
	m_dropTarget.Register(this);
	m_hCsrBusy = AfxGetApp()->LoadStandardCursor(IDC_WAIT);
	HIMAGELIST hImage = GetShellImageList(SHGFI_SMALLICON);
	TreeView_SetImageList(GetTreeCtrl().GetSafeHwnd(), hImage, TVSIL_NORMAL);

	OpenSHChangeNotify();
	return 0;
}


void CZBTreeView::OnDestroy() 
{
	TreeView_SetImageList(GetTreeCtrl().GetSafeHwnd(), NULL, TVSIL_NORMAL);

	CloseSHChangeNotify();
	CTreeView::OnDestroy();
}

/////////////////////////////////////////////////////////////////////////////
// CZBTreeView initializing

void CZBTreeView::OnInitialUpdate() 
{
	CTreeView::OnInitialUpdate();
	// TODO: Add your specialized code here and/or call the base class

	HRESULT hr = ::SHGetDesktopFolder(&g_pSFRoot);
	if(SUCCEEDED(hr)) {
//		FillItems(g_pSFRoot, TVI_ROOT);
		InsertRootItems();
	}
 //   if(pSF)pSF->Release();
	if(!options.m_sBrowseFolder.IsEmpty())
		SelectFrom(options.m_sBrowseFolder);
}

void CZBTreeView::InsertRootItems()
{
	m_hMyComputer = InsertRootItem(CSIDL_DRIVES);
	m_hNetwork = InsertRootItem(CSIDL_NETWORK);
#ifndef NO_FTP
	m_hFTPServer = InsertFTP();
#endif
	SetSHChangeNotify(GetSafeHwnd());		// using "SHChangeNotifyRegister" ###1.20
}

/////////////////////////////////////////////////////////////////////////////
// CZBTreeView drawing

void CZBTreeView::OnDraw(CDC* pDC)
{
	// TODO: add draw code here
}

/////////////////////////////////////////////////////////////////////////////
// CZBTreeView tree managing

static LPSHELLFOLDER pSF = NULL;
static HTREEITEM hParent = NULL;
static LPITEMIDLIST pIDLParent = NULL;

int CALLBACK CZBTreeView::CompareProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	return ComparePIDs(g_pSFRoot, (LPITEMIDLIST)lParam1, (LPITEMIDLIST)lParam2);
}

UINT CZBTreeView::FillItemsThread(LPVOID pParam)
{
	CZBTreeView* pView = (CZBTreeView*)pParam;
	CTreeCtrl& tree = pView->GetTreeCtrl();
	HWND hWnd = AfxGetMainWnd()->m_hWnd;
	LPITEMIDLIST pIDL = NULL;
	LPENUMIDLIST pEnumIDL = NULL;
	int nItemCount = 0;
	if(SUCCEEDED(pSF->EnumObjects(hWnd, SHCONTF_FOLDERS | SHCONTF_INCLUDEHIDDEN, &pEnumIDL)) &&  pEnumIDL) {
		while(pEnumIDL->Next(1, &pIDL, NULL) == NOERROR) {
			ULONG attr = SFGAO_FILESYSTEM | SFGAO_FILESYSANCESTOR;
			pSF->GetAttributesOf(1, (LPCITEMIDLIST*)&pIDL, &attr);
			if(attr & (SFGAO_FILESYSTEM | SFGAO_FILESYSANCESTOR)) {
				pView->InsertItem(pIDL, pSF, hParent, pIDLParent);
				g_pMalloc->Free(pIDL);
				pIDL = NULL;
				nItemCount++;
			}
		}
		pEnumIDL->Release();	
	}
	if(pSF != g_pSFRoot)
		pSF->Release();

	if(nItemCount == 0)	//###1.24 ???  (if item has A ZIP file)
		pView->SetChildren(hParent, 0);
	else {
//	if(nItemCount) {
		TV_SORTCB tvscb;
		tvscb.hParent = hParent;
		tvscb.lpfnCompare = CompareProc;
		tvscb.lParam = 0;
		tree.SortChildrenCB(&tvscb);
		tree.Expand(hParent, TVE_EXPAND);
	}
	pView->PostMessage(WM_FILL_COMPLETE, (WPARAM)hParent);
	return 1;
}

// CZBFindArchives class ###1.24

class CZBFindArchives : public CFTPFindFiles
{
public:
	virtual BOOL PostFind(CFindData* pData);
	HTREEITEM m_hParent;
	CString m_sSelectFile;
} m_arcFinder;

BOOL CZBFindArchives::PostFind(CFindData* pData)
{
	return m_archiver.IsArchive(pData->GetFileName());
}

BOOL CZBTreeView::IsTreeComress()
{
	if(!m_hSelected) return FALSE;
	return m_dwAttrSelected & SFGAO_FILESYSTEM;
}

UINT CZBTreeView::FillArchivesThread(LPVOID pParam)
{
	CZBTreeView* pView = (CZBTreeView*)pParam;
	CTreeCtrl& tree = pView->GetTreeCtrl();

	HTREEITEM hItem = tree.GetChildItem(m_arcFinder.m_hParent);
	HTREEITEM hSelected = tree.GetSelectedItem();
	while(hItem) {
		HTREEITEM hItemNext = tree.GetNextSiblingItem(hItem);
		if(hItem != hSelected && m_archiver.IsArchive(tree.GetItemText(hItem)))
			tree.DeleteItem(hItem);
		hItem = hItemNext;
	}

	CString sPath = GetNameOf(g_pSFRoot, (LPITEMIDLIST)tree.GetItemData(m_arcFinder.m_hParent), SHGDN_NORMAL|SHGDN_FORPARSING);
	if(!sPath.IsEmpty() && m_arcFinder.FindAll(sPath, NULL)) {
		m_arcFinder.Sort(SortByName);

		HTREEITEM hInsert = tree.GetPrevSiblingItem(hSelected);
		if(hInsert == NULL) hInsert = m_arcFinder.m_hParent;

		CFindData* pData = m_arcFinder.GetHead();
		while(pData) {
			if(m_arcFinder.m_sSelectFile.CompareNoCase(pData->GetFileName()))
				hInsert = pView->InsertFile(m_arcFinder.GetFullPath(pData), m_arcFinder.m_hParent, hInsert);
			else
				hInsert = hSelected;
			pData = pData->GetNext();
		}
	}

	pView->PostMessage(WM_FILL_COMPLETE);
	return 1;
}

void CZBTreeView::FillItems(LPSHELLFOLDER _pSF, HTREEITEM _hParent, LPITEMIDLIST _pIDLParent)
{
	if(m_bBusy) return;
	pSF = _pSF;
	hParent = _hParent;
	pIDLParent = _pIDLParent;
	m_bBusy = TRUE;
	AfxBeginThread(FillItemsThread, (LPVOID)this);
}

HTREEITEM CZBTreeView::InsertItem(LPITEMIDLIST pIDL, LPSHELLFOLDER pSF, HTREEITEM hParent, LPITEMIDLIST pIDLParent)
{
	ULONG attr = SFGAO_HASSUBFOLDER;
	pSF->GetAttributesOf(1, (LPCITEMIDLIST*)&pIDL, &attr);
	TV_INSERTSTRUCT tvi;
	ZeroMem(tvi);
	tvi.hParent = hParent;
	tvi.hInsertAfter = TVI_LAST;
	tvi.item.mask = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM | TVIF_STATE;
	if(attr & SFGAO_HASSUBFOLDER) {
		tvi.item.cChildren=1;
		tvi.item.mask |= TVIF_CHILDREN;
	}
	CString sName = GetNameOf(pSF, pIDL, SHGDN_NORMAL);
	tvi.item.pszText = (LPTSTR)(LPCTSTR)sName;
	LPITEMIDLIST pIDLAbs = ConcatPIDLs(pIDLParent, pIDL);
	tvi.item.lParam = (LONG)pIDLAbs;
	tvi.item.iImage = GetFolderIcon(pIDLAbs, SHGFI_SMALLICON);
	tvi.item.iSelectedImage = GetFolderIcon(pIDLAbs, SHGFI_SMALLICON | SHGFI_OPENICON);
	return GetTreeCtrl().InsertItem(&tvi);
}

HTREEITEM CZBTreeView::InsertItem(LPCTSTR pPath, HTREEITEM hParent, BOOL bFile, HTREEITEM hInsertAfter)
{
	CTreeCtrl& tree = GetTreeCtrl();

	if(!tree.ItemHasChildren(hParent))
		SetChildren(hParent, 1);

	CString sName = CFileName(pPath).GetName();
	BOOL bArchive = m_archiver.IsArchive(sName);
	if(!bArchive) {
		if(!(tree.GetItemState(hParent, TVIS_EXPANDED | TVIS_EXPANDEDONCE) & (TVIS_EXPANDED | TVIS_EXPANDEDONCE)))
			return NULL;
	}

	HTREEITEM hItem = tree.GetChildItem(hParent);
	HTREEITEM hPrevItem = hParent;
	if(hItem) {
		if(hInsertAfter == TVI_LAST || hInsertAfter == TVI_SORT) {
			while(hItem) {
				CString sLabel = tree.GetItemText(hItem);
				if(sName.CompareNoCase(sLabel) == 0)
					return NULL;
				if(!bArchive && m_archiver.IsArchive(sLabel)) {
					hInsertAfter = hPrevItem;
					break;
				}
				hPrevItem = hItem;
				hItem = tree.GetNextSiblingItem(hItem);
			}
		}
	}// else
	 //	SetChildren(hParent, 1);

//	tree.SetItemState(hParent, TVIS_EXPANDEDONCE, TVIS_EXPANDEDONCE);

	TV_INSERTSTRUCT tvi;
	ZeroMem(tvi);
	tvi.hParent = hParent;
	tvi.hInsertAfter = hInsertAfter;
	tvi.item.mask = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM | TVIF_STATE;

	LPITEMIDLIST pIDL = GetPIDLOf(g_pSFRoot, pPath);
	if(!bFile && GetFolderAttr(pIDL) & SFGAO_HASSUBFOLDER) {
		tvi.item.cChildren=1;
		tvi.item.mask |= TVIF_CHILDREN;
	}
	tvi.item.pszText = (LPTSTR)(LPCTSTR)sName;
	tvi.item.lParam = (LONG)pIDL;
	tvi.item.iImage = GetFolderIcon(pPath, SHGFI_SMALLICON);
	tvi.item.iSelectedImage = GetFolderIcon(pPath, SHGFI_SMALLICON | SHGFI_OPENICON);

	return tree.InsertItem(&tvi);
}

void CZBTreeView::DeleteItem(HTREEITEM hItem)	// ###1.20
{
	CTreeCtrl& tree = GetTreeCtrl();

	HTREEITEM hParent = tree.GetParentItem(hItem);
	tree.DeleteItem(hItem);

	if(hParent && !tree.GetChildItem(hParent)) {
		SetChildren(hParent, 0);
	}
}

HTREEITEM CZBTreeView::InsertRootItem(int nFolder)
{
	LPITEMIDLIST pIDL = NULL;
	HTREEITEM hItem = NULL;
	if(::SHGetSpecialFolderLocation(AfxGetMainWnd()->m_hWnd, nFolder, &pIDL) == NOERROR) {
		hItem = InsertItem(pIDL, g_pSFRoot, NULL);
		g_pMalloc->Free(pIDL);
	}
	return hItem;
}

// Find Path Start here!

HTREEITEM CZBTreeView::FindPath(LPCTSTR pPath, BOOL bExpand)
{
	if(!m_hMyComputer || !pPath || !*pPath) return NULL;
	TRACE(_T("*** FindPath %s\n"), pPath);
	CFileName sPath(pPath);
	HTREEITEM hParent;
#ifndef NO_FTP
	if (ftp.IsFTPPath(sPath)) {
		m_sFindPath = sPath;
		hParent = m_hFTPServer;
	} else {
#endif
		sPath.SlashToBackslash();
		m_sFindPath = sPath;
		if(m_pIDLNet) g_pMalloc->Free(m_pIDLNet);
		m_pIDLNet = NULL;
		if(!CFileName::IsSlash(sPath[0])) {
			hParent = m_hMyComputer;
		} else {
			hParent = m_hNetwork;
			if(options.m_iNetWalk) {
				LPITEMIDLIST pIDLParent = (LPITEMIDLIST)(GetTreeCtrl().GetItemData(hParent));
				m_pIDLNet = GetPIDLOf(g_pSFRoot, sPath);
				m_iNetWalk = options.m_iNetWalk;
			}
		}
#ifndef NO_FTP
	}
#endif
	GetMainFrame()->ReportMessage(IDS_INFO_SEARCH, m_sFindPath);
	if(bExpand) {
		FindPath(hParent);
		return NULL;
	} else
		return FindPath(hParent, FALSE);
}

void CZBTreeView::FindPath(HTREEITEM hParent)
{
	CTreeCtrl& tree = GetTreeCtrl();
	if(tree.GetItemState(hParent, TVIS_EXPANDEDONCE) & TVIS_EXPANDEDONCE)
		FindPathNext(hParent);
	else
		tree.Expand(hParent, TVE_EXPAND);
}

CString CZBTreeView::GetPath(HTREEITEM hItem)
{
	CString sPath;
	LPITEMIDLIST pIDL = (LPITEMIDLIST)GetTreeCtrl().GetItemData(hItem);
	if(pIDL) {
		sPath = GetNameOf(g_pSFRoot, pIDL, SHGDN_FORPARSING);
		if(sPath.IsEmpty()) {
			sPath = _T("\\\\");
			sPath += GetNameOf(g_pSFRoot, pIDL, SHGDN_NORMAL);
		}
	} else
		sPath = GetURL(hItem);
	return sPath;
}

LRESULT CZBTreeView::OnFillComplete(WPARAM wParam, LPARAM)
{
	m_bBusy = FALSE;
	PostMessage(WM_SETCURSOR);
	if(!m_sFindPath.IsEmpty())
		FindPathNext((HTREEITEM)wParam);
	else
		GetListView()->DisplayInfo();
	return 1;
}

void CZBTreeView::FindPathNext(HTREEITEM hParent)
{
	HTREEITEM hResult = FindPath(hParent, TRUE);
	if(hResult) {
		GetTreeCtrl().SelectItem(hResult);
		GetTreeCtrl().EnsureVisible(hResult);
		m_sFindPath.Empty();
		m_bFindRetry = FALSE;
	}
}

HTREEITEM CZBTreeView::FindPath(HTREEITEM hParent, BOOL bExpand)
{
	CTreeCtrl& tree = GetTreeCtrl();
	HTREEITEM hItem = tree.GetChildItem(hParent);
	HTREEITEM hResult = hItem ? NULL : hParent;
	while(hItem) {
		if(m_pIDLNet) {
			LPITEMIDLIST pIDLtree = (LPITEMIDLIST)GetTreeCtrl().GetItemData(hItem);
			LPITEMIDLIST pIDLwalk = GetSubPIDL(m_pIDLNet, m_iNetWalk);
			if(pIDLwalk) {
				LPSHELLFOLDER pSFParent = NULL;
				TRACE(_T("Tree=%s\n"), GetNameOf(g_pSFRoot, pIDLtree, SHGDN_NORMAL));
				TRACE(_T("Walk=%s\n"), GetNameOf(g_pSFRoot, pIDLwalk, SHGDN_NORMAL));
				if(!ComparePIDs(g_pSFRoot, pIDLtree, pIDLwalk)) {
					m_iNetWalk++;
					if(!ComparePIDs(g_pSFRoot, pIDLtree, m_pIDLNet)) {
						hResult = hItem;
						break;
					} else {
						if(bExpand) {
							FindPath(hItem);
							break;
						} else {
							hResult = hItem;
							hItem = tree.GetChildItem(hItem);
							continue;
						}
					}
				}
				g_pMalloc->Free(pIDLwalk);
			}
		} else {
			CString sItem = GetPath(hItem);
			int nItemLen = sItem.GetLength();
			if (!_tcsnicmp(sItem, m_sFindPath, nItemLen)) {
				if (!_tcsicmp(sItem, m_sFindPath)) {
					hResult = hItem;
					break;
				} else {
					if(nItemLen <= 3 || m_sFindPath[nItemLen] == TCHAR('\\') || m_sFindPath[nItemLen] == TCHAR('/')) {
						if(!tree.ItemHasChildren(hItem)) { // ###1.24
							hParent = hItem;
							goto NotFound;
						}
						if(bExpand) {
							FindPath(hItem);
							break;
						} else {
							hResult = hItem;
							hItem = tree.GetChildItem(hItem);
							continue;
						}
					}
				}
			}
		}
		hItem = tree.GetNextSiblingItem(hItem);
		if(!hItem) {
NotFound:
			if(m_archiver.IsArchive(m_sFindPath) && ::GetFileAttributes(m_sFindPath) != -1) {	// ###1.24
				hResult = InsertFile(m_sFindPath, hParent);
			} else {
				tree.SelectItem(hParent);
				if(!m_bFindRetry) {
					m_bFindRetry = TRUE;
					AfxGetMainWnd()->PostMessage(WM_COMMAND, ID_VIEW_REFRESH);
					return NULL;	// ###1.24 ???
				} else {
					m_bFindRetry = FALSE;
					CString sMsg;
					sMsg.Format(IDS_ERR_PATH, m_sFindPath);
					AfxMessageBox(sMsg);
					m_sFindPath.Empty();
				}
			}
		}
	}
	return hResult;
}

void CZBTreeView::Refresh()
{
	HTREEITEM hItem = GetTreeCtrl().GetSelectedItem();
	RefreshTree(hItem);
}

void CZBTreeView::RefreshTree(HTREEITEM hItem)
{
	if(!IsExistItem(hItem)) return;

	BOOL bFillList = IsAncestor(hItem, m_hSelected);
	m_bDeleting = TRUE;
	CTreeCtrl& tree = GetTreeCtrl();
	TRACE(_T("RefreshTree: %s\n"), tree.GetItemText(hItem));
	UINT state = 0;
	if(hItem) {
		state = tree.GetItemState(hItem, TVIS_EXPANDED | TVIS_EXPANDEDONCE | TVIS_SELECTED);
		if(tree.GetChildItem(hItem))
			tree.Expand(hItem, TVE_COLLAPSE | TVE_COLLAPSERESET);
	} else {
		tree.DeleteAllItems();
		InsertRootItems();
	}
	m_bDeleting = FALSE;
	GetListView()->SetRefresh();
	if(hItem) {
		int cChildren = 0;
		DWORD lParam = tree.GetItemData(hItem);
		if(lParam == 0 || (lParam && (GetFolderAttr((LPITEMIDLIST)lParam) & SFGAO_HASSUBFOLDER)))	// ### 1.23
			cChildren = 1;
		SetChildren(hItem, cChildren);
		if(state & TVIS_EXPANDED)
			tree.Expand(hItem, TVE_EXPAND);
	}
	if(bFillList) {
		HTREEITEM hItem = tree.GetSelectedItem();
		if(IsExistItem(hItem))
			FillListView(m_hSelected = hItem);
	}
}

#ifndef NO_FTP
void CZBTreeView::RefreshFTP()
{
	GetTreeCtrl().DeleteItem(m_hFTPServer);
	m_hFTPServer = InsertFTP();
}
#endif

LRESULT CZBTreeView::OnRefreshTree(WPARAM hItem, LPARAM)
{
	RefreshTree((HTREEITEM)hItem);
	return 1;
}

BOOL CZBTreeView::IsAncestor(HTREEITEM hParent, HTREEITEM hChild)
{
	while(hChild) {
		if(hChild == hParent) return TRUE;
		hChild = GetTreeCtrl().GetParentItem(hChild);
	}
	return FALSE;
}

BOOL CZBTreeView::IsExistItem(HTREEITEM hItem)
{
	TV_ITEM item;
	item.hItem = hItem;
	item.mask = TVIF_STATE;
	item.stateMask = 0;
	item.state = 0;
	return GetTreeCtrl().SendMessage(TVM_GETITEM, 0, (LPARAM)&item);
}

void CZBTreeView::SetChildren(HTREEITEM hItem, int cChildren)
{
	TV_ITEM item;
	item.mask = TVIF_HANDLE | TVIF_CHILDREN;
	item.hItem = hItem;
	item.cChildren = cChildren;
	GetTreeCtrl().SetItem(&item);
	if(cChildren == 0)
		TRACE(_T(""));
}

void CZBTreeView::FolderUp() 
{
	// TODO: Add your command handler code here
	if(m_hSelected) {
		CTreeCtrl& tree = GetTreeCtrl();
		GetListView()->m_sPrevSelect = tree.GetItemText(tree.GetSelectedItem());
		HTREEITEM hParent =  tree.GetParentItem(m_hSelected);
		if(hParent)
			tree.SelectItem(hParent);
	}
}

HTREEITEM CZBTreeView::FindArchive() 
{
	return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// FTP handlers

#ifndef NO_FTP

HTREEITEM CZBTreeView::InsertFTP()
{
	CTreeCtrl& tree = GetTreeCtrl();
	TV_INSERTSTRUCT tvi;
	ZeroMem(tvi);
	tvi.hInsertAfter = TVI_LAST;
	tvi.item.mask = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM | TVIF_CHILDREN;
	tvi.item.pszText = _T("FTP");
	tvi.item.iImage = GetFolderIcon(NULL, FILE_ATTRIBUTE_DIRECTORY, SHGFI_SMALLICON);
	tvi.item.iSelectedImage = GetFolderIcon(NULL, FILE_ATTRIBUTE_DIRECTORY, SHGFI_SMALLICON | SHGFI_OPENICON);
	tvi.item.cChildren=1;
	HTREEITEM hFTP = tree.InsertItem(&tvi);
	for_to(i, ftp.GetServerCount()) {
		tvi.item.mask |= TVIF_CHILDREN;
		tvi.hParent = hFTP;
		tvi.item.pszText = (LPTSTR)(LPCTSTR)ftp.m_arrFTP[i].m_sServer;
//		tvi.item.lParam = i;
		tvi.item.cChildren=1;
		tree.InsertItem(&tvi);
	}
/*	for_to_(i, 32) {
		tvi.item.mask = TVIF_IMAGE;
		tvi.item.iImage = i;
		tree.InsertItem(&tvi);
	}
*/	return hFTP;
}
#endif // NO_FTP

CString CZBTreeView::GetURL(HTREEITEM hItem)
{
	CTreeCtrl& tree = GetTreeCtrl();
	
	CString sURL;
	while(hItem) {
		CString sItem = (hItem == m_hFTPServer) ? _T("ftp:/") : tree.GetItemText(hItem);
		sURL = (sURL.IsEmpty()) ? sItem : sItem + TCHAR('/') + sURL;
		hItem = tree.GetParentItem(hItem);
	}
	return sURL;
}

#ifndef NO_FTP

UINT CZBTreeView::FillFTPItemsThread(LPVOID pParam)
{
	CZBTreeView* pView = (CZBTreeView*)pParam;
	CTreeCtrl& tree = pView->GetTreeCtrl();
	CString sURL, sServer, sPath;
	sURL = pView->GetURL(hParent);
	ftp.ParseURL(sURL, sServer, sPath);
	TRACE(_T("Tree::FTP Connect to %s\n"), sServer);
	if(ftp.Connect(ftp.FindServer(sServer), 0, TRUE)) {
		CFTPFindFiles m_finder;
		TRACE(_T("Tree::FTP FindAll %s\n"), sPath);
		if(!m_finder.FindAll(sPath, ftp.m_hFTP, FIND_DIR))
			AfxMessageBox(GetErrorMessage());

		if(!(tree.GetItemState(hParent, TVIS_EXPANDED | TVIS_EXPANDEDONCE) & (TVIS_EXPANDED | TVIS_EXPANDEDONCE))) { // ###1.23
			if(!m_finder.GetCount())
				pView->SetChildren(hParent, 0);
			else {
				for_to(i, m_finder.GetCount()) {
					pView->InsertFTPItem(m_finder[i].GetFileName(), hParent);
				}
				tree.Expand(hParent, TVE_EXPAND);
			}
		}
	}
	pView->PostMessage(WM_FILL_COMPLETE, (WPARAM)hParent);
	return 1;
}

void CZBTreeView::FillFTPItems(CFTPFindFiles* pFinder) // ### 1.23
{
	CTreeCtrl& tree = GetTreeCtrl();
	HTREEITEM _hParent = tree.GetSelectedItem();

	TRACE(_T("FillFTPItems finder=%s,  parent=%s\n"), pFinder->GetFindDir(), tree.GetItemText(_hParent));
	//if(tree.GetItemState(_hParent, TVIS_EXPANDED | TVIS_EXPANDEDONCE) & (TVIS_EXPANDED | TVIS_EXPANDEDONCE))
	//	return;

	tree.Expand(_hParent, TVE_COLLAPSE | TVE_COLLAPSERESET);

	for_to(i, pFinder->GetCount()) {
		if(pFinder->ElementAt(i).IsDirectory()) {
			InsertFTPItem(pFinder->ElementAt(i).GetFileName(), _hParent);
		}
	}
	// tree.SetItemState(_hParent, TVIS_EXPANDEDONCE, TVIS_EXPANDEDONCE);
	tree.Expand(_hParent, TVE_EXPAND);
}

HTREEITEM CZBTreeView::InsertFTPItem(LPCTSTR lpszFolder, HTREEITEM _hParent)
{
	if(_hParent == NULL)
		_hParent = GetTreeCtrl().GetSelectedItem();

	TV_INSERTSTRUCT tvi;
	ZeroMem(tvi);
	tvi.hParent = _hParent;
	tvi.hInsertAfter = TVI_SORT;
	tvi.item.mask = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM | TVIF_CHILDREN;
	tvi.item.cChildren = 1;
	tvi.item.pszText = (LPTSTR)(LPCTSTR)lpszFolder;
	tvi.item.iImage = GetFolderIcon(NULL, FILE_ATTRIBUTE_DIRECTORY, SHGFI_SMALLICON);
	tvi.item.iSelectedImage = GetFolderIcon(NULL, FILE_ATTRIBUTE_DIRECTORY, SHGFI_SMALLICON | SHGFI_OPENICON);
	return GetTreeCtrl().InsertItem(&tvi);
}

void CZBTreeView::FillFTPItems(HTREEITEM _hParent)
{
	if(m_bBusy) return;
	hParent = _hParent;
	m_bBusy = TRUE;
	GetMainFrame()->ReportMessage(IDS_INFO_SEARCH, GetURL(hParent));
	AfxBeginThread(FillFTPItemsThread, (LPVOID)this);
}

#endif // NO_FTP

/////////////////////////////////////////////////////////////////////////////
// CZBTreeView message handlers

void CZBTreeView::OnItemExpanding(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
	// TODO: Add your control notification handler code here

	TVITEM& item = pNMTreeView->itemNew;
	CTreeCtrl& tree = GetTreeCtrl();
	if(pNMTreeView->action == TVE_EXPAND && !(item.state & TVIS_EXPANDEDONCE)) {
		if(item.lParam) {
			LPSHELLFOLDER pSFSub = NULL;
			LPITEMIDLIST pIDL = (LPITEMIDLIST)item.lParam;
			HRESULT hResult = g_pSFRoot->BindToObject(pIDL, NULL, IID_IShellFolder, (LPVOID *)&pSFSub);
			if(hResult == NOERROR)
				FillItems(pSFSub, item.hItem, pIDL);
			else
				TRACE(_T("BindToObject failed:%X\n"), hResult);
		} else {
#ifndef NO_FTP
			if(item.hItem == m_hFTPServer) {
				if(!ftp.GetServerCount())
					GetMainFrame()->PostMessage(WM_COMMAND, ID_FTP_SETTING);
				else
					PostMessage(WM_FILL_COMPLETE, (WPARAM)item.hItem);
			} else
				FillFTPItems(item.hItem);
#endif
		}
	}
	*pResult = 0;
}

void CZBTreeView::OnDeleteItem(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
	// TODO: Add your control notification handler code here

	if(pNMTreeView->itemOld.lParam != 0)
		g_pMalloc->Free((LPVOID)pNMTreeView->itemOld.lParam);
	*pResult = 0;
}

void CZBTreeView::OnSelChanged(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
	// TODO: Add your control notification handler code here
	if(!m_bDeleting) {
		FillListView(m_hSelected = pNMTreeView->itemNew.hItem);
	}
	*pResult = 0;
}

void CZBTreeView::OnDblclk(NMHDR* pNMHDR, LRESULT* pResult)	// ###1.24
{
	// TODO: Add your control notification handler code here
	CTreeCtrl& tree = GetTreeCtrl();
	CPoint pt;
	::GetCursorPos(&pt);
	ScreenToClient(&pt);
	UINT uFlags;
	HTREEITEM hItem = tree.HitTest(pt, &uFlags);
	if(hItem) {
		CString sText = tree.GetItemText(hItem);
		TRACE(_T("*** Dblclk %s\n"), sText);
		if(m_archiver.IsArchive(sText)) {
			HTREEITEM hParent = tree.GetParentItem(hItem);
			if(hParent) {
				m_sFindPath.Empty();
				m_arcFinder.m_hParent = hParent;
				m_arcFinder.m_sSelectFile = sText;
				m_bBusy = TRUE;
				AfxBeginThread(FillArchivesThread, (LPVOID)this);
			}
		}
	}
	*pResult = 0;
}

void CZBTreeView::FillListView(HTREEITEM hItem) 
{
	if(!IsExistItem(hItem)) return;
	if(hItem == m_hMyComputer && !m_sFindPath.IsEmpty()) return;
	DWORD lParam = GetTreeCtrl().GetItemData(hItem);
	if(lParam) {
		LPITEMIDLIST pIDL = (LPITEMIDLIST)lParam;

		m_dwAttrSelected = GetFolderAttr(pIDL);
		if(m_dwAttrSelected == 0 || m_dwAttrSelected == INVALID) {
			RefreshTree(GetTreeCtrl().GetParentItem(hItem));
			return;
		}
		BOOL bRoot = hItem == m_hMyComputer || hItem == m_hNetwork;
		CString sPath = GetNameOf(g_pSFRoot, pIDL, bRoot? SHGDN_NORMAL : SHGDN_FORPARSING);		// Win2K
//		CString sMsg; sMsg.Format("%s(%X)\n", sPath, m_dwAttrSelected);
//		GetMainFrame()->ReportMessage(sMsg);
//		GetMainFrame()->SetCaption(sPath);
		if(!sPath.IsEmpty() && m_dwAttrSelected & SFGAO_FILESYSTEM) {
			GetMainFrame()->SetFolder(sPath);
			GetListView()->FillItems(sPath);
		} else {
			GetMainFrame()->m_wndToolBar.m_combo.SetText("");
			if(m_dwAttrSelected & SFGAO_FILESYSANCESTOR) {
				LPSHELLFOLDER pSFSub = NULL;
				if(g_pSFRoot->BindToObject(pIDL, NULL, IID_IShellFolder, (LPVOID *)&pSFSub) == NOERROR) {
					GetListView()->FillDriveItems(sPath, pSFSub, pIDL);
				}
			}
		}
		// TRACE(_T("%s :%X\n"), sPath, m_dwAttrSelected);
	} else {
		m_dwAttrSelected = 0;
		CString sPath = GetURL(hItem);
//		GetMainFrame()->SetCaption(sPath);
		GetMainFrame()->SetFolder(sPath);
#ifndef NO_FTP
		if (hItem != m_hFTPServer && (ftp.IsConnect() || !m_sFindPath.IsEmpty()))
#else
		if (!m_sFindPath.IsEmpty())
#endif
			GetListView()->FillItems(sPath, TRUE);
		else
			GetListView()->Init(LIST_FILE);
	}
}

void CZBTreeView::OnBeginLabelEdit(NMHDR* pNMHDR, LRESULT* pResult) 
{
	TV_DISPINFO* pTVDispInfo = (TV_DISPINFO*)pNMHDR;
	// TODO: Add your control notification handler code here
	if(pTVDispInfo->item.lParam	 && !((GetFolderAttr((LPITEMIDLIST)pTVDispInfo->item.lParam))
	 & SFGAO_CANRENAME))
		*pResult = 1;
	else
		*pResult = 0;
}

void CZBTreeView::OnEndLabelEdit(NMHDR* pNMHDR, LRESULT* pResult) 
{
	TV_DISPINFO* pTVDispInfo = (TV_DISPINFO*)pNMHDR;
	// TODO: Add your control notification handler code here
	BOOL bResult = FALSE;
	if(pTVDispInfo->item.pszText != NULL) {
		HTREEITEM hItem = pTVDispInfo->item.hItem;
		LPCITEMIDLIST pIDL = (LPCITEMIDLIST)pTVDispInfo->item.lParam;
		if(pIDL) {
			LPSHELLFOLDER pSFParent = NULL;
			HTREEITEM hParent = GetTreeCtrl().GetParentItem(hItem);
			if(hParent) {
				LPITEMIDLIST pIDLParent = (LPITEMIDLIST)(GetTreeCtrl().GetItemData(hParent));
				g_pSFRoot->BindToObject(pIDLParent, NULL, IID_IShellFolder, (LPVOID *)&pSFParent);
			} else
				pSFParent = g_pSFRoot;
			if(pSFParent) {
				LPITEMIDLIST pIDLNew = NULL;
				USES_CONVERSION;
				if(pSFParent->SetNameOf(AfxGetMainWnd()->m_hWnd, GetLastPIDL(pIDL)
					, T2OLE(pTVDispInfo->item.pszText), SHGDN_FORPARSING, &pIDLNew) == NOERROR) {
					GetTreeCtrl().SetItemData(hItem, (DWORD)ConcatPIDLs(pIDLParent, pIDLNew));
					bResult = TRUE;
				}
				g_pMalloc->Free(pIDLNew);
				if(pSFParent != g_pSFRoot)
					pSFParent->Release();
			}
		} else {
#ifndef NO_FTP
			CFTPFile file;
			file.m_hFTP = ftp.m_hFTP;
			CFileName sPathOld = GetPath(hItem);
			CFileName sPathNew = sPathOld.GetDir() + pTVDispInfo->item.pszText;
			sPathOld = ftp.GetRoot(sPathOld);
			sPathNew = ftp.GetRoot(sPathNew);
			bResult = file.Rename(sPathOld, sPathNew);
#endif
		}
		if(bResult)
			FillListView(hItem);	// ###1.23
	}

	*pResult = bResult;
}

void CZBTreeView::OnFileRename() 
{
	// TODO: Add your command handler code here
	GetTreeCtrl().EditLabel(GetTreeCtrl().GetSelectedItem());
}

void CZBTreeView::OnUpdateFileRename(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(GetTreeCtrl().GetSelectedItem() != NULL);	
}

void CZBTreeView::OnRClick(NMHDR* pNMHDR, LRESULT* pResult) 
{
	// TODO: Add your control notification handler code here
	*pResult = 0;
	TV_HITTESTINFO tvhti;
	POINT pt;
	::GetCursorPos(&pt);
	tvhti.pt = pt;
	ScreenToClient(&tvhti.pt);
	GetTreeCtrl().HitTest(&tvhti);
	if(tvhti.hItem)
		DoContextMenu(tvhti.hItem, pt);
}

void CZBTreeView::OnFileContextMenu() 
{
	// TODO: Add your command handler code here
	CTreeCtrl& tree = GetTreeCtrl();
	HTREEITEM hItem = tree.GetSelectedItem();
	if(hItem) {
		CRect rItem;
		tree.GetItemRect(hItem, rItem, TRUE);
		POINT pt = rItem.CenterPoint();
		ClientToScreen(&pt);
		DoContextMenu(hItem, pt);
	}
}

void CZBTreeView::DoContextMenu(HTREEITEM hItem, POINT pt)
{
	CTreeCtrl& tree = GetTreeCtrl();
	LPCITEMIDLIST pIDL = (LPCITEMIDLIST)tree.GetItemData(hItem);
	if(pIDL) {
		LPCITEMIDLIST pIDLItem = GetLastPIDL(pIDL);
		HTREEITEM hParent = tree.GetParentItem(hItem);
		if(!hParent)
			GetMainFrame()->PopupContextMenu(g_pSFRoot, 1, &pIDLItem, pt);
		else {
			LPITEMIDLIST pIDLParent = (LPITEMIDLIST)(tree.GetItemData(hParent));
			LPSHELLFOLDER pSFParent = NULL;
			if(g_pSFRoot->BindToObject(pIDLParent, NULL, IID_IShellFolder, (LPVOID *)&pSFParent) == NOERROR) {
				GetMainFrame()->PopupContextMenu(pSFParent, 1, &pIDLItem, pt);
				pSFParent->Release();
			}
		}
	}
}

BOOL CZBTreeView::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message) 
{
	// TODO: Add your message handler code here and/or call default
	if(nHitTest == HTCLIENT && pWnd == this && m_bBusy) {
		SetCursor(m_hCsrBusy);
		return TRUE;
	}
	return CTreeView::OnSetCursor(pWnd, nHitTest, message);
}

void CZBTreeView::OnKeyDown(NMHDR* pNMHDR, LRESULT* pResult) 
{
	TV_KEYDOWN* pTVKeyDown = (TV_KEYDOWN*)pNMHDR;
	// TODO: Add your control notification handler code here
	switch(pTVKeyDown->wVKey) {
	case VK_TAB:
		if(::GetKeyState(VK_SHIFT) < 0)
			GetMainFrame()->m_wndToolBar.SetFocus();
		else
			GetMainFrame()->SetActiveView(GetListView());
		*pResult = 1;
		return;
	}
	*pResult = 0;
}

int CZBTreeView::OnMouseActivate(CWnd* pDesktopWnd, UINT nHitTest, UINT message)
{
	int nResult = CWnd::OnMouseActivate(pDesktopWnd, nHitTest, message);

	if (nResult == MA_NOACTIVATE || nResult == MA_NOACTIVATEANDEAT
		|| nHitTest == HTHSCROLL ||  nHitTest == HTVSCROLL)
		return nResult;   // frame does not want to activate

	CFrameWnd* pParentFrame = GetParentFrame();
	if (pParentFrame != NULL)
	{
		// eat it if this will cause activation
		ASSERT(pParentFrame == pDesktopWnd || pDesktopWnd->IsChild(pParentFrame));

		// either re-activate the current view, or set this view to be active
		CView* pView = pParentFrame->GetActiveView();
		HWND hWndFocus = ::GetFocus();
		if (pView == this &&
			m_hWnd != hWndFocus && !::IsChild(m_hWnd, hWndFocus))
		{
			// re-activate this view
			OnActivateView(TRUE, this, this);
		}
		else
		{
			// activate this view
			pParentFrame->SetActiveView(this);
		}
		if(pView != this) {
			TV_HITTESTINFO tvhti;
			::GetCursorPos(&tvhti.pt);
			ScreenToClient(&tvhti.pt);
			if(GetTreeCtrl().HitTest(&tvhti) == m_hSelected && tvhti.flags & TVHT_ONITEMLABEL)
				nResult = MA_NOACTIVATEANDEAT;	// suppress Label Edit
		}	
	}
	return nResult;
}

/////////////////////////////////////////////////////////////////////////////
// CZBTreeView OLE Drag and Drop

#define SCROLL_MGN 16
#define SCROLL_INTERVAL 100
#define IDTIMER 2

CImageList* g_pDragImage = NULL;

void CZBTreeView::OnBeginDrag(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
	*pResult = 0;
	// TODO: Add your control notification handler code here
	m_hTarget = pNMTreeView->itemNew.hItem;
	if(pNMTreeView->itemNew.lParam) {
		DWORD dwAttr = GetFolderAttr((LPITEMIDLIST)pNMTreeView->itemNew.lParam);
		if(!(dwAttr & (SFGAO_CANCOPY + SFGAO_CANMOVE))) return ;	// ###1.01
	}
	if(g_pDragImage) delete g_pDragImage;
	g_pDragImage = GetTreeCtrl().CreateDragImage(m_hTarget);
	g_pDragImage->BeginDrag(0, CPoint(8, 8));
	CPoint pt(pNMTreeView->ptDrag);
	ClientToScreen(&pt);
	g_pDragImage->DragEnter(NULL, pt);
	COleDataSource dataSrc;
	HGLOBAL hDrop = CreateHDrop(GetPath(m_hTarget));
	dataSrc.CacheGlobalData(CF_HDROP, hDrop);
	dataSrc.DoDragDrop(DROPEFFECT_COPY | DROPEFFECT_MOVE);
	g_pDragImage->DragLeave(NULL);
	g_pDragImage->EndDrag();
	delete g_pDragImage;
	g_pDragImage = NULL;
	GetTreeCtrl().SelectDropTarget(NULL);
}

DROPEFFECT CZBTreeView::OnDragEnter(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point) 
{
	// TODO: Add your specialized code here and/or call the base class
	if(!pDataObject->IsDataAvailable(CF_HDROP)) return DROPEFFECT_NONE;
	HDROP hDrop = (HDROP)pDataObject->GetGlobalData(CF_HDROP);
	if(!hDrop) return DROPEFFECT_NONE;
	m_sDropSrc = HDropGetFileName(hDrop);
	::GlobalFree(hDrop);
	m_hDropItem = NULL;
	return OnDragOver(pDataObject, dwKeyState, point);
}

DROPEFFECT CZBTreeView::OnDragOver(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point) 
{
	// TODO: Add your specialized code here and/or call the base class
	if(!pDataObject->IsDataAvailable(CF_HDROP)) return DROPEFFECT_NONE;
	if(g_pDragImage) {
		CPoint pt(point);
		ClientToScreen(&pt);
		g_pDragImage->DragMove(pt);
	}
	CTreeCtrl& tree = GetTreeCtrl();
	UINT uFlags;
	HTREEITEM hItem = tree.HitTest(point, &uFlags);
	if(hItem != m_hDropItem) {
		m_hDropItem = hItem;
		if(g_pDragImage)
			g_pDragImage->DragShowNolock(FALSE);
		tree.SelectDropTarget(hItem);
		if(g_pDragImage)
			g_pDragImage->DragShowNolock(TRUE);
	}
	DROPEFFECT de = DROPEFFECT_NONE;
	if(hItem) {
		CRect rClient;
		GetClientRect(rClient);
		int nScroll = 0;
		if(point.y < SCROLL_MGN) nScroll = -1;
		else if(point.y >= rClient.Height() - SCROLL_MGN) nScroll = 1;
		if(nScroll) {
			if(!m_nScroll)
				SetTimer(IDTIMER, SCROLL_INTERVAL, NULL);
			m_nScroll = nScroll;
		} else if(m_nScroll) {
			KillTimer(IDTIMER);
			m_nScroll = 0;
		}
		de = DROPEFFECT_COPY;
		LPITEMIDLIST pIDL = (LPITEMIDLIST)tree.GetItemData(hItem);
		if(pIDL) {
			if(!(GetFolderAttr(pIDL) & SFGAO_FILESYSTEM))
				de = DROPEFFECT_NONE;
			else {
				CString sDrive = ((CFileName&)GetPath(hItem)).GetDrive();
				if(sDrive == m_sDropSrc.GetDrive())
					de = DROPEFFECT_MOVE;
				CFileName sDropSrcDir = m_sDropSrc;
				sDropSrcDir.GetParentDirectory();
				if(GetPath(hItem) == sDropSrcDir)
					de = DROPEFFECT_NONE;
			}
		}
		if(de != DROPEFFECT_NONE) {
			if(dwKeyState & MK_CONTROL) de = DROPEFFECT_COPY;
			if(dwKeyState & MK_SHIFT)   de = DROPEFFECT_MOVE;
		}
	}
	return de;
}

void CZBTreeView::OnDragLeave() 
{
	// TODO: Add your specialized code here and/or call the base class
	EndDrag();
	CTreeView::OnDragLeave();
}

void CZBTreeView::EndDrag() 
{
	CTreeCtrl& tree = GetTreeCtrl();
	if(g_pDragImage)
		g_pDragImage->DragShowNolock(FALSE);
	if(m_hDropItem)
		tree.SelectDropTarget(NULL);
	if(m_nScroll) {
		KillTimer(IDTIMER);
		m_nScroll = 0;
	}
}

void CZBTreeView::OnTimer(UINT nIDEvent) 
{
	// TODO: Add your message handler code here and/or call default
	CTreeCtrl& tree = GetTreeCtrl();
	HTREEITEM hItem = tree.GetDropHilightItem();
	if(hItem) {
		if(g_pDragImage)
			g_pDragImage->DragShowNolock(FALSE);
		if(m_nScroll > 0)
			hItem = tree.GetNextVisibleItem(hItem);
		else if(m_nScroll < 0)
			hItem = tree.GetPrevVisibleItem(hItem);
		else
			hItem = NULL;
		if(hItem) {
			tree.EnsureVisible(hItem);
			tree.SelectDropTarget(hItem);
		}
		if(g_pDragImage)
			g_pDragImage->DragShowNolock(TRUE);
	}
	CTreeView::OnTimer(nIDEvent);
}

BOOL CZBTreeView::OnDrop(COleDataObject* pDataObject, DROPEFFECT dropEffect, CPoint point) 
{
	// TODO: Add your specialized code here and/or call the base class
	HTREEITEM hItem = GetTreeCtrl().GetDropHilightItem();
	if(!hItem) return FALSE;
	m_fileOp.hDrop = (HDROP)pDataObject->GetGlobalData(CF_HDROP);
	if(!m_fileOp.hDrop) return FALSE;
	m_fileOp.wFunc = (dropEffect == DROPEFFECT_MOVE) ? FO_MOVE : FO_COPY;
	m_fileOp.sDstDir = GetPath(hItem);
	m_fileOp.hwnd = GetSafeHwnd();
	m_fileOp.pFinder = GetListView()->m_pDragFinder;
	m_bBusy = TRUE;
	m_fileOp.Start();
	EndDrag();
	return TRUE;
}

void CZBTreeView::OnFileRemove() 
{
	// TODO: Add your command handler code here
	HTREEITEM hItem = GetTreeCtrl().GetSelectedItem();
//	m_hTarget = GetTreeCtrl().GetParentItem(hItem);
	m_hTarget = hItem;	// ###1.24
	if(!hItem) return;
	CFileName sPath = GetPath(hItem);
	CString sMsg;
	sMsg.Format(IDS_REMOVE_TREE, sPath.GetName());
	if(AfxGetMainWnd()->MessageBox(sMsg, _T("Delete"), MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2) != IDYES)
		return;
	m_fileOp.hDrop = (HDROP)CreateHDrop(sPath);
	m_fileOp.wFunc = FO_DELETE;
	m_fileOp.sDstDir.Empty();
	m_fileOp.hwnd = GetSafeHwnd();
	m_fileOp.pFinder = NULL;
	m_bBusy = TRUE;
	m_fileOp.Start();
}

void CZBTreeView::OnUpdateFileRemove(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(m_dwAttrSelected & SFGAO_CANDELETE && !m_bBusy);

}

HGLOBAL CZBTreeView::CreateHDrop(LPCTSTR pPath)
{
	int len = _tcslen(pPath);
	DWORD size = sizeof(DROPFILES) + (len + 2) * sizeof(TCHAR);
	HGLOBAL hGlobal = ::GlobalAlloc(GHND | GMEM_SHARE, size);
	LPDROPFILES pDropFiles = (LPDROPFILES)::GlobalLock(hGlobal);
   	pDropFiles->pFiles = sizeof(DROPFILES);
#ifndef _UNICODE
	pDropFiles->fWide = FALSE;
#else
	pDropFiles->fWide = TRUE;
#endif
	_tcscpy((LPTSTR)(pDropFiles + 1), pPath);
	*((LPTSTR)(pDropFiles + 1) + len + 1) = 0;
	::GlobalUnlock(hGlobal);
 	return hGlobal;
}

LRESULT CZBTreeView::OnFileOpComplete(WPARAM bDone, LPARAM)
{
	m_bBusy = FALSE;
	PostMessage(WM_SETCURSOR);
	m_fileOp.pDlg->DestroyWindow();
	if(bDone) {
		GetListView()->SetRefresh();
		HTREEITEM hItemSrc = NULL;
		HTREEITEM hItemDst = NULL;
		switch(m_fileOp.wFunc) {
		case FO_MOVE:
			hItemSrc = m_fileOp.pFinder ? m_hSelected : GetTreeCtrl().GetParentItem(FindPath(m_sDropSrc, FALSE));
		case FO_COPY:
			hItemDst = m_hDropItem;
			break;
		case FO_DELETE:
			hItemDst = m_hTarget;	// ###1.24
			DWORD lParam = GetTreeCtrl().GetItemData(hItemDst);
			if(lParam) {
				::SHChangeNotify(SHCNE_RMDIR, SHCNF_IDLIST, (LPITEMIDLIST)lParam, NULL);
				return 1;
			}
			hItemDst = GetTreeCtrl().GetParentItem(hItemDst);	// FTP
			break;
		}
		if(hItemSrc) {
			if(IsAncestor(hItemSrc, hItemDst))
				hItemDst = hItemSrc;
			else if(!IsAncestor(hItemDst, hItemSrc))
				RefreshTree(hItemSrc);
		}
		RefreshTree(hItemDst);
	}
	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// using "SHChangeNotifyRegister" ###1.20

#define SHCNRF_InterruptLevel		0x0001
#define SHCNRF_ShellLevel			0x0002
#define SHCNRF_RecursiveInterrupt	0x1000
#define SHCNRF_NewDelivery			0x8000

/*
typedef struct {
	LPCITEMIDLIST pidl;
	BOOL fRecursive;
} SHChangeNotifyEntry;
*/

typedef struct {
    DWORD dwItem1;    // dwItem1 contains the previous PIDL or name of the folder. 
    DWORD dwItem2;    // dwItem2 contains the new PIDL or name of the folder. 
} SHNOTIFYSTRUCT;

typedef ULONG(WINAPI*LP_SHChangeNotifyRegister)(HWND, int, LONG, UINT, int, SHChangeNotifyEntry*);
typedef BOOL (WINAPI*LP_SHChangeNotifyDeregister)(ULONG);

static HINSTANCE m_hDllShell32 = NULL;
static LP_SHChangeNotifyRegister   m_pSHChangeNotifyRegister = NULL;
static LP_SHChangeNotifyDeregister m_pSHChangeNotifyDeregister = NULL;
static UINT m_ulIDSHChangeNotify = 0;

static void OpenSHChangeNotify()
{
	m_hDllShell32 = ::LoadLibrary(_T("SHELL32.DLL"));
	if(m_hDllShell32) {
		m_pSHChangeNotifyRegister = (LP_SHChangeNotifyRegister)::GetProcAddress(m_hDllShell32, MAKEINTRESOURCEA(2)); // ###_UNICODE
		m_pSHChangeNotifyDeregister = (LP_SHChangeNotifyDeregister)::GetProcAddress(m_hDllShell32, MAKEINTRESOURCEA(4));
	}
}

static void CloseSHChangeNotify()
{
	if(m_ulIDSHChangeNotify)
		VERIFY(m_pSHChangeNotifyDeregister(m_ulIDSHChangeNotify));

	if(m_hDllShell32) 
		::FreeLibrary(m_hDllShell32);
}

static void SetSHChangeNotify(HWND hwnd)
{
	SHChangeNotifyEntry shCNE;

	LPITEMIDLIST pIDL = NULL;
	if(::SHGetSpecialFolderLocation(AfxGetMainWnd()->m_hWnd, CSIDL_DESKTOP, &pIDL) == NOERROR) {
		shCNE.pidl = pIDL;
		shCNE.fRecursive = TRUE;
		if(m_pSHChangeNotifyRegister) {
			if(m_ulIDSHChangeNotify)
				m_pSHChangeNotifyDeregister(m_ulIDSHChangeNotify);

			m_ulIDSHChangeNotify = 
				m_pSHChangeNotifyRegister(hwnd, SHCNRF_ShellLevel, SHCNE_MKDIR | SHCNE_RMDIR | SHCNE_RENAMEFOLDER
					// | SHCNE_UPDATEDIR // ###1.23
					| SHCNE_DRIVEREMOVED | SHCNE_DRIVEADD	// ###1.23
					// | SHCNE_ALLEVENTS 
					, WM_SHCHANGENOTIFY, 1, &shCNE);
			ASSERT(m_ulIDSHChangeNotify);
		}
		g_pMalloc->Free(pIDL);
	}
}

LRESULT CZBTreeView::OnShellChangeNotify(WPARAM wParam, LPARAM lParam)
{
	SHNOTIFYSTRUCT *shns = (SHNOTIFYSTRUCT*)wParam;
	LPITEMIDLIST pIDL1 = (LPITEMIDLIST)shns->dwItem1;
	LPITEMIDLIST pIDL2 = (LPITEMIDLIST)shns->dwItem2;

	static LPITEMIDLIST pIDLInsert = NULL;

	HTREEITEM hItem = FindTreeItem(pIDL1, lParam == SHCNE_MKDIR);

	CString sPath = GetPathFromPIDL(pIDL1);
	TRACE(_T("Notify=%X, (%X, %X), hItem=%X, '%s'\n"), lParam, pIDL1, pIDL2, hItem, sPath);

	if(lParam == SHCNE_DRIVEADD) {
		InsertItem(pIDL1, g_pSFRoot, m_hMyComputer, NULL);
		return 0;
	}

	if(hItem) {
		CTreeCtrl& tree = GetTreeCtrl();
		switch(lParam) {
		case SHCNE_MKDIR:
			InsertFolder(sPath, hItem);
			//PostMessage(WM_REFRESH_TREE, (WPARAM)hItem);
			break;
		case SHCNE_RMDIR:
		case SHCNE_DRIVEREMOVED:
			DeleteItem(hItem);
			break;
		case SHCNE_RENAMEFOLDER: {
			CString sNewText = GetNameOfPIDL(pIDL2);
			tree.SetItemText(hItem, sNewText);
			LPITEMIDLIST pIDLNew = DuplicatePIDL(pIDL2);
			g_pMalloc->Free((LPITEMIDLIST)tree.GetItemData(hItem));
			tree.SetItemData(hItem, (DWORD)pIDLNew);
			if(hItem == m_hSelected)
				FillListView(hItem);	// ###1.23
			break;
			}
		}
	}
/*	TCHAR buffer[MAX_PATH];
	::SHGetPathFromIDList((LPCITEMIDLIST)shns->dwItem1, buffer);
	TRACE(_T("dwItem1=%X, dwItem2=%X, hParent=%X, %s\n"), shns->dwItem1, shns->dwItem2, hParent, buffer);
*/	return 0;
}

HTREEITEM CZBTreeView::FindTreeItem(LPITEMIDLIST pIDL, BOOL bParent)
{
	CTreeCtrl& tree = GetTreeCtrl();

	int nLength = GetPIDLLength(pIDL);
	if(nLength <= 1)  return NULL;	// ###1.23
	HTREEITEM hItem = NULL;
	HTREEITEM hParent = m_hMyComputer;

	for(int nIndex=1; nIndex < nLength; nIndex++) {
		LPITEMIDLIST pIDLwalk = GetSubPIDL(pIDL, nIndex);
		if(!pIDLwalk) break;
		CTreeCtrl& tree = GetTreeCtrl();
		hItem = tree.GetChildItem(hParent);
		while(hItem) {
			LPITEMIDLIST pIDLtree = (LPITEMIDLIST)tree.GetItemData(hItem);
			if(!ComparePIDs(g_pSFRoot, pIDLtree, pIDLwalk)) break;
			hItem = tree.GetNextSiblingItem(hItem);
		}
		g_pMalloc->Free(pIDLwalk);
		if(!hItem) break;
		hParent = hItem;
	}
	if(nIndex != nLength - bParent) return NULL;
	TRACE(_T("Found Tree:%s\n"), tree.GetItemText(hParent));
	return hParent;
}


CString CZBTreeView::GetNameOfPIDL(LPITEMIDLIST pIDLAbs) // ###1.23
{
//	SHFILEINFO sfi;
//	::SHGetFileInfo((LPCSTR)pIDL1, 0, &sfi, sizeof(SHFILEINFO), SHGFI_DISPLAYNAME | SHGFI_PIDL);
	CString sName = GetNameOf(g_pSFRoot, (LPITEMIDLIST)GetLastPIDL(pIDLAbs), SHGDN_NORMAL);
	if(sName.IsEmpty())
		sName = GetNameOf(g_pSFRoot, (LPITEMIDLIST)pIDLAbs, SHGDN_NORMAL);
	return sName;
}
