// ZBTreeVw.h : interface of the CZBTreeView class
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_ZBTREEVW_H__7A179302_9BCF_11D2_AE75_00400533897F__INCLUDED_)
#define AFX_ZBTREEVW_H__7A179302_9BCF_11D2_AE75_00400533897F__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include <afxcview.h>
#include <afxole.h>
#include <shlobj.h>
#include "filename.h"

class CZBrowseDoc;
class CZBListView;
class CFTPFindFiles;

#define FTPICON_MAX 4

class CZBTreeView : public CTreeView
{
protected:
	CZBTreeView();           // protected constructor used by dynamic creation
	DECLARE_DYNCREATE(CZBTreeView)

// Attributes
private:
	HTREEITEM m_hMyComputer;
	HTREEITEM m_hNetwork;
	HTREEITEM m_hFTPServer;
	HTREEITEM m_hSelected;
	HTREEITEM m_hTarget;
	HTREEITEM m_hDropItem;
	BOOL m_bBusy;
	HCURSOR m_hCsrBusy;
	CString m_sFindPath;
//	CString m_sFindArchive;		// ###1.24
	BOOL m_bFindRetry;
	BOOL m_bDeleting;
	DWORD m_dwAttrSelected;
	int m_nScroll;
	CFileName m_sDropSrc;
	COleDropTarget m_dropTarget;
	LPITEMIDLIST m_pIDLNet;
	int m_iNetWalk;
	LPITEMIDLIST m_pIDLDrive;	// ###1.23

// Operations
public:
	void SelectFrom(LPCTSTR pPath) { FindPath(pPath, TRUE); }
	void FolderUp();
	void Refresh();
	void RefreshTree(HTREEITEM hItem);
#ifndef NO_FTP
	void RefreshFTP();
	void FillFTPItems(CFTPFindFiles* pFinder);
	HTREEITEM InsertFTPItem(LPCTSTR lpszFolder, HTREEITEM _hParent = NULL);
#endif
	BOOL IsTreeComress();

// Implementation
protected:
	void FillItems(LPSHELLFOLDER lpSF, HTREEITEM hParent, LPITEMIDLIST pIDLParent = NULL);
	void FillFTPItems(HTREEITEM _hParent);
	HTREEITEM InsertItem(LPITEMIDLIST pIDL, LPSHELLFOLDER lpSF, HTREEITEM hParent, LPITEMIDLIST pIDLParent = NULL);
	HTREEITEM InsertItem(LPCTSTR pPath, HTREEITEM hParent, BOOL bFile, HTREEITEM hInsertAfter);
	HTREEITEM InsertFolder(LPCTSTR pPath, HTREEITEM hParent, HTREEITEM hInsertAfter = TVI_SORT) {
		return InsertItem(pPath, hParent, FALSE, hInsertAfter); }
	HTREEITEM InsertFile(LPCTSTR pPath, HTREEITEM hParent, HTREEITEM hInsertAfter = TVI_LAST) {
		return InsertItem(pPath, hParent, TRUE, hInsertAfter); }
	HTREEITEM InsertRootItem(int nFolder);
	void InsertRootItems();
#ifndef NO_FTP
	HTREEITEM InsertFTP();
#endif
	void DeleteItem(HTREEITEM hItem);	// ###1.20
	CZBListView* GetListView() { return ((CZBListView*)GetNextWindow()); }
	CString GetURL(HTREEITEM hItem);
	CString GetPath(HTREEITEM hItem);
	void SetChildren(HTREEITEM hItem, int cChildren);
	void FillListView(HTREEITEM hItem);
	BOOL IsExistItem(HTREEITEM hItem);
	BOOL IsAncestor(HTREEITEM hParent, HTREEITEM hChild);
	CString GetNameOfPIDL(LPITEMIDLIST pIDLAbs); // ###1.23
	HTREEITEM FindArchive(); // ###1.23

private:
	static UINT FillItemsThread(LPVOID pParam);
#ifndef NO_FTP
	static UINT FillFTPItemsThread(LPVOID pParam);
#endif
	static UINT FillArchivesThread(LPVOID pParam);	// ###1.24
	static int CALLBACK CompareProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);
	HTREEITEM FindPath(LPCTSTR pPath, BOOL bExpand);
	void FindPath(HTREEITEM hParent);
	void FindPathNext(HTREEITEM hParent);
	HTREEITEM FindPath(HTREEITEM hParent, BOOL bExpand);
	HGLOBAL CreateHDrop(LPCTSTR pPath);
	void EndDrag();
	void DoContextMenu(HTREEITEM hTree, POINT pt);
	HTREEITEM FindTreeItem(LPITEMIDLIST pIDL, BOOL bParent);

protected:
	virtual ~CZBTreeView();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CZBTreeView)
	public:
	virtual void OnInitialUpdate();
	virtual DROPEFFECT OnDragEnter(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point);
	virtual DROPEFFECT OnDragOver(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point);
	virtual void OnDragLeave();
	virtual BOOL OnDrop(COleDataObject* pDataObject, DROPEFFECT dropEffect, CPoint point);
	protected:
	virtual void OnDraw(CDC* pDC);      // overridden to draw this view
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
	//}}AFX_VIRTUAL

	// Generated message map functions
protected:
	//{{AFX_MSG(CZBTreeView)
	afx_msg void OnItemExpanding(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnDeleteItem(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnSelChanged(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg BOOL OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message);
	afx_msg void OnFileRemove();
	afx_msg void OnUpdateFileRemove(CCmdUI* pCmdUI);
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg void OnTimer(UINT nIDEvent);
	afx_msg void OnBeginDrag(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnKeyDown(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnDestroy();
	afx_msg int OnMouseActivate(CWnd* pDesktopWnd, UINT nHitTest, UINT message);
	afx_msg void OnBeginLabelEdit(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnEndLabelEdit(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnRClick(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnFileRename();
	afx_msg void OnUpdateFileRename(CCmdUI* pCmdUI);
	afx_msg void OnFileContextMenu();
	afx_msg void OnDblclk(NMHDR* pNMHDR, LRESULT* pResult);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
	afx_msg LRESULT OnFillComplete(WPARAM hItem, LPARAM);
	afx_msg LRESULT OnRefreshTree(WPARAM hItem, LPARAM);
	afx_msg LRESULT OnFileOpComplete(WPARAM bDone, LPARAM);
	afx_msg LRESULT OnShellChangeNotify(WPARAM wParam, LPARAM lParam);
};

extern LPSHELLFOLDER g_pSFRoot;

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_ZBTREEVW_H__7A179302_9BCF_11D2_AE75_00400533897F__INCLUDED_)
