/////////////////////////////////////////////////////////////////////////////
// ZCFind.cpp

#include "stdafx.h"
#include "ZCopy.h"
#include "ZCInfo.h"
#include "LogDlg.h"
#include "ZCFind.h"
#include "ErrorDlg.h"
#include "base.h"
#ifndef _UNICODE
#include <mbstring.h>
#endif

#define MASK_INCMASK	0x01
#define MASK_EXCMASK	0x02
#define MASK_INCDIR		0x04
#define MASK_EXCDIR		0x08
#define MASK_INCPATH	0x10
#define MASK_EXCPATH	0x20
#define MASK_EXCROOT	0x80

#define MASK_EACHFILES	0x0F

#define MOVE_INDRIVE 1
#define MOVE_TREE    2

#define OPT_TOTALDISK		(OPT_SUBDIR)
#define OPT_TOTALDISK_MASK	(OPT_INCLUDE | OPT_SUBDIR | OPT_NOOVW | OPT_OVW | OPT_BYTARGET | OPT_DAYAFTER)

// #define FO_COMPARE 8		###1.20

LONG CompareFileTime2(CONST FILETIME *lpFileTime1, CONST FILETIME *lpFileTime2); // ###1.05

CZCFind::CZCFind(CLogDlg* pLogDlg)
{
	m_pLogDlg = pLogDlg;
	m_pInfo = pLogDlg->m_pInfo;
	m_bExec = pLogDlg->m_bExec;
	m_bDummy = FALSE;
//	m_pBuffer2 = NULL;
}

CZCFind::~CZCFind()
{
//	if(m_pBuffer2)
//		MemFree(m_pBuffer2);
}

BOOL CZCFind::Init()
{
	m_sOpMode = sOpMode[m_pInfo->m_opmode];
	m_sSrcPath = ParseEnv(m_pInfo->m_sSrcPath);
	m_sDstPath = ParseEnv(m_pInfo->m_sDstPath);
#ifndef NO_FTP
	if(m_pInfo->m_option & OPT_SRCFTP) {
		m_sSrcPath = ftp.GetRoot(ParseEnv(m_pInfo->m_sSrcPath));
		m_fileSrc.m_hFTP = ftp.m_hFTP;
	}
	if(m_pInfo->m_option & OPT_DSTFTP) {
		m_sDstPath = ftp.GetRoot(ParseEnv(m_pInfo->m_sDstPath));
		m_fileDst.m_hFTP = ftp.m_hFTP;
	}
#endif
	m_bDest = m_pInfo->m_opmode != OPE_DELETE && m_pInfo->m_opmode != OPE_FIND;
	m_maskstate = 0;
	if(m_pInfo->m_option & OPT_INCLUDE)
		ParseMask(m_pInfo->m_sIncMask, TRUE);
	if(m_pInfo->m_option & OPT_EXCLUDE)
		ParseMask(m_pInfo->m_sExcMask, FALSE);
	m_movemode = (m_pInfo->m_opmode == OPE_MOVE && !(m_pInfo->m_option & OPT_FTP)
		&& m_sSrcPath.GetLength() >= 2 && m_sDstPath.GetLength() >= 2
		&& _tcsncicmp(m_sSrcPath, m_sDstPath, 2) == 0
		&& m_sSrcPath[1] == TCHAR(':')) ? MOVE_INDRIVE : 0;
	if(m_bExec && m_movemode && !(m_maskstate & MASK_EACHFILES)) m_movemode |= MOVE_TREE;
	if(m_bExec && m_pInfo->m_opmode == OPE_DELETE && m_pInfo->m_option & OPT_SUBDIR
		&& AfxGetMainWnd()->IsWindowVisible()) {
		CString sMsg;
		sMsg.Format(IDS_ASK_DELETE, m_sSrcPath);
		if(m_pLogDlg->MessageBox(sMsg, sOpMode[m_pInfo->m_opmode], MB_OKCANCEL) != IDOK)
			return FALSE;
	}
	if(m_pInfo->m_option & OPT_BYTARGET) {
FindTarget:
		if(!m_finderDest.FindAll(m_sDstPath, m_fileDst.m_hFTP, FIND_DIR, &m_pLogDlg->m_bCanceled)) {
			if(ErrorMessageBox(m_sDstPath) == IDRETRY) goto FindTarget;
			return FALSE;
		}
		for_to(i, m_finderDest.GetCount()) {
			m_Paths.sInc.Add(m_finderDest[i].GetFileName());
		}
		m_maskstate |= MASK_INCPATH + MASK_EXCROOT;
		m_finderDest.RemoveAll();
	}

	m_bValidTotal = (m_pInfo->m_opmode != OPE_DELETE && m_pInfo->m_opmode != OPE_FIND && !(m_movemode & MOVE_INDRIVE));
	m_bDummy = FALSE;
	m_bDeleteOnSync = FALSE;

	m_nFileCount = 0;
	m_bLogOut = m_pInfo->m_opmode == OPE_UPDATE || m_pInfo->m_opmode == OPE_SYNC || m_pInfo->m_opmode == OPE_FIND
		|| m_maskstate & MASK_INCMASK || m_pInfo->m_option & (OPT_FTP + OPT_OVW + OPT_NOOVW + OPT_DAYAFTER + OPT_BYSIZE);
	InitFileOp();
	if(m_pInfo->m_option & OPT_IGNORE_ERROR) m_bIgnoreError = TRUE;	// ###0.93

	m_dwlBySize = m_pInfo->m_dwBySize;	// ###1.23
	if(m_pInfo->m_modeBySize & OP_SIZE_KB)
		m_dwlBySize <<= 10;
	else if(m_pInfo->m_modeBySize & OP_SIZE_MB)
		m_dwlBySize <<= 20;
	return TRUE;
}

void CZCFind::CalcTotalBytes()	// ###1.20
{
	if(!(m_pInfo->m_option & OPT_SUBDIR)) return;	// ###1.22
	if(m_pInfo->m_opmode == OPE_COPY && ((m_pInfo->m_option & OPT_TOTALDISK_MASK) == OPT_TOTALDISK) && InitTotalDiskBytes(m_sSrcPath)) return;
	if(options.m_bNoCalcTotal) return;
	m_pLogDlg->SetInitializeMode(TRUE);
	m_bDummy = TRUE;
	ScanDir();
	m_pLogDlg->SetInitializeMode(FALSE);
	m_bDummy = FALSE;
	if(!IsCanceled())
		SetTotalParameters();
	else {
		InitFileOp();
		m_nFileCount = 0;
	}
}

void CZCFind::DrawStatus(DWORDLONG dwlDoneBytes)
{
	if(m_bDummy || !m_bValidTotal) return;
	if(m_dwlTotalBytes) {
		m_pLogDlg->m_dwlTotalBytes = m_dwlTotalBytes;
		m_pLogDlg->m_dwlDoneBytes = (m_dwBytesPerCluster ? m_dwlDoneDiskBytes : m_dwlDoneBytes) + dwlDoneBytes;
		m_pLogDlg->DrawStatus(m_dwlTotalBytes, m_pLogDlg->m_dwlDoneBytes, m_nTotalFiles, m_nDoneFiles);
	}
	else {
		if(!(m_pInfo->m_option & OPT_SUBDIR)) {		// ###1.22
			m_pLogDlg->m_dwlTotalBytes = m_dwlSubTotalBytes;
			m_pLogDlg->m_dwlDoneBytes = m_dwlSubDoneBytes + dwlDoneBytes;
		}
		m_pLogDlg->DrawStatus(m_dwlSubTotalBytes, m_dwlSubDoneBytes + dwlDoneBytes, m_nSubTotlaFiles, m_nSubDoneFiles);
	}
}

void CZCFind::ScanDir()	// ###1.23
{
	if(m_maskstate & MASK_INCPATH) {
		if(!(m_maskstate & MASK_EXCROOT)) {
			UINT option = m_pInfo->m_option;
			m_pInfo->m_option &= ~OPT_SUBDIR;
			ScanDir(NULL);
			m_pInfo->m_option = option;
			if(IsCanceled()) return;
		}
		for_to(i, m_Paths.sInc.GetSize()) {
			ScanDir(m_Paths.sInc[i]);
			if(IsCanceled()) return;
		}
	} else
		ScanDir(NULL);
}

void CZCFind::ScanDir(LPCTSTR pDir)
{
	BOOL bScanDest;
	CStringArray sSubdirs, sSubdirs2;
	BOOL bOutPath = FALSE;
	CFindData* pDstHead = NULL;
	CFindData* pDstNext = NULL;
	TRACE(_T("%s\n"), pDir);

	CFileName sSrcPath = m_fileSrc.AddPath((CFileName)m_sSrcPath, pDir);
	CFileName sDstPath;
	if(m_bDest)
		sDstPath = (m_pInfo->m_opmode == OPE_GATHER) ? (CFileName)m_sDstPath : m_fileDst.AddPath((CFileName)m_sDstPath, pDir);
	if(bScanDest = m_bDest) {
		if(m_pInfo->m_opmode == OPE_GATHER 
			|| ((m_pInfo->m_opmode == OPE_COPY || m_pInfo->m_opmode == OPE_MOVE)
				&& !(m_pInfo->m_option & (OPT_NOOVW + OPT_OVW)))) bScanDest = FALSE;
	}
	m_pLogDlg->DrawPath(sSrcPath, sDstPath);
FindSrc:
	if(!FindAll(sSrcPath, m_fileSrc.m_hFTP, m_pInfo->IsSubDir() ? FIND_DIRFILE : FIND_FILE, &m_pLogDlg->m_bCanceled)) {
		if(ErrorMessageBox(sSrcPath) == IDRETRY) goto FindSrc;
		m_bError = TRUE;
		Cancel();
	}
	if(IsCanceled()) return;
	if(bScanDest) {
FindDst:
		UINT findtype = (m_pInfo->m_opmode == OPE_SYNC || m_pInfo->m_opmode == OPE_COMPARE)  ? FIND_DIRFILE : FIND_FILE;
		if(!m_finderDest.FindAll(sDstPath, m_fileDst.m_hFTP, findtype, &m_pLogDlg->m_bCanceled)) {
			int nErr = LOWORD(::GetLastError());
			if(nErr != ERROR_PATH_NOT_FOUND && nErr != ERROR_INTERNET_EXTENDED_ERROR) {		// ###1.23
				if(ErrorMessageBox(sDstPath) == IDRETRY) goto FindDst;
				m_bError = TRUE;
				Cancel();
				return;
			}
			if(m_pInfo->m_opmode == OPE_COMPARE) return;
			bScanDest = FALSE;
		} else {
			if(IsCanceled()) return;
			for_to(i, GetCount()) {
				CFindData& src = ElementAt(i);
				if(src.IsDirectory() && m_pInfo->m_opmode != OPE_COMPARE && m_pInfo->m_opmode != OPE_SYNC) continue;	// ###0.92
				for_to(j, m_finderDest.GetCount()) {
					CFindData& dest = m_finderDest[j];
					if(dest.IsMarked()) continue;
					if(_tcsicmp(src.GetFileName(), dest.GetFileName()) == 0) {
						if(src.IsDirectory()) {		// ###0.92
							if(m_pInfo->m_opmode == OPE_COMPARE) src.SetMark();	 // ###0.93
						} else {	// ### 040619
							if(!(m_pInfo->m_option & OPT_NOOVW)) {
								if(m_pInfo->m_opmode == OPE_UPDATE || m_pInfo->m_opmode == OPE_SYNC) {
									if(CompareFileTime2(&src.ft.ftLastWrite, &dest.ft.ftLastWrite) > 0)
										src.SetMark();
								} else if(m_pInfo->m_opmode == OPE_COMPARE || m_pInfo->m_option & OPT_OVW)
									src.SetMark();
							}
							if(m_pInfo->m_opmode == OPE_COMPARE && src.IsMarked()) {	// ###1.20
								if(!pDstHead) pDstHead = &dest;
								else pDstNext->pNext = &dest;
								pDstNext = &dest;
								dest.pNext = NULL;
							}
						}
						dest.SetMark();
						break;
					}
				}
				if(j == m_finderDest.GetCount()) {
					if(!(m_pInfo->m_option & OPT_OVW) && (m_pInfo->m_option & OPT_NOOVW || m_pInfo->m_opmode == OPE_UPDATE || m_pInfo->m_opmode == OPE_SYNC))
						src.SetMark();
				}
				if(m_pInfo->m_opmode == OPE_SYNC && !IsIncExc(src.GetFileName(), m_Masks, MASK_INCMASK))  // ###0.92
					src.ClearMark();
			}
		}
	}
	InitSubCounter();
	BOOL bExcRoot = (pDir == NULL) && (m_maskstate & MASK_EXCROOT);
	for_to(i, GetCount()) {
		CFindData& src = ElementAt(i);
		if(src.IsDirectory()) {
			if(m_pInfo->m_opmode != OPE_COMPARE || src.IsMarked())
				sSubdirs.Add(src.GetFileName());
		} else if(!bExcRoot && (!bScanDest || bScanDest && src.IsMarked())) {
			m_dwlSubTotalBytes += src.dwlSize;
			m_nSubTotlaFiles ++;
		}
	}
	if(m_nSubTotlaFiles) {
		CFindData* pDst = pDstHead;
		for_to_(i, GetCount()) {
			CFindData& src = ElementAt(i);
			if(!src.IsDirectory() && (!bExcRoot && (!bScanDest || bScanDest && src.IsMarked()))) {
#ifndef NO_FTP
				m_transfer = ftp.m_transfer ? ftp.m_transfer : GetFileType(src.GetFileName());
#else
				m_transfer = FTP_TRANSFER_TYPE_BINARY;
#endif
				BOOL bOperate = TRUE;
				if(!m_bDummy) {
					m_sLogMsg.Empty();
					if(m_bExec && m_pInfo->m_opmode != OPE_FIND) {
						m_pLogDlg->DrawFileName(src.GetFileName());
						DrawStatus();
						bOperate = Operate(m_pInfo->m_opmode, sSrcPath, sDstPath, &src, pDst);
					}
					if(IsCanceled()) return;
					if(m_bLogOut || !bOperate || !m_sLogMsg.IsEmpty()) {
						CString sType;
						if(m_pInfo->m_option & OPT_FTP)
							sType.Format(_T(" (%c) "), TCHAR('A') + m_transfer - 1);
						if(!bOutPath) {
							AddLog(sSrcPath);
							bOutPath = TRUE;
						}
						if(!bOperate)
							m_sLogMsg.Format(IDS_ERRORCODE, m_nErrCode);
						AddLog(CString(_T("  ")) + src.GetFileName() + sType + m_sLogMsg);
					}
				}
				if(bOperate) {
					if(m_bDeleteOnSync)	// ###1.22
						m_nFileCount++;
					else {
						m_dwlSubDoneBytes += src.dwlSize;
						m_nSubDoneFiles++;
						m_dwlDoneBytes += src.dwlSize;
						if(m_dwBytesPerCluster) // ###1.24
							m_dwlDoneDiskBytes += ((src.dwlSize + (m_dwBytesPerCluster - 1)) & ~(m_dwBytesPerCluster-1));
						m_nDoneFiles++;
					}
					if(m_bExec && m_pInfo->m_opmode != OPE_FIND)
						DrawStatus();
				}
				if(pDst) pDst = pDst->pNext;
			}
		}	
	}
	if(m_pInfo->m_opmode == OPE_SYNC) {
		for_to(i, m_finderDest.GetCount()) {
			CFindData& dest = m_finderDest[i];
			if(!dest.IsMarked()) {
				if(m_bDummy) {
					m_nFileCount++;		// ###1.22
				} else {
					if(!bOutPath) {
						AddLog(sDstPath);
						bOutPath = TRUE;
					}
					if(dest.IsDirectory()) {	// ###0.92
						if(m_bExec && m_pInfo->IsSubDir()) {
							sSubdirs2.Add(dest.GetFileName());
						}
					} else {
						if(!m_bExec || Operate(OPE_DELETE, NULL, sDstPath, &dest))
							AddLog(CString(_T("  ")) + dest.GetFileName() + " ... Removed");
						m_nFileCount++;
					}
				}
			}
		}
	}	
	if(!m_bDummy) {	// ### 1.23
		if(m_pInfo->m_opmode == OPE_COMPARE) {	// ###0.93
			if(options.m_fCompare & OP_CMP_IN_SRC) {
				for_to(i, GetCount()) {
					CFindData& src = ElementAt(i);
					if(!src.IsMarked()) {
						if(!bOutPath) {
							AddLog(sSrcPath);
							bOutPath = TRUE;
						}
						CString sHead = _T("  < Source > ");
						if(src.IsDirectory()) sHead += TCHAR('\\');
						AddLog(sHead + src.GetFileName());
					}
				}
			}
			if(options.m_fCompare & OP_CMP_IN_DST) {
				for_to_(i, m_finderDest.GetCount()) {
					CFindData& dest = m_finderDest[i];
					if(!dest.IsMarked()) {
						if(!bOutPath) {
							AddLog(sDstPath);
							bOutPath = TRUE;
						}
						CString sHead = _T("  < Target > ");
						if(dest.IsDirectory()) sHead += TCHAR('\\');
						AddLog(sHead + dest.GetFileName());
					}
				}
			}
		}
	}
	RemoveAll();
	m_finderDest.RemoveAll();
	for_to_(i, sSubdirs.GetSize()) {
		if(!(IsIncExc((CFileName&)sSubdirs[i], m_Dirs, MASK_INCDIR))) continue;
		CFileName sDir;
		if(pDir) {
			sDir += pDir;
			sDir.AddSlash();
		}
		sDir += sSubdirs[i];
		if(!(IsIncExc(sDir, m_Paths, MASK_INCPATH))) continue;
		if(m_bDummy) {
			ScanDir(sDir);
		} else {
			CFileName sSrcDir = m_fileSrc.AddPath((CFileName)m_sSrcPath, sDir);
			CFileName sDstDir = m_fileDst.AddPath((CFileName)m_sDstPath, sDir);
			if(m_movemode & MOVE_TREE) {
				if(!Operate(OPE_MOVE, sSrcDir, sDstDir)) {
					m_movemode &= ~MOVE_TREE;
				} else
					m_nFileCount++;
			}
			if(!(m_movemode & MOVE_TREE)) {
				if(m_bExec && (m_pInfo->m_opmode == OPE_COPY || m_pInfo->m_opmode == OPE_SYNC)) // ###1.03
					m_fileDst.CreateDirectory(sDstDir);
				ScanDir(sDir);
				if(IsCanceled()) return;
				if(m_bExec && !(m_maskstate & MASK_INCMASK+MASK_EXCMASK)
				   && (m_pInfo->m_opmode == OPE_MOVE || m_pInfo->m_opmode == OPE_DELETE))
					Operate(OPE_DELETE, sSrcDir);
			}
		}
		if(IsCanceled()) return;
	}
	for_to_(i, sSubdirs2.GetSize()) {			// ###0.92
		CZCInfo info;
		info.m_opmode = OPE_DELETE;
		info.m_option = m_pInfo->m_option;
		CFileName sDir;
		if(pDir) {
			sDir += pDir;
			sDir.AddSlash();
		}
		sDir += sSubdirs2[i];
		CZCInfo* pInfo0 = m_pInfo;
		m_pInfo = &info;
		m_bDest = FALSE;
		BOOL bLogOut0 = m_bLogOut;
		m_bLogOut = FALSE;
		UINT maskstate0 = m_maskstate;
		m_maskstate = 0;
		int nDoneFiles = m_nDoneFiles;	// ###1.22
		m_bDeleteOnSync = TRUE;
		Swap(m_sSrcPath, m_sDstPath);
		Swap(m_fileSrc, m_fileDst);
		ScanDir(sDir);
		m_bDest = TRUE;
		Swap(m_sSrcPath, m_sDstPath);
		Swap(m_fileSrc, m_fileDst);
		m_pInfo = pInfo0;
		m_bLogOut = bLogOut0;
		m_maskstate = maskstate0;
		m_nDoneFiles = nDoneFiles;
		m_bDeleteOnSync = FALSE;
		CFileName sDstDir = m_fileDst.AddPath((CFileName)m_sDstPath, sDir);
		if(Operate(OPE_DELETE, sDstDir))
			AddLog(CString(_T("  ")) + sDstDir + " ... Removed");
	}
}

BOOL CZCFind::Done()
{
	if(!m_bError && !m_pLogDlg->m_bLogOpen && (IsCanceled() && m_nDoneFiles == 0
	 /* || !AfxGetMainWnd()->IsWindowVisible()*/)) return FALSE;
	CString sDstPath;
	if(m_bDest) sDstPath = m_sDstPath;
	m_pLogDlg->DrawPath(m_sSrcPath, sDstPath);

	CString sTotal;
	DWORD dwKBytes = (DWORD)(m_dwlDoneBytes / 1024);
	if(dwKBytes == 0) dwKBytes = m_dwlDoneBytes ? 1 : 0;
	CString sKBytes = SeparateByComma(dwKBytes);

	if(IsCanceled())
		sTotal.LoadString(IDS_CANCEL);
	else if(m_pInfo->m_opmode == OPE_COMPARE) {
		if(m_nFileCount)
			sTotal.Format(IDS_RESULT_MISMATCH, m_nDoneFiles, m_nFileCount);
		else
			sTotal.Format(IDS_RESULT_MATCH, m_nDoneFiles);
	} else if(m_movemode & MOVE_TREE && m_nFileCount)
		sTotal.Format(IDS_RESULT_MOVEDIR, m_nDoneFiles, m_nFileCount);
	else if(m_pInfo->m_opmode == OPE_SYNC && m_nFileCount)
		sTotal.Format(IDS_RESULT_SYNC, m_nDoneFiles, sKBytes, m_nFileCount);
	else
		sTotal.Format(IDS_RESULT, m_nDoneFiles, sKBytes);
	m_pLogDlg->DrawFileName(sTotal);
	AddLog(sTotal);
	if(m_bError) {
		CString sError;
		sError.Format(IDS_ERRORCODE, m_nErrCode);
		sTotal += sError;
	}
	if(m_bExec) {
		m_pInfo->SetReport(sTotal);
		if(!IsCanceled() && m_pInfo->m_opmode != OPE_FIND)	// ###1.21
			m_pInfo->SetDate();
		if(options.m_bSaveLog)
			m_pLogDlg->SaveLog();
		if(m_bExec != DO_AUTOEXEC)
			MessageBeep(IsCanceled() ? MB_NOFOCUS : MB_ICONASTERISK);
	}
	return TRUE;
}

BOOL CZCFind::Operate(OpMode opmode, LPCTSTR pSrcDir, LPCTSTR pDstDir, CFindData* pSrcData, CFindData* pDstData)
{
	UINT wFunc = 0;
	switch(opmode) {
	case OPE_COPY:
	case OPE_UPDATE:
	case OPE_SYNC:
	case OPE_GATHER:
		wFunc = FO_COPY;
		break;
	case OPE_MOVE:
		if(m_movemode & MOVE_INDRIVE)
			wFunc = FO_RENAME;
		else
			wFunc = FO_MOVE;
		break;
	case OPE_DELETE:
		wFunc = FO_DELETE;
		break;
	case OPE_COMPARE:
		wFunc = FO_COMPARE;
		break;
	}
	BOOL bResult = FileOperate(wFunc, pSrcDir, pDstDir, pSrcData, pDstData);
	if(m_bError && m_bIgnoreError && !m_pLogDlg->m_bLogOpen && m_bExec != DO_AUTOEXEC)
		m_pLogDlg->OpenLog(TRUE);
	return bResult;
}

BOOL CZCFind::OperateDefault(UINT wFunc, LPCTSTR pSrcFile, LPCTSTR pDstFile, CFindData* pSrcData, CFindData* pDstData)
{
	if(wFunc == FO_COMPARE)
		return CompareFile(pSrcFile, pDstFile, pSrcData, pDstData);
	return CFTPFileOp::OperateDefault(wFunc, pSrcFile, pDstFile, pSrcData, pDstData);
}

void CZCFind::ParseMask(LPCTSTR pMask, BOOL bInc)
{
	for(;;) {
		CString sTerm = GetNextTerm(pMask);
		if(sTerm.IsEmpty()) break;
		CStringArray* psArr = NULL;
		UINT f = 0;
		if(sTerm[0] == _T('\\') || sTerm[0] == _T('/')) {
			sTerm = sTerm.Mid(1);
			if(bInc) { psArr = &m_Paths.sInc; f = MASK_INCPATH; }
			else     { psArr = &m_Paths.sExc; f = MASK_EXCPATH; }
		} else if(((CFileName&)sTerm).IsTailSlash()) {
			sTerm = sTerm.Left(sTerm.GetLength()-1);
			if(bInc) { psArr = &m_Dirs.sInc; f = MASK_INCDIR; }
			else     { psArr = &m_Dirs.sExc; f = MASK_EXCDIR; }
		} else {
			if(sTerm[0] == _T('.'))
				sTerm = CString(_T('*')) + sTerm;
			if(bInc) { psArr = &m_Masks.sInc; f = MASK_INCMASK; }
			else     { psArr = &m_Masks.sExc; f = MASK_EXCMASK; }
		}
		if(!sTerm.IsEmpty()) {
			psArr->Add(sTerm);
			m_maskstate |= f;
	#ifdef _DEBUG
			AddLog(sTerm);
	#endif
		} else if(f == MASK_EXCPATH)
			m_maskstate |= MASK_EXCROOT;
	}
}

BOOL CZCFind::IsIncExc(CFileName sPath, IncExcArray& arr, UINT mask)
{
	if(m_maskstate & (mask<<1)) {
		for_to(i, arr.sExc.GetSize()) {
			if(!sPath.CompareMask(arr.sExc[i], mask != MASK_INCMASK)) return FALSE;
		}
	}
	if(m_maskstate & mask) {
		if(mask == MASK_INCPATH) return TRUE;	// ###1.23
		for_to(i, arr.sInc.GetSize()) {
			if(!sPath.CompareMask(arr.sInc[i], mask != MASK_INCMASK)) return TRUE;
		}
		return FALSE;
	}
	return TRUE;
}

BOOL CZCFind::PostFind(CFindData* pData)
{
	// ###1.24 comment out
	// if(m_dwBytesPerCluster && !pData->IsDirectory()) // ###1.20
	//	m_dwlDoneDiskBytes += ((pData->dwlSize + (m_dwBytesPerCluster - 1)) & ~(m_dwBytesPerCluster-1));

	if(pData->IsHidden() && !(m_pInfo->m_option & OPT_HIDDEN)) return FALSE;
	if(pData->IsDirectory()) return TRUE;
	if(m_pInfo->m_option & OPT_DAYAFTER) {	// ###1.20
		if(CompareFileTime2(&pData->ft.ftLastWrite, &m_pInfo->m_ftOperate) <= 0) return FALSE;
	}
	if(m_pInfo->m_option & OPT_BYSIZE) {	// ###1.23
		if(m_pInfo->m_modeBySize & OP_SIZE_EQUAL) {
			if(pData->dwlSize != m_dwlBySize) return FALSE;
		} else if(m_pInfo->m_modeBySize & OP_SIZE_SMALL) {
			if(pData->dwlSize >= m_dwlBySize) return FALSE;
		} else {
			if(pData->dwlSize <= m_dwlBySize) return FALSE;
		}
	}
	CFileName sName(pData->GetFileName());
	if(m_pInfo->m_opmode == OPE_SYNC) return TRUE;	// ###0.92
	return(IsIncExc(sName, m_Masks, MASK_INCMASK));
}

enum CompResult { CMP_MATCH, CMP_SHORT, CMP_LONG, CMP_MISMATCH };
static const TCHAR *sCompMsg[] = { _T("Shorter"), _T("Longer"), _T("Mismatch") };

BOOL CZCFind::CompareFile(LPCTSTR pSrcFile, LPCTSTR pDstFile, CFindData* pSrcData, CFindData* pDstData)
{
	ASSERT(pDstData);
	if(!pDstData) return TRUE;
	BOOL bResult = TRUE;
	CompResult nMatch = CMP_MATCH;
	if(pSrcData->dwlSize < pDstData->dwlSize) nMatch = CMP_SHORT;
	else if(pSrcData->dwlSize > pDstData->dwlSize) nMatch = CMP_LONG;
	else if(pSrcData->dwlSize == 0 && pDstData->dwlSize == 0) nMatch = CMP_MATCH;	// ###0.93
	else if(options.m_fCompare & OP_CMP_MISMATCH) {
		int nResult = CFTPFileOp::CompareFile(pSrcFile, pDstFile);
		bResult = (nResult != 0);
		if(nResult == INVALID) nMatch = CMP_MISMATCH;
	}
	if(nMatch) {
		if(nMatch == CMP_MISMATCH || nMatch != CMP_MISMATCH && options.m_fCompare & OP_CMP_LENGTH) {
			m_pLogDlg->OpenLog(TRUE);
			m_sLogMsg = CString(_T("  <<< ")) + sCompMsg[nMatch - 1] + _T(" >>>");
			m_nFileCount++;
		}
	} else if(options.m_fCompare & OP_CMP_TIME) {
		int r = CompareFileTime2(&pSrcData->ft.ftLastWrite, &pDstData->ft.ftLastWrite);
		if(r) {
			m_pLogDlg->OpenLog(TRUE);
			m_sLogMsg = CString(r > 0 ? _T("  < newer >") : _T("  < older >"));
			m_nFileCount++;
		}
	}
	return bResult;
}

BOOL CZCFind::AllocBuffer()
{
	if(!m_pBuffer) {
		m_dwBuffer = ((m_pInfo->m_option & OPT_FTP) ? options.m_dwBufferFTP : options.m_dwBuffer) * 1024;
		m_pBuffer = MemAlloc(m_dwBuffer);
	}
	if(m_pInfo->m_opmode == OPE_COMPARE && !m_pBuffer2) {
		m_pBuffer2 = MemAlloc(m_dwBuffer);
	}
	return m_pBuffer != NULL;
}

/*
///// ###1.05 samba time compare ver. (revised code by Michiyasu Odaki 2003/3/16)

LONG CompareFileTime2(CONST FILETIME *lpFileTime1, CONST FILETIME *lpFileTime2)
{
	if(!options.m_uRoundFT)
		return ::CompareFileTime(lpFileTime1, lpFileTime2);
 ULARGE_INTEGER ult;
 FILETIME ft1;
 FILETIME ft2;

 ult.LowPart = lpFileTime1->dwLowDateTime & 0xFFF00000;
 ult.HighPart = lpFileTime1->dwHighDateTime;
 ult.QuadPart = ((ult.QuadPart+5000000i64) / 10000000i64) * 10000000i64;
 ft1.dwLowDateTime = ult.LowPart;
 ft1.dwHighDateTime = ult.HighPart;

 ult.LowPart = lpFileTime2->dwLowDateTime & 0xFFF00000;
 ult.HighPart = lpFileTime2->dwHighDateTime;
 ult.QuadPart = ((ult.QuadPart+5000000i64) / 10000000i64) * 10000000i64;
 ft2.dwLowDateTime = ult.LowPart;
 ft2.dwHighDateTime = ult.HighPart;

 return ::CompareFileTime(&ft1, &ft2);
}
*/

// ###050605 Qbۂ

#define FT_2SECOND (1000i64 * 1000 * 10 * 2)

LONG CompareFileTime2(const FILETIME *pFT1, const FILETIME *pFT2)
{
	if(!options.m_uRoundFT)
		return ::CompareFileTime(pFT1, pFT2);

	if(pFT1->dwLowDateTime == pFT2->dwLowDateTime && pFT1->dwHighDateTime == pFT2->dwHighDateTime) return 0;

	LONGLONG llDiff = *(LONGLONG*)pFT1 - *(LONGLONG*)pFT2;

	if(llDiff > 0)
		return (llDiff < FT_2SECOND) ? 0 : 1;
	return (llDiff > -FT_2SECOND) ? 0 : -1;
}
