/////////////////////////////////////////////////////////////////////////////
// ZCInfo.cpp

#include "stdafx.h"
#include "ZCopy.h"
#include "ZCInfo.h"

TCHAR *sOpMode[OPECOUNT] = { _T("Update"), _T("Copy"), _T("Sync"), _T("Move"), _T("Delete"), _T("Gather"), _T("Compare"), _T("Find") };

CZCInfo::CZCInfo()
{
	m_sLabel = _T("Untitled");
	m_opmode = OPE_UPDATE;
	m_option = OPT_DEFAULT;
	m_bReportUpdated = FALSE;
	m_ftOperate.dwLowDateTime = m_ftOperate.dwHighDateTime = 0;
	m_dwBySize = 0;
	m_modeBySize = 0;
}

CZCInfo::~CZCInfo()
{
}

CString CZCInfo::GetOpModeString()
{
	CString s(sOpMode[m_opmode]);
	return s;
}

CString CZCInfo::GetTitle()
{
	CString s = GetOpModeString() + _T(" ") + m_sLabel;
	return s;
}

#define BUFFER_MAX 128

CString GetDateTimeString(SYSTEMTIME *pST, DWORD dwTimeFlags)
{
	CString sDate;
	SYSTEMTIME st;
	if(pST == NULL) {
		::GetLocalTime(&st);
		pST = &st;
	}
	LPTSTR pBuff = sDate.GetBuffer(BUFFER_MAX);
    ::GetDateFormat(LOCALE_USER_DEFAULT, DATE_SHORTDATE, pST, NULL, pBuff, BUFFER_MAX);
	int len = _tcslen(pBuff);
	pBuff += len;
	*pBuff++ = TCHAR(' ');
    ::GetTimeFormat(LOCALE_USER_DEFAULT, dwTimeFlags, pST, NULL, pBuff, BUFFER_MAX - len - 1);
	sDate.ReleaseBuffer();
	return sDate;
}

void CZCInfo::SetReport(LPCTSTR msg)
{
	CString sDate = GetDateTimeString(NULL, TIME_NOSECONDS);
	m_sReport.Format(_T("%s  %s %s"), sDate, GetOpModeString(), msg);
	m_bReportUpdated = TRUE;
}

void CZCInfo::SetDate()  // ###1.20
{
	SYSTEMTIME st;
	::GetSystemTime(&st);
	::SystemTimeToFileTime(&st, &m_ftOperate);
	m_bReportUpdated = TRUE;
}

CZCInfoCtrl::CZCInfoCtrl()
{
	m_iSel = 0;
	m_bModified = FALSE;
}

CZCInfoCtrl::~CZCInfoCtrl()
{
	m_arrInfo.RemoveAll();	
}

BOOL CZCInfoCtrl::Load()
{
	m_bModified = FALSE;
	CWinApp* pApp = AfxGetApp();
	int nSheet = pApp->GetProfileInt(sRegDefault, _T("SheetCount"), 0);
	if(!nSheet) return FALSE;

	m_iSel = pApp->GetProfileInt(sRegDefault, _T("SheetActive"), 0);
	m_arrInfo.RemoveAll();	// ###1.21
	for_to(i, nSheet) {
		CZCInfo info;
		CString sRegSheet;
		sRegSheet.Format(_T("Sheet%d"), i+1);
		info.m_sLabel   = pApp->GetProfileString(sRegSheet, _T("Label"));
		info.m_sSrcPath = pApp->GetProfileString(sRegSheet, _T("SrcPath"));
		info.m_sDstPath = pApp->GetProfileString(sRegSheet, _T("DstPath"));
		info.m_sIncMask = pApp->GetProfileString(sRegSheet, _T("Include"));
		info.m_sExcMask = pApp->GetProfileString(sRegSheet, _T("Exclude"));
		info.m_opmode   = (OpMode)pApp->GetProfileInt(sRegSheet, _T("OpMode"), OPE_UPDATE); 
		info.m_option   = pApp->GetProfileInt(sRegSheet, _T("Options"), OPT_DEFAULT);
		info.m_sReport  = pApp->GetProfileString(sRegSheet, _T("Report"));
		info.m_ftOperate.dwLowDateTime  = pApp->GetProfileInt(sRegSheet, _T("DateLow"), 0); 
		info.m_ftOperate.dwHighDateTime = pApp->GetProfileInt(sRegSheet, _T("DateHigh"), 0); 
		info.m_dwBySize = pApp->GetProfileInt(sRegSheet, _T("BySize"), 0);
		info.m_modeBySize = pApp->GetProfileInt(sRegSheet, _T("BySizeMode"), 0);
		m_arrInfo.Add(info);
	}
	LoadHistory(CString(_T("Path")), m_sHistPath);
	LoadHistory(CString(_T("Mask")), m_sHistMask);
	LoadHistory(CString(_T("FTP")),  m_sHistFTP);
	return TRUE;
}

void CZCInfoCtrl::Save()
{
	if(!m_bModified) {
		SaveReport();
		return;
	}
	m_bModified = FALSE;
	CWinApp* pApp = AfxGetApp();
	int nSheet = GetInfoCount();
	pApp->WriteProfileInt(sRegDefault, _T("SheetCount"), nSheet);
	pApp->WriteProfileInt(sRegDefault, _T("SheetActive"), m_iSel);
	for_to(i, nSheet) {
		CString sRegSheet;
		sRegSheet.Format(_T("Sheet%d"), i+1);
		pApp->WriteProfileString(sRegSheet, _T("Label"), m_arrInfo[i].m_sLabel);
		pApp->WriteProfileString(sRegSheet, _T("SrcPath"), m_arrInfo[i].m_sSrcPath);
		pApp->WriteProfileString(sRegSheet, _T("DstPath"), m_arrInfo[i].m_sDstPath);
		pApp->WriteProfileString(sRegSheet, _T("Include"), m_arrInfo[i].m_sIncMask);
		pApp->WriteProfileString(sRegSheet, _T("Exclude"), m_arrInfo[i].m_sExcMask);
		pApp->WriteProfileInt(sRegSheet, _T("OpMode"), m_arrInfo[i].m_opmode); 
		pApp->WriteProfileInt(sRegSheet, _T("Options"), m_arrInfo[i].m_option); 
		pApp->WriteProfileString(sRegSheet, _T("Report"), m_arrInfo[i].m_sReport);
		pApp->WriteProfileInt(sRegSheet, _T("DateLow"), m_arrInfo[i].m_ftOperate.dwLowDateTime);
		pApp->WriteProfileInt(sRegSheet, _T("DateHigh"), m_arrInfo[i].m_ftOperate.dwHighDateTime);
		pApp->WriteProfileInt(sRegSheet, _T("BySize"), m_arrInfo[i].m_dwBySize);
		pApp->WriteProfileInt(sRegSheet, _T("BySizeMode"), m_arrInfo[i].m_modeBySize);
	}
	SaveHistory(CString(_T("Path")), m_sHistPath);
	SaveHistory(CString(_T("Mask")), m_sHistMask);
	SaveHistory(CString(_T("FTP")),  m_sHistFTP);
}

void CZCInfoCtrl::SaveReport()
{
	CWinApp* pApp = AfxGetApp();
	int nSheet = GetInfoCount();
	for_to(i, nSheet) {
		if(m_arrInfo[i].m_bReportUpdated) {
			CString sRegSheet;
			sRegSheet.Format(_T("Sheet%d"), i+1);
			pApp->WriteProfileString(sRegSheet, _T("Report"), m_arrInfo[i].m_sReport);
			pApp->WriteProfileInt(sRegSheet, _T("DateLow"), m_arrInfo[i].m_ftOperate.dwLowDateTime);
			pApp->WriteProfileInt(sRegSheet, _T("DateHigh"), m_arrInfo[i].m_ftOperate.dwHighDateTime);
			m_arrInfo[i].m_bReportUpdated = FALSE;
		}
	}
}

void CZCInfoCtrl::LoadHistory(CString sHistName, CStringArray& sArrHist)
{
	CWinApp* pApp = AfxGetApp();
	sArrHist.RemoveAll();
	int nCount = (int)pApp->GetProfileInt(sRegHistory, sHistName + _T("Count"), 0);
	for_to(i, nCount) {
		CString sRegKey;
		sRegKey.Format(sHistName + _T("%d"), i+1);
		CString sHist = pApp->GetProfileString(sRegHistory, sRegKey);
		if(!sHist.IsEmpty())
			sArrHist.Add(sHist);
	}
}

void CZCInfoCtrl::SaveHistory(CString sHistName, CStringArray& sArrHist)
{
	CWinApp* pApp = AfxGetApp();
	pApp->WriteProfileInt(sRegHistory, sHistName + _T("Count"), sArrHist.GetSize());
	for_to(i, sArrHist.GetSize()) {
		CString sRegKey;
		sRegKey.Format(sHistName + _T("%d"), i+1);
		pApp->WriteProfileString(sRegHistory, sRegKey, sArrHist[i]);
	}
}

BOOL CZCInfoCtrl::New()
{
	CZCInfo info;
	m_arrInfo.InsertAt(0, info);
	m_iSel = 0;
	return TRUE;
}


BOOL CZCInfoCtrl::Remove(int i)
{
	if(i <= 0 || i >= GetInfoCount()) return FALSE;
	m_arrInfo.RemoveAt(i);
	if(m_iSel >= GetInfoCount()) m_iSel--;
	return TRUE;
}

#ifndef NO_FTP
BOOL CZCInfoCtrl::GetServer(CString& sServer, CZCInfo* pInfo)
{
	CString sURL;
	if(pInfo == NULL) pInfo = GetInfoPtr(m_iSel);
	if(pInfo->m_option & OPT_SRCFTP) sURL = ParseEnv(pInfo->m_sSrcPath);
	else if(pInfo->m_option & OPT_DSTFTP) sURL = ParseEnv(pInfo->m_sDstPath);
	if(sURL.IsEmpty()) return FALSE;
	CString sObject;
	return ftp.ParseURL(sURL, sServer, sObject);
}
#endif

int CZCInfoCtrl::GetBackupCount(int iSel)
{
	int nCount = 0;
	for_to(i, GetInfoCount()) {
		if(i != iSel && GetInfoPtr(i)->m_option & OPT_BACKUP)
			nCount++;
	}
	return nCount;
}
