/////////////////////////////////////////////////////////////////////////////
// ZCInfo.h

#ifndef ZCINFO_H
#define ZCINFO_H

#include "FTP.h"
#include "zczb.h"

enum OpMode { OPE_UPDATE, OPE_COPY, OPE_SYNC, OPE_MOVE, OPE_DELETE, OPE_GATHER, OPE_COMPARE, OPE_FIND, OPECOUNT };

#define OPT_INCLUDE	0x01
#define OPT_EXCLUDE	0x02
#define OPT_SUBDIR	0x04
#define OPT_HIDDEN	0x08
#define OPT_NOOVW	0x10
#define OPT_OVW		0x20
#define OPT_BYTARGET 0x40
#define OPT_IGNORE_ERROR 0x80
#define OPT_BACKUP   0x100
#define OPT_DAYAFTER 0x200
#define OPT_BYSIZE	0x400	// ### 1.23
#define OPT_SRCFTP  0x1000
#define OPT_DSTFTP  0x2000
#define OPT_FTP (OPT_SRCFTP+OPT_DSTFTP)

#define OPT_DEFAULT OPT_HIDDEN

#define TAB_MAX 32

#define OP_SIZE_LARGE 0
#define OP_SIZE_EQUAL 1
#define OP_SIZE_SMALL 2
#define OP_SIZE_MODEMASK 0x0F

#define OP_SIZE_BYTE 0x00
#define OP_SIZE_KB   0x10
#define OP_SIZE_MB   0x20
#define OP_SIZE_UNITMASK 0xF0
#define OP_SIZE_UNITSHIFT 4

class CZCInfo
{
public:
	CZCInfo();
	~CZCInfo();
public:
	CString m_sLabel;
	CString m_sSrcPath;
	CString m_sDstPath;
	CString m_sIncMask;
	CString m_sExcMask;
	OpMode	m_opmode;
	UINT	m_option;
	CString m_sReport;
	BOOL m_bReportUpdated;
	FILETIME m_ftOperate;
	DWORD   m_dwBySize;
	UINT	m_modeBySize;

	BOOL IsSubDir() { return (m_option & OPT_SUBDIR) != 0; }
	CString GetOpModeString();
	CString GetTitle();
	void SetReport(LPCTSTR msg);
	void SetDate();
	BOOL IsValidDate() { return m_ftOperate.dwLowDateTime || m_ftOperate.dwHighDateTime; }
};

class CZCInfoCtrl
{
public:
	CArray<CZCInfo, CZCInfo&> m_arrInfo;
	int m_iSel;
	CStringArray m_sHistPath;
	CStringArray m_sHistMask;
	CStringArray m_sHistFTP;
	BOOL m_bModified;
public:
	CZCInfoCtrl();
	~CZCInfoCtrl();
	CZCInfo* GetInfoPtr(int i) { return &m_arrInfo[i]; }
	BOOL New();
	BOOL Remove(int i);
	BOOL Load();
	void Save();
	int GetInfoCount() { return m_arrInfo.GetSize(); }
	void Touch() { m_bModified = TRUE; }
#ifndef NO_FTP
	BOOL GetServer(CString& sServer, CZCInfo* pInfo = NULL);
#endif
	void SaveReport();

	int GetBackupCount(int iSel = -1);
private:
	void LoadHistory(CString sHistName, CStringArray& sArrHist);
	void SaveHistory(CString sHistName, CStringArray& sArrHist);
};

extern TCHAR *sOpMode[];

enum DoMode {
	DO_LIST,
	DO_EXEC,
	DO_AUTOEXEC,
};

CString GetDateTimeString(SYSTEMTIME *pST, DWORD dwTimeFlags = 0);

#endif // ZCINFO_H



