// archiver.cpp : implementation of the CArchiver class
//

#include "stdafx.h"
#include "archiver.h"
#ifndef _UNICODE
#include <mbstring.h>
#endif

enum ArcType { ARC_ZIP, ARC_LHA, ARC_RAR, ARC_CAB, ARC_7Z, ARCTYPE_MAX };

enum OptionTable { OPT_DLLNAME, OPT_PROCNAME, OPT_EXTRACT, OPT_EXTRACTALL, OPT_MAX };

static const TCHAR s_szExtTable[] =	_T("zip lzh rar cab 7z jar");

static const TCHAR s_szWildAll[] =	_T("*.*");

static const TCHAR *s_szProcTable[] = {
	_T(""),
	_T("GetCursorMode"),
	_T("SetCursorMode"),
	_T("OpenArchive"),
	_T("CloseArchive"),
	_T("FindFirst"),
	_T("FindNext"),
	_T("SetOwnerWindowEx"),
	_T("KillOwnerWindowEx"),
};

static const TCHAR *s_szOptionTable[ARCTYPE_MAX][OPT_MAX] = {
	{ _T("UnZip"), _T("UnZip"), _T("-x --i -j -o -qq"), _T("-x -o") },
	{ _T("Unlha"), _T("Unlha"), _T("e -n -c1 -m1"), _T("x -c1 -r2") },
	{ _T("Unrar"), _T("Unrar"), _T("-e -o"), _T("-x -o") },
	{ _T("Cab"),   _T("Cab"),   _T("-x -o -i -j"), _T("-x -o") },
	{ _T("7-zip"), _T("SevenZip"), _T("e -o"), _T("x -o") }
};

typedef int	 (WINAPI* LP_CMDLINE)(const HWND, LPCSTR, LPSTR, const DWORD);
typedef BOOL (WINAPI* LP_GET_CURSOR_MODE)();
typedef BOOL (WINAPI* LP_SET_CURSOR_MODE)(const BOOL);
typedef HARC (WINAPI* LP_OPEN_ARCHIVE)(const HWND, LPCSTR, const DWORD);
typedef int	 (WINAPI* LP_CLOSE_ARCHIVE)(HARC);
typedef int  (WINAPI* LP_FIND_FIRST)(HARC , LPCSTR, INDIVIDUALINFO*);
typedef int  (WINAPI* LP_FIND_NEXT)(HARC, INDIVIDUALINFO*);
typedef BOOL (WINAPI* LP_SET_OWNER_WINDOW_EX)(HWND, LPARCHIVERPROC);
typedef BOOL (WINAPI* LP_KILL_OWNER_WINDOW_EX)(HWND);

#define ArcProc(name) reinterpret_cast<LP_##name##>(m_proc[ARC_##name##])

CArchiver::CArchiver()
{
	m_hDll = NULL;
	m_hArc = NULL;
	m_dwError = 0;
	m_lpszExtTable = s_szExtTable;
	m_nArcType = -1;
	m_szExtractDir[0] = TCHAR('\0');
}

CArchiver::~CArchiver()
{
	if(m_hDll) FreeDll();
}

void CArchiver::SetExtTable(LPCTSTR lpszExtTable)
{
	m_lpszExtTable = lpszExtTable ? lpszExtTable : s_szExtTable;
}

int CArchiver::GetArcType(LPCTSTR lpszFileName)
{
	if(lpszFileName == NULL) return -1;
	LPCTSTR lpszExt = _tcsrchr(lpszFileName, TCHAR('.'));
	if(lpszExt) lpszExt++;
	else lpszExt = lpszFileName;
	int len = _tcslen(lpszExt);
	if(len == 0) return -1;

	int index = 0;
	LPCTSTR p = m_lpszExtTable;
	while(*p) {
		if(_tcsnicmp(p, lpszExt, len) == 0) {
			if(*(p + len) <= TCHAR(' ')) {
				if (index >= ARCTYPE_MAX) { // jar
					index = ARC_ZIP;
				}
				return index;
			}
		}
		while(*p && *p != TCHAR(' ')) p++;
		while(*p == TCHAR(' ')) p++;
		index++;
	}
	return -1;
}

BOOL CArchiver::LoadDll(LPCTSTR lpszFileName)
{
	int nArcType = GetArcType(lpszFileName);
	if(nArcType == -1) return FALSE;
	if(m_hDll && nArcType == m_nArcType) return TRUE;

	FreeDll();

	TCHAR szDllName[32];
	_tcscpy(szDllName, s_szOptionTable[nArcType][OPT_DLLNAME]);
	_tcscat(szDllName, _T("32.DLL"));

	if((m_hDll = ::LoadLibrary(szDllName)) == NULL) return FALSE;
	TRACE(_T("CArchiver: Load \"%s\" OK\n"), szDllName);
	m_nArcType = nArcType;

	TCHAR szProcName[64];
	LPCTSTR pszPrefix = s_szOptionTable[nArcType][OPT_PROCNAME];
	int len = _tcslen(pszPrefix);
	_tcsncpy(szProcName, pszPrefix, len);
	for(int i = 0; i < ARC_PROC_MAX; i++) {
		_tcscpy(szProcName + len, s_szProcTable[i]);
		m_proc[i] = ::GetProcAddress(m_hDll, CW2A(szProcName));
		if(m_proc[i] == NULL)
			TRACE(_T("CArchiver: [%s] not found\n"), szProcName);
	}
	return TRUE;
}

void CArchiver::FreeDll()
{
	if(m_hDll) {
		::FreeLibrary(m_hDll);
		m_hDll = NULL;
	}
}

BOOL CArchiver::SetError(int result,  LPCTSTR lpszFunc)
{
	if((m_dwError = result) == 0) return TRUE;
	TRACE(_T("CArchiver: *** %s Error(%X)\n"), lpszFunc, result);
	return FALSE;
}

HARC CArchiver::OpenArchive(HWND hWnd, LPCTSTR lpszArcName, UINT uMode)
{
	if(m_hArc) CloseArchive();
	m_hArc = ArcProc(OPEN_ARCHIVE)(hWnd, CW2A(lpszArcName), uMode);
	if(m_nArcType == ARC_ZIP && m_szExtractDir[0] != TCHAR('\0')) {	// UnZIP32 bug...
		::RemoveDirectory(m_szExtractDir);
	}
	return m_hArc;
}

void CArchiver::CloseArchive()
{
	if(m_hArc) {
		ArcProc(CLOSE_ARCHIVE)(m_hArc);
		m_hArc = NULL;
	}
}

int CArchiver::FindFirst(LPCTSTR lpszWildName)
{
	if(lpszWildName == NULL) lpszWildName = s_szWildAll;
	int result = ArcProc(FIND_FIRST)(m_hArc, CW2A(lpszWildName), &m_info);
	if(result == -1) return -1;
	return SetError(result, s_szProcTable[ARC_FIND_FIRST]);
}

int CArchiver::FindNext()
{
	int result = ArcProc(FIND_NEXT)(m_hArc, &m_info);
	if(result == -1) return -1;
	return SetError(result, s_szProcTable[ARC_FIND_NEXT]);
}


LPCTSTR CArchiver::GetFileName()
{
#ifndef _UNICODE
	return m_info.szFileName;
#else
	::MultiByteToWideChar(CP_ACP, 0, m_info.szFileName, -1, m_szFileName, FNAME_MAX32);
	return m_szFileName;
#endif
}

int CArchiver::GetFileFolderLength()
{
	LPCTSTR pName = GetFileName();
	LPCTSTR pSlash = _tcsrchr(pName, TCHAR('/'));
	if(!pSlash)
		pSlash = (LPCTSTR)_mbsrchr((const UCHAR*)pName, TCHAR('\\'));
	return pSlash ? pSlash - pName + 1 : 0;
}

void CArchiver::GetFileTime(FILETIME& ft)
{
	FILETIME ftLocal;

	::DosDateTimeToFileTime(GetFileDosDate(), GetFileDosTime(), &ftLocal);
	::LocalFileTimeToFileTime(&ftLocal, &ft);
}

WORD CArchiver::GetFileDosAttr()
{
	WORD wAttr = 0;
	for_to(i, 8) {
		if(m_info.szAttribute[i] == '\0') break;
		switch(m_info.szAttribute[i]) {
			case 'R': wAttr |= FA_RDONLY; break;
			case 'H': wAttr |= FA_HIDDEN; break;
			case 'S': wAttr |= FA_SYSTEM; break;
			case 'L': wAttr |= FA_LABEL; break;
			case 'D': wAttr |= FA_DIREC; break;
			case 'A': wAttr |= FA_ARCH; break;
		}
	}
	return wAttr;
}

DWORD CArchiver::GetFileAttributes()
{
	DWORD dwAttr = 0;
	for_to(i, 8) {
		if(m_info.szAttribute[i] == '\0') break;
		switch(m_info.szAttribute[i]) {
			case 'R': dwAttr |= FILE_ATTRIBUTE_READONLY; break;
			case 'H': dwAttr |= FILE_ATTRIBUTE_HIDDEN; break;
			case 'S': dwAttr |= FILE_ATTRIBUTE_SYSTEM; break;
			case 'D': dwAttr |= FILE_ATTRIBUTE_DIRECTORY; break;
			case 'A': dwAttr |= FILE_ATTRIBUTE_ARCHIVE; break;
		}
	}
	return dwAttr;
}

BOOL CArchiver::SetArchive(HWND hWnd, LPCTSTR lpszArcName, UINT uMode)
{
	m_hWnd = hWnd;
	_tcscpy(m_szArcName, lpszArcName);
	// ArcProc(SET_OWNER_WINDOW_EX)(m_hWnd, NULL);
	return TRUE;
}

BOOL CArchiver::Extract(LPCTSTR lpszFileName, LPCTSTR lpszDirName)
{
	char buffer[512];
	char log[512];
	log[0] = 0;
#ifndef _UNICODE
	wsprintf(buffer, _T("%s \"%s\" \"%s\" \"%s\""),
		s_szOptionTable[m_nArcType][OPT_EXTRACT], m_szArcName, lpszDirName, lpszFileName);
#else
	sprintf(buffer, "%S \"%S\" \"%S\" \"%S\"",
		s_szOptionTable[m_nArcType][OPT_EXTRACT], m_szArcName, lpszDirName, lpszFileName);
#endif
	TRACE(_T("%S\n"), buffer);

	ArcProc(SET_OWNER_WINDOW_EX)(m_hWnd, NULL);
	//ArcProc(SET_CURSOR_MODE)(FALSE);
	int result = ArcProc(CMDLINE)(m_hWnd, buffer, log, 500);
	ArcProc(KILL_OWNER_WINDOW_EX)(m_hWnd);

	TRACE(_T("%S"), log);
	return SetError(result, _T("Extract"));
}

BOOL CArchiver::ExtractAll(LPCTSTR lpszDirName)
{
	if(m_nArcType == ARC_ZIP)
		_tcsncpy(m_szExtractDir, lpszDirName, MAX_PATH - 1);

	char buffer[512];
#ifndef _UNICODE
	wsprintf(buffer, _T("%s \"%s\" \"%s\""),
		s_szOptionTable[m_nArcType][OPT_EXTRACTALL], m_szArcName, lpszDirName);
#else
	sprintf(buffer, "%S \"%S\" \"%S\"",
		s_szOptionTable[m_nArcType][OPT_EXTRACTALL], m_szArcName, lpszDirName);
#endif
	TRACE(_T("%S\n"), buffer);
	int result = ArcProc(CMDLINE)(m_hWnd, buffer, NULL, 0);
	return SetError(result, _T("ExtractAll"));
}
