// filename.cpp : implementation of the CFileName class
//

#include "stdafx.h"
#include "filename.h"
#ifndef _UNICODE
#include <mbstring.h>
#endif

UINT CFileName::GetElementBits()
{
	UINT bits;
	SplitPath(0, &bits);
	return bits;
}

CString CFileName::SplitPath(UINT fnbits, UINT* pbits)
{
	TCHAR sDrive[_MAX_DRIVE], sDir[_MAX_DIR], sFname[_MAX_FNAME], sExt[_MAX_EXT];
	_tsplitpath(GetString(), sDrive, sDir, sFname, sExt);
	if(pbits) {
		UINT bits = 0;
		if(*sDrive) bits |= FN_DRV;
		if(*sDir)   bits |= FN_DIR;
		if(*sFname) bits |= FN_TITLE;
		if(*sExt)   bits |= FN_EXT;
		*pbits = bits;
	}
	CString sResult;
	if(fnbits & FN_DRV) sResult += sDrive;
	if(fnbits & FN_DIR) sResult += sDir;
	if(fnbits & FN_TITLE) sResult += sFname;
	if(fnbits & FN_EXT) sResult += sExt;
	return sResult;
}

int CFileName::CompareMask(CString& sMask, BOOL bDir, UINT uFlags)	// ###1.20
{
	if(bDir)
		return MatchWildcard(sMask, bDir);
	else {
		if(uFlags == 0 || uFlags & (MBST_WILDCARD|MBST_PATH)) {
			return (!sMask.IsEmpty()
				&& !CompareMaskBit((CFileName&)sMask, FN_ROOT)
				&& !CompareMaskBit((CFileName&)sMask, FN_TITLE)
				&& !CompareMaskBit((CFileName&)sMask, FN_EXT)) ? 0 : INVALID;
		}
		CFileName sTitle = SplitPath(FN_TITLE);
#ifndef _UNICODE
		return FindMBString(sTitle, sMask, uFlags) ? 0 : INVALID;
#else
		return sTitle.MakeLower().Find(sMask) != -1 ? 0 : INVALID;
#endif
	}
}

int CFileName::CompareMaskBit(CFileName& sMask, UINT fnbits)
{
	CString sMask1 = sMask.SplitPath(fnbits);
	if(sMask1.IsEmpty() && fnbits != FN_EXT) return 0;
	CFileName sPath = SplitPath(fnbits);
	return sPath.MatchWildcard(sMask1);
}

int CFileName::MatchWildcard(LPCTSTR pMask, BOOL bDir)
{
	LPCTSTR pPath = GetString();
	TCHAR cMsk, c; // ###_UNICODE
	while(cMsk = *pMask++) {
		if (cMsk == TCHAR('*')) { // ###050630 *file* Ή
			if(!*pMask) return 0;
			TCHAR szPath[MAX_PATH];
			TCHAR szMask[MAX_PATH];
			_tcscpy(szPath, pPath);
			_tcslwr(szPath);
			_tcscpy(szMask, pMask);
			_tcslwr(szMask);
			TCHAR *p = szMask;
			while (*p && *p != TCHAR('*')) p++;
			TCHAR cEnd = *p;
			*p = TCHAR('\0');
			p = _tcsstr(szPath, szMask);
			if(!p) return INVALID;
			return (cEnd != TCHAR('\0') || *(p + _tcslen(szMask)) == TCHAR('\0')) ? 0 : INVALID;
		}
		else if (cMsk == TCHAR('?')) {
			pPath++;
			if (*pPath == TCHAR('\0'))
				return (*pMask == TCHAR('\0')) ? 0 : INVALID;
		} else {
			c = *pPath++;
			if(toupper(cMsk) != toupper(c)) return INVALID;
		}
	}
	return (*pPath == TCHAR('\0') || bDir && IsSlash(*pPath)) ? 0 : INVALID;
}

BOOL CFileName::IsTailSlash()	// ###0.92
{
	if(!IsEmpty()) {
#ifndef _UNICODE
		const UCHAR* p = (const UCHAR*)GetString();
		const UCHAR* pTail = p + GetLength() - 1;
		if(*pTail == '/') return TRUE;
		if(*pTail == '\\') {
			const UCHAR* p1 = _mbsrchr((const UCHAR*)p, UINT('\\'));
			if(p1 == pTail) return TRUE;
		}
#else
		TCHAR c = GetAt(GetLength() - 1);
		return c == TCHAR('/') || c == TCHAR('\\');
#endif
	}
	return FALSE;
}

void CFileName::AddSlash(BOOL bSlash)
{
	if(!IsTailSlash())
		operator+=(bSlash ? TCHAR('/') : TCHAR('\\'));
}

BOOL CFileName::DeleteSlash()
{
	if(IsTailSlash()) {
		LPTSTR p = GetBuffer(0);
		*(p + GetLength() - 1) = TCHAR('\0');
		ReleaseBuffer();
		return TRUE;
	}
	return FALSE;
}

void CFileName::SlashToBackslash()
{
	LPTSTR p = GetBuffer(0);
	while(*p) {
		if(*p == TCHAR('/')) *p = TCHAR('\\');
		p++;
	}
	ReleaseBuffer();
}

void CFileName::BackslashToSlash()
{
	LPTSTR p = GetBuffer(0);
	while(*p) {
#ifndef _UNICODE
		if(_istlead(*p)) p++;
		else {
			if(*p == TCHAR('\\')) *p = TCHAR('/');
		}
#else
		if (*p == TCHAR('\\')) *p = TCHAR('/');
#endif
		p++;
	}
	ReleaseBuffer();
}

BOOL CFileName::GetParentDirectory()
{
	BOOL bResult = FALSE;
	int nLen = GetLength();
	LPTSTR p = GetBuffer(0);
#ifndef _UNICODE
	LPTSTR p1 = (LPTSTR)_mbsrchr((const UCHAR*)p, UINT('\\'));
	LPTSTR p2 = (LPTSTR)_mbsrchr((const UCHAR*)p, UINT('/'));
#else
	LPTSTR p1 = _tcsrchr(p, TCHAR('\\'));
	LPTSTR p2 = _tcsrchr(p, TCHAR('/'));
#endif
	if(p2 > p1) p1 = p2;
	if(p1) {
		if((p1 - p) == 2) p1++;		// ###1.20 "C:"->"C:\"
		*p1 = '\0';
		bResult = TRUE;
	}
	ReleaseBuffer();
	if(GetLength() == nLen) return FALSE;
	return bResult;
}

BOOL CFileName::IsRoot()	// ###1.20
{
	return GetLength() >= 2 && GetLength() <= 3 && GetAt(1) == TCHAR(':');
}

#ifdef _UNICODE
UINT GetMBStringType(LPCTSTR p)
{
	UINT uFlags = MBST_SB;
	while (*p) {
		TCHAR c = *p++;
		if (c == TCHAR('*') || c == TCHAR('?')) uFlags |= MBST_WILDCARD;
		else if (c == TCHAR('.') || c == TCHAR('/') || c == TCHAR(':') || c == TCHAR('\\')) uFlags |= MBST_PATH;
	}
	return uFlags;
}
#endif