// filename.h : interface of the CFileName class
//

#ifndef FILENAME_H
#define FILENAME_H

#ifdef _UNICODE
#define MBST_SB			0x0001
#define MBST_WILDCARD	0x0002	// '*' '?'
#define MBST_PATH		0x0004	// ':' '/' '\\' '.'
#else
#include "zctype.h"
#endif

enum FileNameBits {
	FN_DRV = 1,
	FN_DIR = 2,
	FN_ROOT = 3,
	FN_TITLE = 4,
	FN_EXT = 8,
	FN_NAME = 12,
	FN_FULLPATH = 15,
};

class CFileName : public CString
{
public:
	CFileName() { }
	CFileName(const CString& stringSrc) : CString(stringSrc) { }
	CFileName(LPCTSTR lpsz) : CString(lpsz) { }
	CFileName(const unsigned char* psz) : CString(psz) { }
	~CFileName() { }

	UINT GetElementBits();
	CString GetDrive() { return SplitPath(FN_DRV); }
	CString GetDir() { return SplitPath(FN_ROOT); }
	CString GetName() { return SplitPath(FN_NAME); }
	CString GetTitle() { return SplitPath(FN_TITLE); }
	CString GetExt() { return SplitPath(FN_EXT); }
	CString GetNoExt() { return SplitPath(FN_ROOT|FN_TITLE); }
	int CompareMask(CString& sMask, BOOL bDir = FALSE, UINT uFlags = 0);
	int MatchWildcard(LPCTSTR pMask, BOOL bDir = FALSE);
	BOOL IsTailSlash();
	static BOOL IsSlash(TCHAR c) { return c == TCHAR('\\') || c == TCHAR('/'); }
	void AddSlash(BOOL bSlash = FALSE);
	BOOL DeleteSlash();
	void SlashToBackslash();
	void BackslashToSlash();
	BOOL GetParentDirectory();
	int ComparePath(LPCTSTR pPath) { return CompareNoCase(pPath); }
	BOOL IsRoot();	// ###1.20
protected:
	CString SplitPath(UINT fnbits, UINT* pbits = NULL);
	int CompareMaskBit(CFileName& sMask, UINT fnbits);
};

#ifdef _UNICODE
UINT GetMBStringType(LPCTSTR p);
#endif

#endif
