/////////////////////////////////////////////////////////////////////////////
// zctype.h  (03.08.26 ver. for ZBrowse)

#ifndef _ZCTYPE_H_
#define _ZCTYPE_H_

#pragma warning(disable: 4244)

#ifndef __CHAR_DEFINED
#define __CHAR_DEFINED
typedef unsigned char _CHAR;
typedef _CHAR *LPTEXT;
typedef const _CHAR *LPCTEXT;
#endif //__CHAR_DEFINED

#define _CHAR_TAB	'\t'
#define _CHAR_LF	'\n'
#define _CHAR_CR	'\r'
#define _CHAR_SPC	' '
#define _CHAR_EOF	0x1A
#define _CHAR_MBSPC	0x8140
#define _CHAR_PAGE	0x0C
#define _CHAR_CRLF  0x0A0D
#define _CHAR_LFCR  0x0D0A
#define MBCHAR_SPACE 0x4081
#define MBCHAR_LF	 0x0A0D
#define STR_MBSPC	 "@"

#define CT_NULL			0
#define CT_CTRL			0x01
#define CT_BLANK		0x02
#define CT_SYMBOL		0x04
#define CT_URL			0x08
#define CT_UPPER		0x10
#define CT_LOWER		0x20
#define CT_DIGIT		0x40
#define CT_UNDERSCORE	0x80
#define CT_KATAS		0x100
#define CT_KATAL		0x200
#define CT_KATA			0x400
#define CT_MBLEAD		0x800

#define CT_ALPHA		(CT_UPPER|CT_LOWER)
#define CT_ALPHADIGIT	(CT_ALPHA|CT_DIGIT)
#define CT_LABEL		(CT_ALPHA|CT_DIGIT|CT_UNDERSCORE)
#define CT_LABELHEAD	(CT_ALPHA|CT_UNDERSCORE)
#define CT_SYMURL		(CT_SYMBOL|CT_URL)
#define CT_WORD			0xEEF0

#define MBCT_SPACE	0x0001
#define MBCT_BREAK	0x0002
#define MBCT_SYMBOL	0x0004
#define MBCT_TRAIL	0x0008
#define MBCT_UPPER	0x0010
#define MBCT_LOWER	0x0020
#define MBCT_DIGIT	0x0040
#define MBCT_ALPHA	(MBCT_UPPER|MBCT_LOWER)
#define MBCT_ALPHADIGIT	(MBCT_ALPHA|MBCT_DIGIT)
#define MBCT_ASCII	0x0040	// ASCII(0x20`0x7F) for MBST

#define MBCT_HIRA	0x0100
#define MBCT_HIRA_D	0x0200	// 
#define MBCT_HIRA_H	0x0400	// 
#define MBCT_HIRA_S	0x0800	// X

#define MBCT_KATA	0x1000
#define MBCT_KATA_D	0x2000
#define MBCT_KATA_H	0x4000
#define MBCT_KATA_S	0x8000

#define MBCT_KANJI	0x10000

#define MIN_MBCHAR	0x8140
#define MAX_MBCTABLE ''
#define MBCTABLE_COUNT (MAX_MBCTABLE + 1 - MIN_MBCHAR)

#define MIN_KANJI ''
#define MIN_HIRA  ''
#define MAX_HIRA  ''
#define MIN_KATA  '@'
#define MAX_KATA  ''
#define MIN_KANJI_H 0x88

extern WORD _zctype[];
extern WORD* _pzmbctype;

#define SWAPBYTE(c) __asm {__asm mov ax, c __asm xchg al, ah __asm mov c, ax}

inline WORD char_type(_CHAR c) { return _zctype[c]; }
inline BOOL is_upper(_CHAR c)	{ return c >='A' && c <= 'Z'; }
inline BOOL is_lower(_CHAR c)	{ return c >='a' && c <= 'z'; }
inline BOOL is_alpha(_CHAR c)	{ return (_zctype[c] & CT_ALPHA) != 0; }
inline BOOL is_digit(_CHAR c)	{ return c >= '0' && c <= '9'; }
inline BOOL is_alphadigit(_CHAR c) { return (_zctype[c] & CT_ALPHADIGIT) != 0; }
inline BOOL is_mblead(_CHAR c)	{ return (_zctype[c] & CT_MBLEAD) != 0; }
inline BOOL is_hankana(_CHAR c)	{ return c >=0xA1 && c <= 0xDF; }

inline WORD& mbchar_table(WORD c) { return _pzmbctype[c - MIN_MBCHAR]; }
inline UINT mbchar_type(WORD c) { return (c < MAX_MBCTABLE) ? mbchar_table(c) : c >= MIN_KANJI ? MBCT_KANJI : MBCT_SYMBOL; }
inline BOOL is_mbchar(WORD c) { return (c & 0xFF00) != 0; }
inline BOOL is_mbc_kata(WORD c) { return c >= MIN_KATA && c <= MAX_KATA; }
inline BOOL is_mbc_hira(WORD c) { return c >= MIN_HIRA && c <= MAX_HIRA; }
inline BOOL is_mbc_kanji(WORD c) { return c >= MIN_KANJI; }
inline BOOL is_mbc_alpha(WORD c) { return c >= '`' && c <= ''; }

AFX_EXT_CLASS CString HanToZen(LPCTSTR pHan, int len = 0, BOOL bLower = FALSE);
WORD HanToZen(_CHAR c);
UINT GetMBCharType(LPCTSTR p);
inline WORD GetMBChar(LPCTSTR p) { return MAKEWORD(*(p+1), *p); }	// GetMBChar(p++)̓_!!
AFX_EXT_CLASS LPCSTR FindMBString(LPCSTR str1, LPCSTR str2, UINT uFlags = 0);

// MBC String Type

#define MBST_SB			0x0001
#define MBST_WILDCARD	0x0002	// '*' '?'
#define MBST_PATH		0x0004	// ':' '/' '\\' '.'
#define MBST_MB			0x0100
#define MBST_MB_ASCII	0x0200

AFX_EXT_CLASS UINT GetMBStringType(LPCTSTR p);

#endif //_ZCTYPE_H_
